#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/sendfile.h>
#include <sys/wait.h>
#include <sys/sysinfo.h>
#include <signal.h>
#include <time.h> /*DB!*/ 
#include <readline/readline.h>
#include <readline/history.h>

#define __RDWR_BUFFER_SIZE__ 4096
#include <neurocad/engine/libnce.h>
#include <neurocad/engine/nce_common.h>
#include <neurocad/engine/asmmanager.h>
#include <neurocad/assemblers/basic/asmbasic.h>
#include "getparam.m"

static int ch_pid = -1, decoder_pid = -1, main_pid = -1;
static sig_atomic_t engine_ret, stop_signal;

static nce_context nce_ctx;


void nce_ret ( int signal )
{
	engine_ret = 1;
	return;
}

void nce_sim_stop ( int signal )
{
/*DB*/	fprintf(stderr,"nce_sim_stop\n");
	stop_signal = 1;
	return;
}

void nce_deaded ( int signal )
{
	engine_ret  = 2;
	stop_signal = 1;
}

static size_t nce_rdwr(int hto, int hin, size_t size)
{
	char buffer[__RDWR_BUFFER_SIZE__];
	size_t all_sr = 0, curr_sr = 0, cnt;
	int wr_ret, rd_ret;
	while(all_sr != size){
		curr_sr = ( (size - all_sr) < __RDWR_BUFFER_SIZE__ ) ?  (size - all_sr) : __RDWR_BUFFER_SIZE__;
		cnt = 0;
		while( cnt != curr_sr){
			rd_ret = read(hin, buffer + cnt, curr_sr - cnt);
			if(rd_ret < 0)	return all_sr;
			cnt += rd_ret;
		}
		cnt = 0;
		while( cnt != curr_sr){
			wr_ret = write(hto, buffer + cnt, curr_sr - cnt);
			if(wr_ret < 0)	return all_sr;
			cnt += wr_ret;
		}
		all_sr += curr_sr;
	}
	return all_sr;
}

static int new_engine()
{
	fprintf(stderr,"_ncc_:Engine is dead!\n_ncc_:Restart Engine...\n");
	int result;
	waitpid(ch_pid, &result, 0);
	if( (ch_pid = fork()) == 0 ){
		asm_drop();
		asm_model_grow_pointer 	= &asm_model_grow;
		asm_messenger_pointer	= &asm_messenger;
		asm_printinfo_pointer	= &asm_printinfo;
		nce( &nce_ctx );
		exit (0);
	}
	return 0;
}

static int binary_decoder(int hfile)
{
	char cmd[COMMAND_CONTEXT_SIZE];
	char message[__MAX_MESSAGE_LENGTH__];
	command_header command;
	command.command = cmd;
	command.length  = (int*)(cmd+1);
	
	char empty;
	int i_value;
	double d_value;
	
	while( read(hfile, cmd, COMMAND_CONTEXT_SIZE) == COMMAND_CONTEXT_SIZE ){
		switch(*command.command){
			case NC__RET_MESSAGE:
				memset(message,0,__MAX_MESSAGE_LENGTH__);
				web_read(hfile, message, *command.length);
				fprintf(stderr,"%s",message);
				break;
			case NC__RET_DEBUG:
				memset(message,0,__MAX_MESSAGE_LENGTH__);
				read(hfile, message, *command.length);
				fprintf(stderr,"%s",message);
				break;
			case NC__RET_OK:
				while( ((*command.length)-- ) > 0 )
					read(hfile, &empty, sizeof(char) );
				fprintf(stderr,"_ncc_:OK!\n");
				kill(main_pid, SIGUSR1);
				break;
			case NCE_RET_DEAD:
				kill(main_pid, SIGALRM);
				return 0;
				break;
			case NC__RET_ERROR:
				memset(message,0,__MAX_MESSAGE_LENGTH__);
				read(hfile, message, *command.length);
				fprintf(stderr,"%s",message);
				fprintf(stderr,"_ncc_:ERROR!\n");
				kill(main_pid, SIGUSR1);
				kill(main_pid, SIGUSR2);
				break;
			case NC__RET_WARNING:
				memset(message,0,__MAX_MESSAGE_LENGTH__);
				read(hfile, message, *command.length);
				fprintf(stderr,"%s",message);
				fprintf(stderr,"_ncc_:WARNING!\n");
				kill(main_pid, SIGUSR1);
				kill(main_pid, SIGUSR2);
				break;
			case NCE_RET_SYSTEM_GET_PID:
				if( *command.length != sizeof( int ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
					
				}
				else{
					read(hfile, &i_value, sizeof(int));
					fprintf(stderr,"_ncc_:Engine PID: %d\n",i_value);
				}
				break;
			case ASMBASIC_RET_DATA_GET:
				if( (*command.length) != sizeof( double ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
				}
				else{
					read(hfile, &d_value, sizeof(double));
					fprintf(stderr,"_ncc_:Data is: %G\n",d_value);
				}
				break;
			case NCE_RET_SIMULATION_IS_END:
				if( (*command.length) != sizeof( int ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
				}
				else{
					read(hfile, &i_value, sizeof(int));
					fprintf(stderr,"\r_ncc_:Simulation is completed in %f cpu secconds\n_ncc_>", ((double)i_value)*1e-3);
					kill(main_pid, SIGUSR2);
				}
				break;
			case NCE_RET_SIMULATION_GET_MODEL_TIME:
				if( (*command.length) != sizeof( double ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
				}
				else{
					
					read(hfile, &d_value, sizeof(double));
					fprintf(stderr,"_ncc_:Current model time is: %G\n",d_value);
				}
				break;
			case NCE_RET_SIMULATION_GET_MODEL_TIMESTEP:
				if( (*command.length) != sizeof( double ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
				}
				else{
					read(hfile, &d_value, sizeof(double));
					fprintf(stderr,"_ncc_:Current model timestep is: %G\n",d_value);
				}
				break;
			case NCE_RET_WATCHER_ADD:
				if( *command.length != sizeof( int ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
				}
				else{
					read(hfile, &i_value, sizeof(int));
					fprintf(stderr,"_ncc_:Watcher ID: %d\n",i_value);
				}
				break;
			case ASMBASIC_RET_WATCHER_ITEM_ADD_ID:
				if( (*command.length) != sizeof( int ) ){
					while( ((*command.length)-- ) > 0 )
						read(hfile, &empty, sizeof(char) );
				}
				else{
					read(hfile, &i_value, sizeof(int));
					fprintf(stderr,"_ncc_:Watcher Item ID is: %d\n",i_value);
				}
				break;
			default:
				while( ((*command.length)-- ) > 0 )
					read(hfile, &empty, sizeof(char) );
				fprintf(stderr,"_ncc_:Unknown command\n");
		}
	}
	return 0;
}

static int usage(char *progname)
{
	fprintf(stderr,"\nUSAGE: %s [MODE] [OPTIONS]\n\nMODE:\n",progname);
	fprintf(stderr,"\t%s                         \t\t- for simulation in terminal mode\n",progname);
	fprintf(stderr,"\t%s isf=<input script file> \t\t- for simulation in  script mode\n\nOPTIONS:\n",progname);
	fprintf(stderr,"\twbf=<watcher binary file> \t\t- for set the watcher output file\n\n");
	fprintf(stderr,"%s -h or --help              \t\t- this message\n",progname);
	return 0;
}

static int help()
{
	/*
	TODO: ReWrite using readline function. 
	TODO: Clean Disply after each help list.
	TODO: Make `help general' chousing help page
	*/	
	fprintf(stderr,"\nGENERAL COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"quit\n\texit from program\n-----\n");
	fprintf(stderr,"pid\n\tget pid of neurocad engine\n-----\n");
//	fprintf(stderr,"msg <filename>\n\tredirect message stream to <filename> - file or $<filename> -socket\n-----\n");
//	fprintf(stderr,"dbmsg <filename>\n\tredirect debug stream to <filename> - file \n\tor $<filename> -socket \n-----\n");
	fprintf(stderr,"\n\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nMODEL MANIPULATION COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"upload <file name>\n\tload XML file <file name> with model and init it\n-----\n");
	fprintf(stderr,"reinit\n\treinit loaded model\n-----\n");
	fprintf(stderr,"release\n\trelease current model\n-----\n");
	fprintf(stderr,"destroy\n\tdestroy current model and reset engine\n-----\n");
	fprintf(stderr,"save <file name>\n\tsave current model condition in file\n-----\n");
	fprintf(stderr,"restore <file name>\n\trestore model condition from file\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nSIMULATION COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"step\n\tone step of simulation\n-----\n");
	fprintf(stderr,"start ### \n\tstart simulation and go #### times\n-----\n");
	fprintf(stderr,"go ###\n\tstart simulation and go #### steps\n-----\n");
	fprintf(stderr,"until ###\n\tstart simulation and go untile model time < ####\n-----\n");
	fprintf(stderr,"stop\n\tstop simulation\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nDEBUG SIMULATION COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"dbit\n\tone iteration\n-----\n");
	fprintf(stderr,"dbmodit  ###\n\tone iteration for module with id ###\n-----\n");
	fprintf(stderr,"dbnextmodit\n\tnext iteration for current module \n-----\n");
	fprintf(stderr,"getsimtime\n\tget simulation time\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nSIMULATION TIME AND STEP COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"setsimtime ###\n\tset up simulation time in ### value\n-----\n");
	fprintf(stderr,"getstepsim\n\tget simulation time step\n-----\n");
	fprintf(stderr,"setstepsim ###\n\tset up simulation time step in ### value\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nDATA MANIPULATION COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"get ### ***\n\tget data for ### object, *** name id\n-----\n");
	fprintf(stderr,"set ### *** +++\n\tset data for ### object, *** name id in +++ value\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nWATCHER(stream data reading) COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"watcher ###\n\tadd watcher with ### period. Use it if you set 'wbf=' item \n\tthrough starting ncc. (see ./ncc --help)\n-----\n");
	fprintf(stderr,"watcherfile ### <filename>\n\tadd watcher with ### period and redirect output\n\tto <filename> - file or $<filename> -socket\n-----\n");
	fprintf(stderr,"watcherkill ###\n\tkill watcher with ### id\n-----\n");
	fprintf(stderr,"watcherkillall\n\tkill all watchers\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nWATCHER(stream data reading) COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"itemadd  ### *** +++\n\tadd item to ### watcher and set new item to *** object +++ name\n-----\n");
	fprintf(stderr,"itemminimal  ### *** +++ \n\tadd item to ### watcher and set new item to *** object +++ name.\n\tWatcher will write minimal value of +++ name.\n-----\n");
	fprintf(stderr,"itemmaximal  ### *** +++\n\tadd item to ### watcher and set new item to *** object +++ name\n\tWatcher will write maximal value of +++ name.\n-----\n");
	fprintf(stderr,"\titemaverage  ### *** +++\n\tadd item to ### watcher and set new item to *** object +++ name\n\tWatcher will write average value of +++ name.\n-----\n");
	fprintf(stderr,"itemkill ### *** +++\n\tkill item to ### watcher with *** object +++ name\n-----\n");
	fprintf(stderr,"\n\nPress ENTER key for continue or print q and ENTER for quit from help\n");
	if( rl_read_key() == 'q' ) return 0;
	fprintf(stderr,"\nINFORMATION AND HELP COMMANDS: \n------------------------------------------\n");
	fprintf(stderr,"info <flags>\n\tprint information. <flags> can be: asm, pars, mod, obj, sim, mem, watch\n-----\n");
	fprintf(stderr,"help\n\tshow this message\n-----\n");
	return 0;
}



static int terminal(char *progname, FILE* console, int hOut)
{
	char *command = NULL, filename[1024], *command_shift;
	char cmd[COMMAND_CONTEXT_SIZE];
	command_header hc;
	hc.command = cmd;
	hc.length  = (int*)(cmd+1);
	
	struct timespec tsv;
	tsv.tv_sec = 0;
	tsv.tv_nsec = 500000;
	
	rl_catch_signals = rl_catch_sigwinch = 0;
	rl_instream = console;
	int i_value, ii_value, iii_value;
	double d_value;
	fprintf(stderr,"\n===============================================================");
	fprintf(stderr,"\n                    NeuroCAD console application.");
	fprintf(stderr,"\n                               2004 - 2007");
	fprintf(stderr,"\n                               VERSION %s", APP_VERSION);
	fprintf(stderr,"\n===============================================================");
	fprintf(stderr,"\n Copyright(C) Ruben Tikidji - Hanburyan (rth@nisms.krinc.ru)");
	fprintf(stderr,"\n Copyright(C) Sergey Markin             (markin@nisms.krinc.ru)");
	fprintf(stderr,"\n                           Licence: GPL");
	fprintf(stderr,"\n                     Use it without any wary!!!\n\n");

/*	fprintf(stderr,"_ncc_:LOAD BASIC ASSMEBLER: ......\n");
	engine_ret = 0;
	*(hc.command) = NCE_ASSEMBLER_LOAD;
	*(hc.length) = strlen(assname)+1;
	write(hOut, cmd, COMMAND_CONTEXT_SIZE);
	write(hOut, assname, strlen(assname)+1);
	while(engine_ret == 0)sleep(1);
	if(engine_ret == 2){
		fprintf(stderr," FAILED!\n\n");
		return -1;
	}
*/	while(1){
		if (command != NULL){
			free(command);
			command = NULL;
		}
		if( ( command = readline ("_ncc_>") ) == NULL ){
			if(stop_signal != 1) kill(ch_pid,NCE_STOP_SIMULATION);
			if(console != NULL){
				fprintf(stderr,"_ncc_:NeuroCAD in \'script mode\'. Wating the end of simulation!\n");
				while(stop_signal == 0)sleep(1);
			}
			else fprintf(stderr,"\n_ncc_:QUIT from ncc?\n");
			*(hc.command) = NCE_SYSTEM_QUIT;
			*(hc.length) = 0;
			write(hOut, cmd, COMMAND_CONTEXT_SIZE);
			while(engine_ret != 2)sleep(1);
			fprintf(stderr,"_ncc_>yes\n_ncc_:Are You sure?\n_ncc_>yes\n_ncc_:Are You really sure?\n_ncc_>Yes, YeS, YES .......!\n_ncc_:Good luck!\n");
			return 0;
		}
		if (*command)add_history (command);

		engine_ret = 0;
		if( command[0] == '#') continue;
		else if(strstr(command,"help")) help();
		else if(strstr(command, "quit")){
			if(stop_signal != 1) kill(ch_pid,NCE_STOP_SIMULATION);
			if(console != NULL){
				fprintf(stderr,"_ncc_:NeuroCAD in \'script mode\'. Wating the end of simulation!\n");
				while(stop_signal == 0)sleep(1);
			}
			else fprintf(stderr,"_ncc_:QUIT from ncc?\n");
			*(hc.command) = NCE_SYSTEM_QUIT;
			*(hc.length) = 0;
			write(hOut, cmd, COMMAND_CONTEXT_SIZE);
			while(engine_ret != 2)sleep(1);
			fprintf(stderr,"_ncc_:Are You sure?\n_ncc_>yes\n_ncc_:Are You really sure?\n_ncc_>Yes, YeS, YES .......!\n_ncc_:Good luck!\n");
			return 0;
		}
		else if(strstr(command, "pid")){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_SYSTEM_GET_PID;
				*(hc.length) = 0;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
/*		else if( (command_shift = strstr(command, "msg ")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				sscanf(command_shift + strlen("msg "),"%s",filename);
				if(filename[0] == '$'){
					*(hc.command) = NCE_OPEN_MESSAGE_LOCAL_SOCKET;
					*(hc.length) = strlen(filename)-1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, filename + 1, strlen(filename) - 1);
				}
				else{
					*(hc.command) = NCE_OPEN_MESSAGE_FILE;
					*(hc.length) = strlen(filename);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, filename, strlen(filename));
				}
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if( (command_shift = strstr(command, "dbmsg ")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				sscanf(command_shift + strlen("dbmsg "),"%s",filename);
				if(filename[0] == '$'){
					*(hc.command) = NCE_OPEN_DB_LOCAL_SOCKET;
					*(hc.length) = strlen(filename)-1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, filename + 1, strlen(filename) - 1);
				}
				else{
					*(hc.command) = NCE_OPEN_DB_FILE;
					*(hc.length) = strlen(filename);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, filename, strlen(filename));
				}
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
*/		else if( (command_shift = strstr(command, "upload ")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				sscanf(command_shift + strlen("upload "),"%s",filename);
				int fd = open(filename,O_RDONLY);
				if( fd < 0 ){
					fprintf(stderr,"_ncc_:Cannot open file %s\n", filename);
				}
				else{
					size_t size = lseek(fd, 0, SEEK_END);
					lseek(fd, 0, SEEK_SET);
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = (int) size + 1;			write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_MODEL_UPLOAD;
					write(hOut, cmd, 1);
					if(nce_rdwr(hOut, fd, size) != size)
						fprintf(stderr,"_ncc_:Cannot send file to engine. Pipe unstable!\n");
					else{
						while(engine_ret == 0)sleep(1);
						if(engine_ret == 2) new_engine();
					}
					close( fd );
				}
			}
		}
		else if(strstr(command, "reinit")){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_ASSEMBLER_COMMAND;
				*(hc.length) =  1;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				*(hc.command) = ASMBASIC_MODEL_REINIT;
				write(hOut, cmd, 1);
				
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if(strstr(command, "release")){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_ASSEMBLER_COMMAND;
				*(hc.length) =  1;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				*(hc.command) = ASMBASIC_MODEL_RELEASE;
				write(hOut, cmd, 1);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if(strstr(command, "destroy")){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_ASSEMBLER_COMMAND;
				*(hc.length) =  1;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				*(hc.command) = ASMBASIC_MODEL_DESTROY;
				write(hOut, cmd, 1);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
				
		else if(strstr(command, "save")){
		/*
			*(hc.command) = ASMBASIC_MODEL_SAVE;
			*(hc.length) = 0;
			write(hOut, cmd, COMMAND_CONTEXT_SIZE);
			while(engine_ret == 0)sleep(1);
		*/
			fprintf(stderr,"_ncc_:\'save\' currently unsupported by front-end!\n");
		}
		else if(strstr(command, "restore")){
		/*
			*(hc.command) = ASMBASIC_MODEL_SAVE;
			*(hc.length) = 0;
			write(hOut, cmd, COMMAND_CONTEXT_SIZE);
			while(engine_ret == 0)sleep(1);
		*/
			fprintf(stderr,"_ncc_:\'restore\' currently unsupported by front-end!\n");
		}
		else if(strstr(command, "step")){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_SIMUALTION_STEP;
				*(hc.length) = 0;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if((command_shift = strstr(command, "start")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"start %lf",&d_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'start\'command! Type help\n");
				else{
					stop_signal = 0;
					*(hc.command) = NCE_SIMULATION_START;
					*(hc.length) = sizeof(double);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &d_value, sizeof(double));
					if(console != NULL){
						fprintf(stderr,"_ncc_:NeuroCAD in \'script mode\'. Wating the end of simulation!\n");
						while(stop_signal == 0)sleep(1);
						if(engine_ret == 2) new_engine();
					}
				}
			}
		}
		else if((command_shift = strstr(command, "go")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"go %d",&i_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'go\'command! Type help\n");
				else{
					stop_signal = 0;
					*(hc.command) = NCE_SIMULATION_GO;
					*(hc.length) = sizeof(int);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &i_value, sizeof(int));
					if(console != NULL){
						fprintf(stderr,"_ncc_:NeuroCAD in \'script mode\'. Wating the end of simulation!\n");
						while(stop_signal == 0)sleep(1);
						if(engine_ret == 2) new_engine();
					}
				}
			}
		}
		else if((command_shift = strstr(command, "until")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"until %lf",&d_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'until\'command! Type help\n");
				else{
					stop_signal = 0;
					*(hc.command) = NCE_SIMULATION_UNTIL;
					*(hc.length) = sizeof(double);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &d_value, sizeof(double));
					if(console != NULL){
						fprintf(stderr,"_ncc_:NeuroCAD in \'script mode\'. Wating the end of simulation!\n");
						while(stop_signal == 0)sleep(1);
						if(engine_ret == 2) new_engine();
					}
				}
			}
		}
		else if((command_shift = strstr(command, "stop")) != NULL) kill(ch_pid,NCE_STOP_SIMULATION);
		else if((command_shift = strstr(command, "getsimtime")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_SIMULATION_GET_MODEL_TIME;
				*(hc.length) = 0;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if((command_shift = strstr(command, "setsimtime")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"setsimtime %lf",&d_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'setsimtime\'command! Type help\n");
				else{
					*(hc.command) = NCE_SIMULATION_SET_MODEL_TIME;
					*(hc.length) = sizeof(double);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &d_value, sizeof(double));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "getstepsim")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_SIMULATION_GET_MODEL_TIMESTEP;
				*(hc.length) = 0;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if((command_shift = strstr(command, "setstepsim")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"setstepsim %lf",&d_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'setstepsim\'command! Type help\n");
				else{
					*(hc.command) = NCE_SIMULATION_SET_MODEL_TIMESTEP;
					*(hc.length) = sizeof (double);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &d_value, sizeof(double));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "get")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"get %d %d",&i_value, &ii_value) != 2)
					fprintf(stderr,"_ncc_:Wrong format \'get\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof(int) * 2 + 1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_DATA_GET;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "set")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"set %d %d %lf",&i_value, &ii_value, &d_value) != 3)
					fprintf(stderr,"_ncc_:Wrong format \'set\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof(double) + sizeof(int) * 2 +1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_DATA_SET;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					write(hOut, &d_value, sizeof(double));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "watcherfile")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"watcherfile %lf %s", &d_value, filename) != 2)
					fprintf(stderr,"_ncc_:Wrong format \'watcherfile\'command! Type help\n");
				else{
					if(filename[0] == '$'){
						*(hc.command) = NCE_WATCHER_ADD_LOCAL_SOCKET;
						*(hc.length) = sizeof(double)+strlen(filename) - 1;
						write(hOut, cmd, COMMAND_CONTEXT_SIZE);
						write(hOut, &d_value, sizeof(double));
						write(hOut, filename + 1, strlen(filename) - 1 );
						while(engine_ret == 0)sleep(1);
						if(engine_ret == 2) new_engine();
					}
					else{
						*(hc.command) = NCE_WATCHER_ADD_FILE;
						*(hc.length) = sizeof(double)+strlen(filename);
						write(hOut, cmd, COMMAND_CONTEXT_SIZE);
						write(hOut, &d_value, sizeof(double));
						write(hOut, filename, strlen(filename));
						DB_INFO("ncc", "Wating engine_ret");
						while(engine_ret == 0)sleep(1);
						DB_INFO("ncc", "engine_ret is DONE");
						if(engine_ret == 2) new_engine();
					}
				}
			}
		}
		else if((command_shift = strstr(command, "watcherkillall")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				*(hc.command) = NCE_WATCHER_KILL_ALL;
				*(hc.length) = 0;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if((command_shift = strstr(command, "watcherkill")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"watcherkill %d", &i_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'watcherkill\'command! Type help\n");
				else{
					*(hc.command) = NCE_WATCHER_KILL;
					*(hc.length) = sizeof(int);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &i_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "watcher")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"watcher %lf", &d_value) != 1)
					fprintf(stderr,"_ncc_:Wrong format \'watcher\'command! Type help\n");
				else{
					*(hc.command) = NCE_WATCHER_ADD;
					*(hc.length) = sizeof(double);
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					write(hOut, &d_value, sizeof(double));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "itemadd")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"itemadd %d %d %d", &i_value, &ii_value, &iii_value) != 3)
					fprintf(stderr,"_ncc_:Wrong format \'itemadd\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof( int ) * 3 + 1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_WATCHER_ADD_ITEM;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					write(hOut, &iii_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "itemminimal")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"itemminimal %d %d %d", &i_value, &ii_value, &iii_value) != 3)
					fprintf(stderr,"_ncc_:Wrong format \'itemminimal\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof( int ) * 3 + 1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_WATCHER_ADD_ITEM_MIN;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					write(hOut, &iii_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "itemmaximal")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"itemmaximal %d %d %d", &i_value, &ii_value, &iii_value) != 3)
					fprintf(stderr,"_ncc_:Wrong format \'itemmaximal\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof( int ) * 3 + 1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_WATCHER_ADD_ITEM_MAX;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					write(hOut, &iii_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}

		else if((command_shift = strstr(command, "itemaverage")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"itemaverage %d %d %d", &i_value, &ii_value, &iii_value) != 3)
					fprintf(stderr,"_ncc_:Wrong format \'itemaverage\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof( int ) * 3 + 1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_WATCHER_ADD_ITEM_AVERAGE;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					write(hOut, &iii_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if((command_shift = strstr(command, "itemkill")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				if( sscanf(command_shift ,"itemkill %d %d %d", &i_value, &ii_value, &iii_value) != 3)
					fprintf(stderr,"_ncc_:Wrong format \'itemkill\'command! Type help\n");
				else{
					*(hc.command) = NCE_ASSEMBLER_COMMAND;
					*(hc.length) = sizeof( int ) * 3 + 1;
					write(hOut, cmd, COMMAND_CONTEXT_SIZE);
					*(hc.command) = ASMBASIC_WATCHER_KILL_ITEM;
					write(hOut, cmd, 1);
					write(hOut, &i_value, sizeof(int));
					write(hOut, &ii_value, sizeof(int));
					write(hOut, &iii_value, sizeof(int));
					while(engine_ret == 0)sleep(1);
					if(engine_ret == 2) new_engine();
				}
			}
		}
		else if( (command_shift = strstr(command, "info")) != NULL){
			if(stop_signal != 1)
				fprintf(stderr,"_ncc_:Simulation is already started!\n");
			else{ 
				memset(filename, 0, 1024);
				sscanf(command_shift + strlen("info "),"%s",filename);
				if( !strcmp(filename,"asm") )
					*(hc.command) = NCE_PRINT_INFO_ASM;
				else if (!strcmp(filename,"mod") )
					*(hc.command) = NCE_PRINT_INFO_MODULES;
/*				else if (!strcmp(filename,"obj") )
					*(hc.command) = ASMBASIC_PRINT_INFO_OBJECTS;
				else if (!strcmp(filename,"pars") )
					*(hc.command) = ASMBASIC_PRINT_INFO_PARSER;
*/				else if (!strcmp(filename,"sim") )
					*(hc.command) = NCE_PRINT_INFO_SIMULATOR;
				else if (!strcmp(filename,"mem") )
					*(hc.command) = NCE_PRINT_INFO_MEMORY;
				else if (!strcmp(filename,"watch") )
					*(hc.command) = NCE_PRINT_INFO_WATCHER;
				else
					*(hc.command) = NCE_PRINT_INFO;
				*(hc.length) = 0;
				write(hOut, cmd, COMMAND_CONTEXT_SIZE);
				while(engine_ret == 0)sleep(1);
				if(engine_ret == 2) new_engine();
			}
		}
		else if(strlen(command) == 0){
			if(console == NULL)fprintf(stderr,"\r");
		}
		else
			fprintf(stderr,"_ncc_:UNKNOWN COMMAND. Type help\n");
	}

	return 0;
}

int main (int cv, char **vv)
{
	int exitcode = 0;
	// scan a command line
	char 	**sv = NULL, 
		*iscriptname = NULL, 
		*watchername = NULL;
		
	for( sv = vv; *sv != NULL; sv++ ){
		__FIND_PS__( *sv, "isf=", iscriptname );
		__FIND_PS__( *sv, "wbf=", watchername );
		if( (!strcmp(*sv, "-h")) || (!strcmp(*sv, "--help")) )
			return usage(vv[0]);
		
	}
	
	int i_pipe[2], o_pipe[2];
	// create a context of neurocad engine 
	memset( &nce_ctx, 0, sizeof( nce_context ));
	// get a system statistics
	struct sysinfo si;
	memset( &si, 0, sizeof( sysinfo ));
	if( sysinfo( &si ) < 0 ){
		fprintf( stderr, "Cannot receive the system statistics. \'sysinfo()\' returns an error.\n" );
		return -2;
	}
	// create i/o pipes
	if( ( pipe( i_pipe ) < 0 ) || ( pipe( o_pipe ) < 0 ) ){
		fprintf( stderr, "Cannon open i/o pipes. \'pipe()\' returns an error\n" );
		return -3;
	}
		
	// create watcher file
	nce_ctx.hWatch = ( watchername ) ? 
		open( watchername, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR ) : -1 ;
	nce_ctx.memsize_all = nce_ctx.memsize_free = (( size_t )si.totalram )*(( size_t )si.mem_unit );
	nce_ctx.hOutput = o_pipe[1];


	nce_ctx.hInput = i_pipe[0];
	main_pid = getpid();
	if( (ch_pid = fork()) == 0 ){
		asm_drop();
		asm_model_grow_pointer 	= &asm_model_grow;
		asm_messenger_pointer	= &asm_messenger;
		asm_printinfo_pointer	= &asm_printinfo;
		nce( &nce_ctx );
		return 0;
	}
	if( (decoder_pid = fork()) == 0 ){
		binary_decoder(o_pipe[0]);
		return 0;
	}	

	struct sigaction nce_ret_signal, nce_sim_stop_signal, nce_deaded_signal;
	memset(&nce_ret_signal, 0, sizeof(nce_ret_signal));
	nce_ret_signal.sa_handler = &nce_ret;
	sigaction(SIGUSR1, &nce_ret_signal, NULL);
	
	memset(&nce_sim_stop_signal, 0, sizeof(nce_sim_stop_signal));
	nce_sim_stop_signal.sa_handler = &nce_sim_stop;
	sigaction(SIGUSR2, &nce_sim_stop_signal, NULL);
	
	memset(&nce_deaded_signal, 0, sizeof(nce_deaded_signal));
	nce_deaded_signal.sa_handler = &nce_deaded;
	sigaction(SIGALRM, &nce_deaded_signal, NULL);
	
	stop_signal = 1;

	
	if( iscriptname ){
		FILE *script = fopen(iscriptname, "rt");
		exitcode = terminal( vv[0], script, i_pipe[1] );
		fclose(script);
	}
	else{
		exitcode = terminal( vv[0], NULL,  i_pipe[1] );
	}
	close( nce_ctx.hInput );
	close( nce_ctx.hOutput );
	close( nce_ctx.hWatch );
	return exitcode;
}


