#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <neurocad/engine/simmanager.h>
#include <neurocad/engine/nce_common.h>
#include <neurocad/assemblers/basic/asmbasic.h>
#include <neurocad/assemblers/basic/objmanager.h>
#include <neurocad/assemblers/basic/xmlfn.h>

///////////////////////////////////////////////////////////////////////////////
//	local variables
static int 		fd = -1;			// handle of open model
static unsigned long 	allnames     = 0;		//
static size_t 		allblocksize = 0;		//

///////////////////////////////////////////////////////////////////////////////
//	local functions
//-----------------------------------------------------------------------------
static size_t rdwr(int hto, int hin, size_t size)
{
	char buffer[__RDWR_BUFFER_SIZE__];
	size_t all_sr = 0, curr_sr = 0, cnt;
	int wr_ret, rd_ret;
	while(all_sr != size){
		curr_sr = ( (size - all_sr) < __RDWR_BUFFER_SIZE__ ) ?  (size - all_sr) : __RDWR_BUFFER_SIZE__;
		cnt = 0;
		while( cnt != curr_sr){
			rd_ret = read(hin, buffer + cnt, curr_sr - cnt);
			if(rd_ret < 0)	return all_sr;
			cnt += rd_ret;
		}
		cnt = 0;
		while( cnt != curr_sr){
			wr_ret = write(hto, buffer + cnt, curr_sr - cnt);
			if(wr_ret < 0)	return all_sr;
			cnt += wr_ret;
		}
		all_sr += curr_sr;
	}
	return all_sr;
}

static int test_version(){
	/*It has to be model*/	
	char errmessage[201]; memset(errmessage,0,201);
	char *xmlText = xmlfn_nextnode("model");
	if(xmlText == NULL)
		return ERROR_MESSAGE("test_version","The tag <model> is not specified.", NULL,-1);
	/*app attribute of model must be "neurocad"*/
	if( (xmlText = xmlfn_getprop( "app" )) == NULL)
		return ERROR_MESSAGE("test_version","The property \'app\' in the tag <model> is not specified.", NULL,-2);
	if(strcmp(xmlText, __APPLICATION__)){
		snprintf(errmessage, 200, "The property app=\"%s\" in the tag <module app=\"%s\"> must be %s.", xmlText, xmlText, __APPLICATION__);
		return ERROR_MESSAGE("test_version",errmessage, NULL,-3);
	}
	/*appversion attribute of model must be 0.03*/
	if( (xmlText = xmlfn_getprop( "app_version" )) == NULL){
		snprintf(errmessage, 200, "The property \'app_version\' in the tag <module app=\"%s\" > is not specified.", __APPLICATION__);
		return ERROR_MESSAGE("test_version",errmessage, NULL,-4);
		}
	if(strcmp(xmlText,  __APP_VERSION__)){
		snprintf(errmessage, 200, "The property appversion=\"%s\" in the tag <module app=\"%s\" appversion=\"%s\"> must be %s.", xmlText, __APPLICATION__, xmlText, __APP_VERSION__ );
		return ERROR_MESSAGE("test_version",errmessage,NULL,-5);
	}
	/*version attribute of model must be 0.1*/
	if( (xmlText = xmlfn_getprop( "model_format_version" )) == NULL){
		snprintf(errmessage, 200, "The property \'model_format_version\' in the tag <module app=\"%s\" appversion=\"%s\"> is not specified.", __APPLICATION__, __APP_VERSION__ );
		return ERROR_MESSAGE("test_version",errmessage, NULL,-6);
		}
	if(strcmp(xmlText, __MOD_FORMAT_VERSION__)){
		snprintf(errmessage, 200, "The property model_format_version=\"%s\" in the tag <module app=\"%s\" appversion=\"%s\" model_format_version=\"%s\"> must be %s.", xmlText, __APPLICATION__,__APP_VERSION__, xmlText, __MOD_FORMAT_VERSION__ );
		return ERROR_MESSAGE("TestVersion",errmessage, NULL,-7);
	}
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	compilation_module		!!!done
static int compilation_module()
{
	int mid = -1;
	char *xmlText = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	if( (xmlText = xmlfn_getprop( "id" )) == NULL)
		return ERROR_MESSAGE("basic parser:compilation_module","The property \'id\' in the tag <module> is not specified.", NULL, -1);
	if( sscanf(xmlText,"%d",&mid) != 1){
		snprintf(errmessage, 200, "The property id=\"%s\" in the tag <module id=\"%s\"> is corrupted.", xmlText, xmlText);
		return ERROR_MESSAGE("basic parser:compilation_module",errmessage, NULL, -2);
	}
	if( mid < 0 ){
		snprintf(errmessage, 200, "The property id=\"%d\" in the tag <module id=\"%d\"> is negative.", mid, mid);
		return ERROR_MESSAGE("basic parser:compilation_module",errmessage, NULL, -3);
	}
	if( ((xmlText = xmlfn_getcontent()) == NULL) || (strlen(xmlText) < 3) ){
		snprintf(errmessage, 200, "Cannot read a content of the tag <module id=\"%d\">.", mid);
		return ERROR_MESSAGE("basic parser:compilation_module", errmessage, NULL, -4);
	}
	int retmod = -1;
	if( (retmod = mod_insert(xmlText,mid)) < 0)
		return ERROR_MESSAGE("basic parser:compilation_module","\'mod_insert()\' returns an error", NULL, -5);
	OUTPUT_MESSAGE("nce:[basic parser]:LOADING     MODULE:\t\"%s\"[%03d:%03d] \t\t\t\tOK\n", xmlText, mid, retmod);
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	compilation_object_block		!!!done
static int compilation_object_block()
{
	int nid = -1, mid = -1;
	char *xmlText = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	if( (xmlText = xmlfn_getprop( "mid" )) == NULL )
		return ERROR_MESSAGE("basic parser:compilation_object_block","The property \'mid\' in the tag <block> is not specified.", NULL, -1);
	if( sscanf(xmlText,"%d",&mid) != 1){
		snprintf(errmessage, 200, "The property mid=\"%s\" in the tag <block mid=\"%s\"> is corrupted.", xmlText, xmlText);
		return ERROR_MESSAGE("basic parser:compilation_object_block",errmessage, NULL, -2);
	}
	if( mod_get_item(mid) == NULL){
		snprintf(errmessage, 200, "The property mid=\"%d\" in the tag <block mid=\"%d\"> cannot be loaded. The model is damage", mid, mid);
		return ERROR_MESSAGE("basic parser:compilation_object_block",errmessage, NULL, -3);
	}
	if( mod_init(mid) < 0)
		return ERROR_MESSAGE("basic parser:compilation_object_block","\'mod_init()\' returns an error", NULL, -4);
	xmlfn_stepin();
	for(xmlText = xmlfn_nextnode( NULL ); xmlText != NULL; xmlText = xmlfn_nextnode( NULL )){
		if(!strcmp(xmlText,"name")){
			if( (xmlText = xmlfn_getprop( "id" )) == NULL )
				return ERROR_MESSAGE("basic parser:compilation_object_block","The property \'id\' in the tag <name> is not specified.", NULL, -5);
			if( strstr(xmlText, "++") != NULL ){
				obj_curr->soft_names++;
				continue;
			}
			else if( sscanf(xmlText,"%d",&nid) != 1){
				snprintf(errmessage, 200, "The property id=\"%s\" in the tag <name id=\"%s\"> is corrupted.", xmlText, xmlText);
				return ERROR_MESSAGE("basic parser:compilation_object_block",errmessage, NULL, -6);
			}
			if( nid < 0 ){
				snprintf(errmessage, 200, "The property id=\"%d\" in the tag <name id=\"%d\"> is negative.", nid, nid);
				return ERROR_MESSAGE("basic parser:compilation_object_block",errmessage, NULL, -7);
			}
			if( (nid + 1) > obj_curr->hard_names) obj_curr->hard_names = nid + 1;
		}
	}
	obj_curr->blocksize += mod_get_item(mid)->blocksize;
	xmlfn_stepout();
	return 0;
}

		
///////////////////////////////////////////////////////////////////////////////
//	compilation_object		!!!done
static int compilation_object()
{
	long oid, objcnt = 0; int blkcnt = 0;
	char *xmlText = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	OUTPUT_MESSAGE("nce:[basic parser]:COMPILING   OBJECT:\t");
	if( (xmlText = xmlfn_getprop( "id" )) == NULL)
		return ERROR_MESSAGE("basic parser:compilation_object","The property \'id\' in the tag <object> is not specified.", NULL, -1);
	
	if( sscanf(xmlText,"%ld",&oid) != 1){
		snprintf(errmessage, 200, "The property id=\"%s\" in the tag <object id=\"%s\"> is corrupted.", xmlText, xmlText);
		return ERROR_MESSAGE("basic parser:compilation_object",errmessage, NULL, -2);
	}
	if( (objcnt = obj_add(oid)) < 0)
		return ERROR_MESSAGE("basic parser:compilation_object","\'obj_add()\' returns an error", NULL, -4);
	if( (obj_setactive(oid)) < 0)
		return ERROR_MESSAGE("basic parser:compilation_object","\'obj_setactive()\' returns an error.", NULL, -5);
	obj_curr->datasize = 0;
	obj_curr->soft_names =  0;
	obj_curr->hard_names =  0;
	OUTPUT_MESSAGE("[%07ld:%07ld] ",oid,objcnt);
	xmlfn_stepin();
	for(xmlText = xmlfn_nextnode( NULL ); xmlText != NULL; xmlText = xmlfn_nextnode( NULL )){
		if(!strcmp(xmlText,"block")){
			if( compilation_object_block() < 0){
				snprintf(errmessage,200,"\'compilation_object_block()\' returns an error for the tag <object id=\"%ld\">.",oid);
				return ERROR_MESSAGE("basic parser:compilation_object",errmessage, NULL, -6);
				}
			blkcnt++;
		}
		OUTPUT_MESSAGE(".");
	}
	xmlfn_stepout();
	OUTPUT_MESSAGE("\tnames=%d; blocks=%d \t\t\tOK\n",obj_curr->datasize, blkcnt);
	obj_curr->datasize = obj_curr->hard_names + obj_curr->soft_names;
	allnames += obj_curr->datasize;
	allblocksize += obj_curr->blocksize;
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	compilation_simulation		!!!done
static int compilation_simulation()
{
	char *xmlText = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	OUTPUT_MESSAGE("nce:[basic parser]:SET     PARAMETERS:\t");
	if( (xmlText = xmlfn_getprop("step")) == NULL){
		WARNING_MESSAGE("compilation_simulation","The property \'step\' in the tag <simulation> is not specified. The \'step\' is initialized by default valde 0.01.", NULL, -1);
		time_step = 0.01;
	}
	else if( sscanf(xmlText,"%lf", &time_step) != 1){
		snprintf(errmessage, 200, "The property step=\"%s\" in the tag <simulation step=\"%s\"> is corrupted. The \'step\' is initialized by default valde 0.01.", xmlText, xmlText);
		WARNING_MESSAGE("compilation_simulation",errmessage, NULL, -2);
		time_step = 0.01;
	}
	if(time_step < 0.0){
		snprintf(errmessage, 200, "The property step=\"%lf\" in the tag <simulation step=\"%lf\"> is negative. The \'step\' is initialized by default valde 0.01.", time_step, time_step);
		WARNING_MESSAGE("compilation_simulation",errmessage, NULL, -3);
		time_step = 0.01;
	}
	if( (xmlText = xmlfn_getprop("bufdeep")) == NULL){
		snprintf(errmessage, 200, "The property \'bufdeep\' in the tag <simulation step=\"%lf\"> is not specified. The \'bufdeep\' is initialized by default valde 3.", time_step);
		WARNING_MESSAGE("compilation_simulation",errmessage,NULL, -4);
		_bufdeep = 3;
	}
	else if( sscanf(xmlText,"%d", &_bufdeep) != 1){
		snprintf(errmessage, 200, "The property bufdeep=\"%s\" in the tag <simulation step=\"%lf\" bufdeep=\"%s\"> is corrupted. The \'bufdeep\' is initialized by default valde 3.", xmlText, time_step, xmlText);
		WARNING_MESSAGE("compilation_simulation",errmessage, NULL, -5);
		_bufdeep = 3;
	}
	if(_bufdeep < 0){
		snprintf(errmessage, 200, "The property bufdeep=\"%d\" in the tag <simulation step=\"%lf\" bufdeep=\"%d\"> is corrupted. The \'bufdeep\' is initialized by default valde 3.", _bufdeep, time_step, _bufdeep);
		WARNING_MESSAGE("compilation_simulation",errmessage, NULL, -6);
		_bufdeep = 3;
	}
	OUTPUT_MESSAGE("step = %lg; bufdeep = %d \t\t\t\t\tOK\n",time_step, _bufdeep);
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	compilation		!!!done
static int compilation()
{
	char *xmlText = NULL;
	xmlfn_stepin();
	while( (xmlText = xmlfn_nextnode( NULL )) != NULL){
		if(!strcmp(xmlText,"module")){
			if( compilation_module() < 0 )
				return ERROR_MESSAGE("basic parser:compilation","\'compilation_module()\' returns an error.", NULL, -1);
			else
				continue;
		}
		if(!strcmp(xmlText,"object")){
			if( compilation_object() < 0 )
				return ERROR_MESSAGE("basic parser:compilation","\'compilation_object()\' returns an error.", NULL, -2);
			else
				continue;
		}
		if(!strcmp(xmlText,"simulation")){
			if( compilation_simulation() < 0 )
				return ERROR_MESSAGE("basic parser:compilation","\'compilation_simulation()\' returns an error.", NULL, -5);
			else
				continue;
		}
	}
	xmlfn_stepout();
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	assembling_module		!!!done
static int assembling_module()
{
	int mid = -1, active = 0;
	char *xmlText = NULL;
	char *xmlAttr = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	if( (xmlText = xmlfn_getprop( "id" )) == NULL){
		return ERROR_MESSAGE("basic parser:assembling_module","The property \'id\' in the tag <module> is not specified.", NULL, -1);
	}
	if( sscanf(xmlText,"%d",&mid) != 1){
		snprintf(errmessage, 200, "The property id=\"%s\" in the tag <module id=\"%s\"> is corrupted.",xmlText,xmlText);
		return ERROR_MESSAGE("basic parser:assembling_module",errmessage, NULL, -2);
	}
	if( mid < 0 ){
		snprintf(errmessage,200,"The \'id\' in the tag <module id=\"%d\"> is negative.", mid );
		return ERROR_MESSAGE("basic parser:assembling_module",errmessage, NULL, -3);
	}
	snprintf(errmessage, 200, "The module %s is not active.",xmlText);
	if( !mod_is_acctive(mid) ) 
		return WARNING_MESSAGE("assembling_module",errmessage,NULL,0);
	while( (xmlText = xmlfn_propscan()) != NULL ){
		if( !strcmp(xmlText, "id") ) continue;
		if( (xmlAttr = bufdup( xmlAttr, xmlText)) == NULL)
			return  ERROR_MESSAGE("basic parser:assembling_module","Cannot duplicate buffer by \'bufdup()\'.", NULL, -5);
		if( mod_set_extended_attribute(mid, xmlAttr, xmlText = xmlfn_propscan_getvalue()) < 0)
			return ERROR_MESSAGE("basic parser:assembling_module","\'mod_set_extended_attribute()\' returns an error", NULL, -6);
		active = 1;
		OUTPUT_MESSAGE("nce:[basic parser]:SETTING  ATTRIBUTE:\tmodule %03d %s = %s; ", mid, xmlAttr, xmlText);
	}
	if(active){
		if(mod_flash_extended_attribute(mid) < 0)
			return ERROR_MESSAGE("basic parser:assembling_module","\'mod_flash_extended_attribute()\' returns an error.", NULL, -7);
		else
			OUTPUT_MESSAGE(">FLASHING< \t\t\tOK\n");
	}
	return 0;
}


///////////////////////////////////////////////////////////////////////////////
//	assembling_object_block_get_name_shift		!!! done
static int assembling_object_block_get_name_shift(mip mod, char *name)
{
	if( mod == NULL)
		return ERROR_MESSAGE("basic parser:assembling_object_block_get_name_shift","module pointer is NULL.",NULL,-1);
	if( name == NULL)
		return ERROR_MESSAGE("basic parser:assembling_object_block_get_name_shift","name pointer is NULL.",NULL,-2);
	int namecnt = 0;
	for(; namecnt < mod->names; namecnt++)
		if(!strcmp(mod->parvar[namecnt],name)) 
			return namecnt;
	char error_msg[200]; memset(error_msg, 0, 200);
	snprintf(error_msg, 199, "Cannot find the name in module %d",mod->mid);
	return ERROR_MESSAGE("basic parser:assembling_object_block_get_name_shift",error_msg,name,-3);
}

///////////////////////////////////////////////////////////////////////////////
//	assembling_object_block		!!! done
static int assembling_object_block()
{
	int mid = -1, nid = -1, nameshift = -1, buffcount;
	long link_oid; int link_nid = -1;
	obp link_obj =  NULL;
	mip mod = NULL;
	double init_val;
	char *xmlText = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	if( (xmlText = xmlfn_getprop( "mid" )) == NULL )
		return ERROR_MESSAGE("basic parser:assembling_object_block","The property \'mid\' in the tag <block> is not specified.", NULL, -1);
	if( sscanf(xmlText,"%d",&mid) != 1){
		snprintf(errmessage, 200, "The property mid=\"%s\" in the tag <block mid=\"%s\"> is corrupted.", xmlText, xmlText);
		return ERROR_MESSAGE("basic parser:compilation_object_block",errmessage, NULL, -2);
	}
	if( (mod = mod_get_item(mid)) == NULL){
		snprintf(errmessage, 200, "The module %d was not loaded; the model is corrupted.", mid);
		return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -3);
	}
	if( (obj_curr->blocks = (*mod->modadd)(mid, obj_curr->blocks, memory_heandle)) == NULL){
		snprintf(errmessage, 200, "\'modadd()\' for the module %d (filename %s) returns an error.",mod->mid, mod->filename);
		return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -4);
	}
	obj_curr->blocks->oid = obj_curr->oid;
	obj_curr->blocks->bid = (obj_curr->blocks->next == NULL) ? 0 : obj_curr->blocks->next->bid + 1;
	xmlfn_stepin();
	for(xmlText = xmlfn_nextnode( NULL ); xmlText != NULL; xmlText = xmlfn_nextnode( NULL )){
		if(!strcmp(xmlText,"name")){
			if( (xmlText = xmlfn_getprop( "id" )) == NULL )
				return ERROR_MESSAGE("basic parser:assembling_object_block","The property \'id\' in the tag <name> is not specified.", NULL, -5);
			if( strstr(xmlText, "++") != NULL ){
				nid = obj_curr->hard_names + obj_curr->soft_names;
				obj_curr->soft_names ++;
			}
			else if( sscanf(xmlText,"%d",&nid) != 1){
				snprintf(errmessage, 200, "The property id=\"%s\" in the tag <name id=\"%s\"> is corrupted.",xmlText,xmlText);
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -6);
			}
			if( nid < 0 ){
				snprintf(errmessage,200,"The \'id\' in the tag <name id=\"%d\"> is negative.", nid );
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -7);
			}
			if( nid >= obj_curr->datasize ){
				snprintf(errmessage,200,"The \'id\' in the tag <name id=\"%d\"> is more than object size", nid );
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -8);
			}
			if( (nameshift = assembling_object_block_get_name_shift(mod, xmlText = xmlfn_getcontent())) < 0)
				return ERROR_MESSAGE("basic parser:assembling_object_block", "\'assembling_object_block_get_name_shift()\' returns an error.", NULL, -9);
			obj_curr->blocks->parvar[nameshift].dp = &obj_curr->data[nid * _bufdeep];
			OUTPUT_MESSAGE("*");
			if( (xmlText = xmlfn_getprop( "init" )) == NULL ) continue;
			if( sscanf(xmlText,"%lf",&init_val) != 1){
				snprintf(errmessage,200,"The property init=\"%s\" in the tag <name id=\"%d\" init=\"%s\"> is corrupted.", xmlText, nid, xmlText);
				WARNING_MESSAGE("assembling_object_block",errmessage, NULL, 0);
				continue;
			}
			for(buffcount = 0 ; buffcount < _bufdeep; buffcount++)
				obj_curr->data[nid * _bufdeep + buffcount] = init_val;
		}
		
		if(!strcmp(xmlText,"link")){
			if( (xmlText = xmlfn_getprop( "oid" )) == NULL ){
				snprintf(errmessage,200,"The property \'oid\' in the tag <link> tag is not specified.");
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -10);
			}
			if( sscanf(xmlText,"%ld",&link_oid) != 1){
				snprintf(errmessage,200,"The property \'oid\' in the tag <link oid=\"%s\"> is corrupted.", xmlText);
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -11);
			}
			if( (link_obj = obj_find(link_oid)) == NULL){
				snprintf(errmessage, 200, "Cannot find object %ld for linking.",link_oid);
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -12);
			}
			if( (xmlText = xmlfn_getprop( "nid" )) == NULL ){
				snprintf(errmessage,200,"The property \'nid\' in the tag <link oid=\"%ld\"> tag is not specified.", link_oid);
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -13);
			}
			if( sscanf(xmlText,"%d",&link_nid) != 1){
				snprintf(errmessage,200,"The property nid=\"%s\" in the tag <link oid=\"%ld\" nid=\"%s\"> is corrupted.", xmlText, link_oid, xmlText);
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -14);
			}
			if( link_nid < 0){
				snprintf(errmessage,200,"The \'nid\' in the tag <link oid=\"%ld\" nid=\"%d\"> is negative.", link_oid, link_nid );
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -15);
			}
			if( link_nid >= link_obj->datasize){
				snprintf(errmessage,200,"The \'nid\' in the tag <link oid=\"%ld\" nid=\"%d\"> is more than total amount of names in the object %ld.", link_oid, link_nid, link_oid );
				return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -16);
			}
			if( (nameshift = assembling_object_block_get_name_shift(mod, xmlText = xmlfn_getcontent())) < 0)
				return ERROR_MESSAGE("basic parser:assembling_object_block","\'assembling_object_block_get_name_shift()\' returns an error.", NULL, -9);
			obj_curr->blocks->parvar[nameshift].dp = &link_obj->data[link_nid * _bufdeep];
			OUTPUT_MESSAGE("-");
			if( (xmlText = xmlfn_getprop( "init" )) == NULL ) continue;
			if( sscanf(xmlText,"%lf",&init_val) != 1){
				snprintf(errmessage,200,"The property init=\"%s\" in the tag <link oid=\"%ld\" nid=\"%d\" init=\"%s\"> is corrupted.", xmlText, link_oid, link_nid,  xmlText);
				WARNING_MESSAGE("assembling_simulation",errmessage,NULL,0);
				continue;
			}
			for(buffcount = 0 ; buffcount < _bufdeep; buffcount++)
				link_obj->data[link_nid * _bufdeep + buffcount] = init_val;
			
		}
	}
	
	for(buffcount = 0; buffcount < mod->names; buffcount++) {
		if(obj_curr->blocks->parvar[buffcount].dp == NULL){
			snprintf(errmessage, 200, "not all names is ported in block %d(module id: %d; filename: %s) of object %ld.",
				 obj_curr->blocks->bid, mod->mid, mod->filename, obj_curr->oid);
			return ERROR_MESSAGE("basic parser:assembling_object_block",errmessage, NULL, -9);
		}
	}
	xmlfn_stepout();
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	assembling_object		!!!done
static int assembling_object()
{
	char *xmlText = NULL;
	long oid = -1l;
	char errmessage[201]; memset(errmessage,0,201);
	
	OUTPUT_MESSAGE("nce:[basic parser]:ASSEMBLING  OBJECT:\t");
	if( (xmlText = xmlfn_getprop( "id" )) == NULL)
		return ERROR_MESSAGE("basic parser:assembling_object","The property \'id\' in the tag <object> is not specified.", NULL, -1);
	
	if( sscanf(xmlText,"%ld",&oid) != 1){
		snprintf(errmessage,200,"The tag <object id=\"%s\"> cannot be processed. The property \'%s\' is corrupted.", xmlText, xmlText);
		return ERROR_MESSAGE("basic parser:assembling_object",errmessage, NULL, -2);
	}
	if( obj_setactive(oid) < 0)
		return ERROR_MESSAGE("basic parser:assembling_object","obj_setactive() returns an error.", NULL, -4);
	
	OUTPUT_MESSAGE("[id=%07ld]: ",oid);
	xmlfn_stepin();
	obj_curr->soft_names = 0;
	for(xmlText = xmlfn_nextnode( NULL ); xmlText != NULL; xmlText = xmlfn_nextnode( NULL )){
		if(!strcmp(xmlText,"block")){
			if( assembling_object_block() < 0){
				snprintf(errmessage,200,"assembling_object_block() returns an error. The tag <object id=\"%ld\"> cannot be processed.",oid);
				return ERROR_MESSAGE("basic parser:assembling_object",errmessage, NULL, -5);
			}
		}
	}
	xmlfn_stepout();
	OUTPUT_MESSAGE(" \t\t\t\tOK\n");
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	assembling_simulation	!!!done. 
static int assembling_simulation()
{
	char *xmlText = NULL;
	int mid = -1;
	mip mod = NULL;
	char errmessage[201]; memset(errmessage,0,201);
	
	xmlfn_stepin();
	while( (xmlText = xmlfn_nextnode( "overstep" )) != NULL){
		OUTPUT_MESSAGE("nce:[basic parser]:SETTING  OVERSTEP:\t");
		if( (xmlText = xmlfn_getprop( "mid" )) == NULL){
			WARNING_MESSAGE("assembling_simulation","The property \'mid\' in the tag <overstep> is not specified. The tag <overstep> is ignored.", NULL, 0);
			continue;
		}

		if( sscanf(xmlText,"%d",&mid) != 1){
			snprintf(errmessage,200,"The tag <overstep mid=\"%s\"> cannot be processed. The property \"%s\" is corrupted.", xmlText,xmlText);
			WARNING_MESSAGE("assembling_simulation",errmessage, NULL, 0);
			continue;
		}
		if( mid < 0 )
			return ERROR_MESSAGE("basic parser:assembling_simulation","The module ID is negative.", NULL, -3);
		if( (mod = mod_get_item(mid)) == NULL ){
			snprintf(errmessage,200,"Cannot find <module id=\"%d\"> for oversteping. The tag <overstep mid=\"%d\"> is ignored.",mid, mid);
			WARNING_MESSAGE("assembling_simulation",errmessage,NULL,0);
			continue;
		}
		if( mod->jump == NULL ){
			snprintf(errmessage,200,"The <module id=\"%d\"> does not support the overstep mode.",mid);
			return ERROR_MESSAGE("basic parser:assembling_simulation", errmessage, NULL, -3);
		}
		if( (xmlText = xmlfn_getcontent()) == NULL ){
			snprintf(errmessage,200,"Cannot read a content of the tag <overstep mid=\"%d\">. The tag <overstep mid=\"%d\"> is ignored.",mid,mid);
			WARNING_MESSAGE("assembling_simulation",errmessage,NULL,0);
			continue;
		}
		if( sscanf(xmlText,"%d",&mod->overstep) != 1){
			snprintf(errmessage,200,"The tag <overstep mid=\"%d\">%s</overstep> cannot be processed. The content is corrupted.", mid, xmlText);
			WARNING_MESSAGE("assembling_simulation",errmessage,NULL,0);
			mod->overstep = 0;		// default overstep is 0
			continue;
		}
		OUTPUT_MESSAGE("module = %03d; overstep = %03d \t\t\t\t\tOK\n", mid, mod->overstep );
	}
	xmlfn_stepout();
	return 0;

}

///////////////////////////////////////////////////////////////////////////////
//	assembling	!!!done
static int assembling()
{
	char *xmlText = NULL;
	xmlfn_stepin();

	while( (xmlText = xmlfn_nextnode( NULL )) != NULL){
		if(!strcmp(xmlText,"module")){
			if( assembling_module() < 0 )
				return ERROR_MESSAGE("basic parser:assembling","\'assembling_module()\' returns an error.", NULL, -1);
			else
				continue;
		}
		if(!strcmp(xmlText,"object")){
			if( assembling_object() < 0 )
				return ERROR_MESSAGE("basic parser:assembling","\'assembling_object()\' returns an error.", NULL, -2);
			else
				continue;
		}
		if(!strcmp(xmlText,"simulation")){
			if( assembling_simulation() < 0 )
				return ERROR_MESSAGE("basic parser:assembling","\'assembling_simulation()\' returns an error.", NULL, -5);
			else
				continue;
		}
	}

	xmlfn_stepout();
	return 0;
}


///////////////////////////////////////////////////////////////////////////////
//	free_model	!!!done
static int model_release()
{
	xmlfn_free();
	if(obj_releasehd() < 0)
		WARNING_MESSAGE("basic parser:model_release","obj_releasehd() returns a warning.",NULL,0);
	if( mem_g_release(memory_heandle) != 0)
		WARNING_MESSAGE("basic parser:model_release","mem_g_release() returns a warning.",NULL,0);
	sim_free();
	if( mod_get_number() )mod_removeall();
	watcher_killall();
	return 0;
}

static int model_destroy()
{
	if ( model_release() < 0)
		WARNING_MESSAGE("basic parser:model_destroy","model_release() returns a warning.",NULL,0);
	if( fd >= 0) close(fd);
	fd = -1;
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	asm_model		!!!done
static int model_init()
{
	if( ( fd < 0) || ( lseek(fd, 0, SEEK_SET) != 0) )
		return ERROR_MESSAGE("basic parser: model_init","Handle is wrong file is corrupted.",NULL, -1);
	if ( xmlfn_init(fd) < 0)
		return ERROR_MESSAGE("basic parser: model_init","Cannot initialize MODEL parser.",NULL, -2);
	
	OUTPUT_MESSAGE("\nnce:[basic parser]:VERIFYING  VERSION:");
	if( test_version() < 0 ){
		model_destroy();
		return ERROR_MESSAGE("basic parser:model_init","Either model or application versions are unsupported.",NULL, -3);
	}
	OUTPUT_MESSAGE("\t\t\t\t\t\t\t\t\tOK!\n");

	OUTPUT_MESSAGE("nce:[basic parser]:Compilation of the model is starting. Please wait...\n");
	if( compilation() < 0){
		model_destroy();
		return ERROR_MESSAGE("basic parser:model_init","\'compilation()\' returns an error.",NULL, -4);
	}
	OUTPUT_MESSAGE("nce:[basic parser]:Compilation of the model is DONE\n\n");
		
	//reinit file for Assembling
	OUTPUT_MESSAGE("nce:[basic parser]:REINIT MODEL FILE ...");
	if( xmlfn_reinit() < 0 ){
		model_destroy();
		return ERROR_MESSAGE("basic parser:model","Cannot re-initialize MODEL parser.",NULL, -5);
	}
	if( xmlfn_nextnode( "model" ) == NULL){
		model_destroy();
		return ERROR_MESSAGE("basic parser:model","Cannot find the tag <model>.",NULL, -6);
	}
	OUTPUT_MESSAGE("\t\t\t\t\t\t\t\t\tDONE!\n");
	OUTPUT_MESSAGE("nce:[basic parser]:BUILDING OBJECT HASH ...");
	if( obj_inithd(allnames, allblocksize) < 0){
		model_destroy();
		return ERROR_MESSAGE("basic parser:model","\'obj_inithd()\' returns an error.",NULL, -7);
	}
	OUTPUT_MESSAGE("\t\t\t\t\t\t\t\t\tDONE!\n");

	
	OUTPUT_MESSAGE("nce:[basic parser]:Assembling of the model is starting. Please wait... \n");
	if( assembling() < 0){
		model_destroy();
		return ERROR_MESSAGE("basic parser:model","\'assembling()\' returns an error.",NULL, -8);
	}
	OUTPUT_MESSAGE("nce:[basic parser]:Assembling of the model is DONE \n\n");
	
	xmlfn_free();
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	model_reinit
static int model_reinit()
{
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
//	pars_printinfo
static int pars_printinfo()
{
	OUTPUT_MESSAGE("\n=== Compilation Assembling information ===\n");
	if(obj_root == NULL){
		OUTPUT_MESSAGE("The model is empty\n\n");
		return 0;
	}
	else
		OUTPUT_MESSAGE("\n");
	obp oscan = obj_root;
	while(oscan != NULL){
		OUTPUT_MESSAGE("Object ID: %ld\t\tDatasize: %d\t\t Bufferdeep: %d\n\tDATA TAB\n",oscan->oid,oscan->datasize,_bufdeep);
		OUTPUT_MESSAGE("\t=======================================================================================\n");
		OUTPUT_MESSAGE("\t[OID:MID:BID]   ADDRESS [OID:NID]   VAL[_sw]\t\tVAL[_asw]\t\tNAME\n");
		module *bscan = oscan->blocks;
		int ipr, inid=0;
		while(bscan != NULL){
			mip mit = mod_get_item(bscan->mid);
			for(ipr=0; ipr<mit->names; ipr++){
				obp cobj = obj_root;
				while((cobj != NULL)&&(&cobj->data[inid*_bufdeep] != bscan->parvar[ipr].dp)){
					for(inid=0; (inid < cobj->datasize)&&(&cobj->data[inid*_bufdeep] != bscan->parvar[ipr].dp); inid++);
					if(&cobj->data[inid*_bufdeep] != bscan->parvar[ipr].dp){
						cobj = cobj->next;
						inid=0;
					}
				}
				if(cobj == NULL) continue;
				OUTPUT_MESSAGE("\t[% 3ld:% 3d:% 3d]   %08x[%03ld:%03d]   %e\t%e\t\t%s\n", 
					       bscan->oid, bscan->mid, bscan->bid,
					       (int)bscan->parvar[ipr].dp, cobj->oid,inid,
					       (double)bscan->parvar[ipr].dp[0], (double)bscan->parvar[ipr].dp[_bufdeep-1],mit->parvar[ipr]);
			}
			bscan = bscan->next;
		}
		OUTPUT_MESSAGE("\n");
		oscan = oscan->next;
	}
	return 0;
}


///////////////////////////////////////////////////////////////////////////////
//	global functions
//-----------------------------------------------------------------------------


///////////////////////////////////////////////////////////////////////////////
//	asm_model_grow()
int asm_model_grow()
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	asm_messeger( int pars_command, int command_length)
int asm_messenger(int model_command, int command_length)
{
	char tmpfilename[] = "asmbasic_model.xml.XXXXXX", empty, cmd[COMMAND_CONTEXT_SIZE];
	struct {
		char *command;
		int  *length;
	} command;
	command.command = cmd;
	command.length  = (int*)(cmd + 1);
	double d_value;
	int    i_value, ii_value, iii_value, iv_value;
	
	switch(model_command){
		case ASMBASIC_MODEL_UPLOAD:
			strcpy(tmpfilename,"asmbasic_model.xml.XXXXXX");
			if( fd >= 0) model_destroy();
			if( ( fd = mkstemp(tmpfilename) ) <= 3)
				return WARNING_MESSAGE( "basic assembler", "Cannot create temporary file", NULL, -1);
			unlink(tmpfilename);
			if( rdwr (fd, engine_context.hInput, (size_t)command_length) != command_length )
				return WARNING_MESSAGE("basic assembler", "The pipe unstable!", NULL, -2);
			if( model_init() < 0)
				return ERROR_MESSAGE("basic assembler", "The model is corrupted", NULL, -5);
			if( sim_init() < 0 )
				return ERROR_MESSAGE("basic assembler", "Cannot init simulation", NULL, -6);
			break;
		case ASMBASIC_MODEL_REINIT:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			if( model_reinit() < 0)
				return ERROR_MESSAGE("basic assembler", "Cannot reinit() model", NULL, -7);
			break;
		case ASMBASIC_MODEL_RELEASE:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			if( model_release() < 0)
				return ERROR_MESSAGE("basic assembler", "Cannot release model", NULL, -8);
			break;
		case ASMBASIC_MODEL_DESTROY:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			if( model_destroy() < 0 )
				return ERROR_MESSAGE("basic assembler", "Cannot destroy model", NULL, -9);
			break;
		case ASMBASIC_MODEL_SAVE:
		case ASMBASIC_MODEL_RESTORE:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			return  WARNING_MESSAGE("basic assembler", "Command ASMBASIC_MODEL_SAVE and ASMBASIC_MODEL_RESTORE currently unsupported by modules and basic assemblers", NULL , 3);
			break;
		case ASMBASIC_DATA_GET:
			if( command_length != (sizeof( int ) * 2) ){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
				 return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_DATA_GET command", NULL, -19);
			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			d_value = obj_getdata(i_value, ii_value);
			*(command.command) = ASMBASIC_RET_DATA_GET;
			*command.length = sizeof(datum);
			write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
			write(engine_context.hOutput,&d_value, sizeof(datum) );
			break;
		case ASMBASIC_DATA_GET_BUFFER :
			if(command_length != (sizeof( int ) * 2) ){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
				return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_DATA_GET_BUFFER command", NULL, -19);
			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			*(command.command) = ASMBASIC_RET_DATA_GET_BUFFER;
			*command.length = sizeof(datum) * _bufdeep;
			write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
			for(iii_value=0; iii_value<_bufdeep; iii_value++){
				d_value = obj_getbuffer(i_value, ii_value, iii_value);
				write(engine_context.hOutput,&d_value, sizeof(datum) );
			}
			break;
		case ASMBASIC_DATA_SET:
			if(command_length != (sizeof( int ) * 2 + sizeof(double) ) ){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
				return WARNING_MESSAGE("basic assemebler", "Erroneous format of NCE_DATA_SET command", NULL, -20);
			}
			read(engine_context.hInput, &i_value,  sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			read(engine_context.hInput, &d_value,  sizeof(double) );
			if( obj_setdata(i_value, ii_value, d_value) < 0)
				WARNING_MESSAGE("basic assemebler", "The function obj_getdata() returns an error",NULL, -18);
			break;				
		case ASMBASIC_WATCHER_ADD_ITEM:
			if(command_length != ( sizeof( int ) * 3)){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));	
				return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_WATCHER_ADD_ITEM command", NULL, -31);
			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			read(engine_context.hInput, &iii_value, sizeof(int) );
			iv_value = ii_value * 1000 + iii_value;
			if( watchitem_add(i_value, iv_value, WATCHER_ITEM_LAST, obj_getdatapointer( ii_value, iii_value ) ) < 0 )
				return WARNING_MESSAGE("basic assembler", "Function watchitem_add() returns an error",NULL, -32);
			*command.command = ASMBASIC_RET_WATCHER_ITEM_ADD_ID;
			*command.length  = sizeof( int );
			write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
			write(engine_context.hOutput,&iv_value, sizeof(int) );
			break;
		case ASMBASIC_WATCHER_ADD_ITEM_MIN:
			if(command_length != ( sizeof( int ) * 3)){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));	
				return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_WATCHER_ADD_ITEM command", NULL, -31);
			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			read(engine_context.hInput, &iii_value, sizeof(int) );
			iv_value = ii_value * 1000 + iii_value;
			if( watchitem_add(i_value, iv_value, WATCHER_ITEM_MINIMAL, obj_getdatapointer( ii_value, iii_value ) ) < 0 )
				return WARNING_MESSAGE("basic assembler", "Function watchitem_add() returns an error",NULL, -32);
			*command.command = ASMBASIC_RET_WATCHER_ITEM_ADD_ID;
			*command.length  = sizeof( int );
			write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
			write(engine_context.hOutput,&iv_value, sizeof(int) );
			break;
		case ASMBASIC_WATCHER_ADD_ITEM_MAX:
			if(command_length != ( sizeof( int ) * 3)){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));	
				return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_WATCHER_ADD_ITEM command", NULL, -31);
			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			read(engine_context.hInput, &iii_value, sizeof(int) );
			iv_value = ii_value * 1000 + iii_value;
			if( watchitem_add(i_value, iv_value, WATCHER_ITEM_MAXIMAL, obj_getdatapointer( ii_value, iii_value ) ) < 0 )
				return WARNING_MESSAGE("basic assembler", "Function watchitem_add() returns an error",NULL, -32);
			*command.command = ASMBASIC_RET_WATCHER_ITEM_ADD_ID;
			*command.length  = sizeof( int );
			write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
			write(engine_context.hOutput,&iv_value, sizeof(int) );
			break;
		case ASMBASIC_WATCHER_ADD_ITEM_AVERAGE:
			if(command_length != ( sizeof( int ) * 3)){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));	
				return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_WATCHER_ADD_ITEM command", NULL, -31);
			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			read(engine_context.hInput, &iii_value, sizeof(int) );
			iv_value = ii_value * 1000 + iii_value;
			if( watchitem_add(i_value, iv_value, WATCHER_ITEM_AVERAGE, obj_getdatapointer( ii_value, iii_value ) ) < 0 )
				return WARNING_MESSAGE("basic assembler", "Function watchitem_add() returns an error",NULL, -32);
			*command.command = ASMBASIC_RET_WATCHER_ITEM_ADD_ID;
			*command.length  = sizeof( int );
			write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
			write(engine_context.hOutput,&iv_value, sizeof(int) );
			break;
		case ASMBASIC_WATCHER_KILL_ITEM:
			if(command_length != ( sizeof( int ) * 3)){
				while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
				return WARNING_MESSAGE("basic assembler", "Erroneous format of NCE_WATCHER_KILL_ITEM command", NULL, -33);

			}
			read(engine_context.hInput, &i_value, sizeof(int) );
			read(engine_context.hInput, &ii_value, sizeof(int) );
			read(engine_context.hInput, &iii_value, sizeof(int) );
			if( watchitem_kill(i_value, ii_value * 1000 + iii_value) < 0 )
				return WARNING_MESSAGE("basic assembler", "Function watchitem_kill() returns an error",NULL, -34);
			break;			
		case ASMBASIC_PRINT_INFO:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			pars_printinfo();
			obj_printinfo();
			break;
		case ASMBASIC_PRINT_INFO_PARSER:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			pars_printinfo();
			break;
		case ASMBASIC_PRINT_INFO_OBJECTS:
			while((command_length--) > 0) read(engine_context.hInput, &empty,  sizeof(char));
			obj_printinfo();
			break;

	}
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
//	asm_printinfo
int asm_printinfo()
{
	pars_printinfo();
	obj_printinfo();
	return 0;
}
