#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <neurocad/engine/modmanager.h>
#include <neurocad/engine/nce_common.h>

static mip root = NULL;
static mip curr = NULL;
int mod_number = 0;

inline static int find_modulele(int imodid)
{
	char errmessage[10]; memset(errmessage, 0,10); snprintf(errmessage,9,"%d",imodid);
	if(root == NULL) return ERROR_MESSAGE("find_modulele","List of Modules is empty root = NULL",errmessage,-1);
	if((curr == NULL) || (curr->mid > imodid))curr = root;
	while((curr != NULL) && (curr->mid != imodid)) curr = curr->next;
	if(curr == NULL) return ERROR_MESSAGE("find_modulele","Can not found  module",errmessage,-1);
	return curr->mid;
}


int mod_insert(char *mod_file_name, int mid)
{
	char errmessage[512]; memset(errmessage,0,512);
	snprintf(errmessage,200,"%s, %d",mod_file_name,mid);
	char dlmessage[512]; memset(dlmessage,0,512);
	void *hmod = dlopen(mod_file_name, RTLD_LAZY);
	if(hmod == NULL){
		snprintf(dlmessage, 511, "Cannot dlopen file! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert","dlmeddage",errmessage,-1);
	}
	enum __mt *pmodtype = (enum __mt*)dlsym(hmod,"modtype");
	enum __ml *pmodmultiload = (enum __ml*)dlsym(hmod,"modmulti");
	int *pnbiter = dlsym(hmod,"moditer");
	int *ppriority = dlsym(hmod,"modprior");
	int *pnbnames  = dlsym(hmod,"modnames");
	int *pblocksize = dlsym(hmod,"modblocksize");
	int *pweight = dlsym(hmod,"modweight");
	char *pmodpar = dlsym(hmod,"modpar");
	char *pmodvar = dlsym(hmod,"modvar");
	int (*modinit)(int) = (int (*) (int)) dlsym(hmod,"init");
	int (*modfree)() = (int (*) (int)) dlsym(hmod,"release");
	module *(*modadd)(int, module *, void*) = (module * (*) (int, module *, void*)) dlsym(hmod,"add");
	int (*calculate)(int) = (int(*)(int))dlsym(hmod,"calculate");
	int (*slack)(int)= (int(*)(int))dlsym(hmod,"slack");
	int (*restore)(int,int,int,int)= (int(*)(int,int,int,int))dlsym(hmod,"restore");
	int (*save)(int,int,int,int)= (int(*)(int,int,int,int))dlsym(hmod,"save");

	if((pmodtype == NULL) || (pmodmultiload == NULL)|| (pnbiter == NULL) || (pmodpar == NULL) || (pmodvar == NULL) || (pnbnames == NULL) ||
	   (ppriority == NULL) || (pweight == NULL) || (pblocksize == NULL) ||
	   (modinit == NULL) || (modfree == NULL) || (modadd == NULL) ||
	   (calculate == NULL) || (slack == NULL) || (restore == NULL) || (save == NULL)){
		dlclose(hmod);
		return ERROR_MESSAGE("mod_inserte","Module hasn't markable attribute",errmessage,-2);
	}

	mip scan = root;
	for(; (scan != NULL) && ( strcmp(scan->filename,mod_file_name) ); scan = scan->next);
	if( (scan != NULL) && ( scan->multiload != __multi ) ){
		dlclose(hmod);
		return ERROR_MESSAGE("mod_insert","Loaded module cannot support multiloder mode",errmessage,-3);
	}
	mip newitem = (mip)mem_malloc(sizeof(moditem));
	if(newitem == NULL){
		dlclose(hmod);
		return ERROR_MESSAGE("mod_insert","Can not malloc memory for New Module Item",errmessage,-4);
	}
	
	newitem->type = *pmodtype;
	newitem->multiload = *pmodmultiload;
	newitem->mid = mid;
	newitem->iterations = *pnbiter;
	newitem->names = *pnbnames;
	newitem->priority = *ppriority;
	newitem->overstep = 0;
	newitem->blocksize = *pblocksize;
	newitem->weight = *pweight;

	newitem->modadd = NULL;
	newitem->calculate = NULL;
	newitem->slack = NULL;
	newitem->jump = NULL;
	newitem->restore = NULL;
	newitem->save = NULL;
	newitem->hmod = NULL;
	newitem->filename = NULL;
	newitem->parvar = NULL;
	newitem->next = NULL;
	
	DB_INFO("modmanager"," prepare stirngs");
	if( (newitem->parvar  = (char **)mem_malloc(sizeof(char *)*(newitem->names+1))) == NULL){
		dlclose(hmod);
		mem_free(newitem);
		return ERROR_MESSAGE("mod_insert","Cannot malloc pointers for Parameters and Variables",errmessage,-5);
	}
	int cnt;
	for(cnt=0; cnt < (newitem->names+1); cnt++)newitem->parvar[cnt] = NULL;

	int stringlen = strlen(mod_file_name)+strlen(pmodpar)+strlen(pmodvar)+10;
	newitem->filename = (char*) mem_malloc (stringlen);
	if(newitem->filename == NULL){
		dlclose(hmod);
		mem_free(newitem->parvar);
		mem_free(newitem);
		return ERROR_MESSAGE("mod_insert","Can not malloc string for FileName+Parameters+Variables",errmessage,-6);
	}
	memset(newitem->filename, 0, stringlen);
	
	int shift = snprintf(newitem->filename, stringlen, "%s", mod_file_name) +1;
	char *pparvar = newitem->filename + shift, *scanparvar;
	if(strlen(pmodpar) != 0)
		if(strlen(pmodvar) != 0)
			shift += snprintf(pparvar, stringlen - shift, "%s\t%s", pmodpar, pmodvar) +1;
		else
			shift += snprintf(pparvar, stringlen - shift, "%s", pmodpar) +1;
	else
		if(strlen(pmodvar) != 0)
			shift += snprintf(pparvar, stringlen - shift, "%s", pmodvar) +1;
	newitem->parvar[0] = pparvar;
	
	scanparvar = pparvar;
	for(cnt=1; cnt <newitem->names; cnt++){
		scanparvar = strchr(scanparvar,'\t');
		if((scanparvar != NULL)&&(*scanparvar == '\t')){
			*scanparvar = '\0';
			newitem->parvar[cnt] = ++scanparvar;
		}
		else
			newitem->parvar[cnt] = NULL;
	}
	for(cnt=0; cnt <newitem->names; cnt++){
		if(newitem->parvar[cnt] == NULL){
			dlclose(hmod);
			mem_free(newitem->parvar);
			mem_free(newitem->filename);
			mem_free(newitem);
			return ERROR_MESSAGE("mod_insert","Number of names more then real names in module",errmessage,-7);
		}
	}
	DB_INFO("modmanager","END of parvar's");
	dlclose(hmod);
	
	if(root == NULL){
		root = newitem;
		return ++mod_number;
	}
	
	if(root->mid > mid){
		newitem->next = root;
		root = newitem;
		return ++mod_number;
	}
	
	if(root->mid == mid) return ERROR_MESSAGE("mod_insert","Module with such id has been made",errmessage,-8);
	
	for(curr = root; (curr != NULL) && (curr->next != NULL) && (curr->next->mid < mid); curr=curr->next);
	
	if(curr->next == NULL){
		curr->next = newitem;
		curr->next->next = NULL;
		return ++mod_number;
	}

	if(curr->next->mid == mid) return ERROR_MESSAGE("mod_insert","Module with such id has been made",errmessage,-9);
	
	
	newitem->next = curr->next;
	curr->next = newitem;
	return ++mod_number;
}

int mod_insert_empty(int mid)
{
	char errmessage[512]; memset(errmessage,0,512);
	snprintf(errmessage,200,"%d",mid);
	mip newitem = (mip)mem_malloc(sizeof(moditem));
	if(newitem == NULL)
		return ERROR_MESSAGE("mod_insert_empty","Can not malloc memory for New Module Item",errmessage,-1);
	newitem->mid = mid;
	newitem->type = 0;
	newitem->multiload = 0;
	newitem->iterations = 0;
	newitem->names = NULL;
	newitem->priority = 0;
	newitem->overstep = 0;
	newitem->blocksize = 0;
	newitem->weight = 0.0;

	newitem->modadd = NULL;
	newitem->calculate = NULL;
	newitem->slack = NULL;
	newitem->jump = NULL;
	newitem->restore = NULL;
	newitem->save = NULL;
	newitem->hmod = NULL;
	newitem->filename = NULL;
	newitem->parvar = NULL;
	newitem->next = NULL;

	if(root == NULL){
		root = newitem;
		return ++mod_number;
	}
	
	if(root->mid > mid){
		newitem->next = root;
		root = newitem;
		return ++mod_number;
	}
	
	if(root->mid == mid) return ERROR_MESSAGE("mod_insert_empty","Module with such id has been made",errmessage,-2);
	
	for(curr = root; (curr != NULL) && (curr->next != NULL) && (curr->next->mid < mid); curr=curr->next);
	
	if(curr->next == NULL){
		curr->next = newitem;
		curr->next->next = NULL;
		return ++mod_number;
	}

	if(curr->next->mid == mid) return ERROR_MESSAGE("mod_insert_empty","Module with such id has been made",errmessage,-3);
	
	newitem->next = curr->next;
	curr->next = newitem;
	return ++mod_number;

}


int mod_remove(int mid)
{
	char errmessage[201]; memset(errmessage,0,201);
	mip cur = root, dell = root;
	if(root->mid == mid){
		root = root->next;
		mem_free(dell->filename);
		mem_free(dell->parvar);
		if(dell->hmod != NULL){
			int (*modfree)() = dlsym(cur->hmod,"modfree");
			if( modfree != NULL ) (*modfree)();
			dlclose(dell->hmod);
		}
		mem_free(dell);
		return --mod_number;
	}
	while((cur->next != NULL)&&(cur->next->mid != mid)) cur = cur->next;
	if(cur->next == NULL){
		snprintf(errmessage,200,"%d",mid);
		return  ERROR_MESSAGE("mod_remove","Can not Find ModID",errmessage,-1);
	}
	dell = cur->next;
	cur->next = cur->next->next;
	mem_free(dell->filename);
	mem_free(dell->parvar);
	if(dell->hmod != NULL){
		int (*modfree)() = dlsym(dell->hmod,"modfree");
		if( modfree != NULL ) (*modfree)();
		dlclose(dell->hmod);
	}
	mem_free(dell);
	return --mod_number;
}

int mod_removeall()
{
	if(root == NULL) return ERROR_MESSAGE("mod_removeall","List of Modules is empty root = NULL",NULL,-1);
	mod_number = 0;
	while(root != NULL){
		curr = root;
		root = root->next;
		mem_free(curr);
	}
	return 0;

}

int mod_get_number()
{
	return (root == NULL) ? 0 :  mod_number;
}

mip mod_get_item(int mid)
{
	char errmessage[10];snprintf(errmessage,10,"%d",mid);
	mip cur = root;
	while((cur != NULL) && (cur->mid != mid))cur = cur->next;
	if(cur == NULL) return (mip)ERROR_MESSAGE("mod_get_item","Can not found  module",errmessage,(int)NULL);
	return cur;
}

mip mod_get_list()
{
	return root;
}

int mod_init(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) < 0) return ERROR_MESSAGE("mod_init","find_modulele() return error",errmessage,-1);
	if((curr->hmod != NULL) && (curr->modadd != NULL) && (curr->calculate != NULL) && (curr->slack != NULL))
		return curr->mid;
	curr->hmod = dlopen(curr->filename, RTLD_LAZY);
	if(curr->hmod == NULL) return ERROR_MESSAGE("InitModule","hmod is NULL",errmessage,-2);
	curr->modadd    = dlsym(curr->hmod,"add");
	curr->calculate = dlsym(curr->hmod,"calculate");
	curr->slack	= dlsym(curr->hmod,"slack");
	curr->jump	= dlsym(curr->hmod,"jump");
	curr->restore	= dlsym(curr->hmod,"restore");
	curr->save	= dlsym(curr->hmod,"save");
	if((curr->modadd  == NULL)||
	   (curr->calculate == NULL)||(curr->slack == NULL)){
		dlclose(curr->hmod);
		curr->modadd = NULL;
		curr->calculate = NULL;
		curr->slack = NULL;
		curr->jump = NULL;
		curr->restore = NULL;
		curr->save = NULL;
		curr->hmod = NULL;
		return ERROR_MESSAGE("mod_init","Cannot find function add, del, calculate or slack",errmessage,-2);
	}
	int (*modinit)(int) = dlsym(curr->hmod,"init");
	if(modinit == NULL){
		dlclose(curr->hmod);
		curr->modadd = NULL;
		curr->calculate = NULL;
		curr->slack = NULL;
		curr->jump = NULL;
		curr->restore = NULL;
		curr->save = NULL;
		curr->hmod = NULL;
		return ERROR_MESSAGE("mod_init","Cannot find in module function init()",errmessage,-3);
	}
	int modinitret = (*modinit)(curr->mid);
	if( modinitret < 0){
		dlclose(curr->hmod);
		curr->modadd = NULL;
		curr->calculate = NULL;
		curr->slack = NULL;
		curr->jump = NULL;
		curr->restore = NULL;
		curr->save = NULL;
		curr->hmod = NULL;
		char errmessageInint[100];snprintf(errmessageInint,100,"init() return error %d",modinitret);
		return ERROR_MESSAGE("mod_init",errmessageInint,errmessage,-4);
	}
	return curr->mid;
}

int mod_free(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) < 0) return ERROR_MESSAGE("mod_free","find_modulele() return error",errmessage,-1);
	if(curr->hmod == NULL) ERROR_MESSAGE("mod_free","hmod is NULL",errmessage,-2);
	int (*release)() = (int (*)())dlsym(curr->hmod,"release");
	if( release == NULL) 
		ERROR_MESSAGE("mod_free","Cannot find release",errmessage,-3);
	else{
		int releaseref = (*release)();
		if(releaseref < 0) {
			char cError[100];
			snprintf(cError,100,"%s:release() return error: %d",curr->filename,releaseref);
			ERROR_MESSAGE("mod_free",cError,errmessage,-3);
		}
		dlclose(curr->hmod);
	}
	curr->modadd = NULL;
	curr->calculate = NULL;
	curr->slack = NULL;
	curr->jump = NULL;
	curr->hmod = NULL;
	return 0;	
}

int mod_set_extended_attribute(int mid, char *name, char *value)
{
	char errmessage[201]; memset(errmessage,0,201); snprintf(errmessage,200,"%d, %s, %s",mid,name,value);
	if(find_modulele(mid) < 0)
		return ERROR_MESSAGE("mod_set_extended_attribute","find_modulele() return error",errmessage,-1);
	if(curr->hmod == NULL)
		return ERROR_MESSAGE("mod_set_extended_attribute","hmod is NULL",errmessage,-2);
	char **pAttr = dlsym(curr->hmod,name);
	if(pAttr == NULL){
		char cError[200];
		snprintf(cError,199,"Cannot found attribute \"%s\" in <module id = %d>",name,mid);
		return ERROR_MESSAGE("mod_set_extended_attribute",cError,errmessage,-3);
	}
	if( (*pAttr = bufdup(NULL, value)) == NULL )
		return ERROR_MESSAGE("mod_set_extended_attribute","Cannot set extended attribute",errmessage,-3);
	return 0;
}


int mod_flash_extended_attribute(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) < 0) 
		return ERROR_MESSAGE("mod_flash_extended_attribute","find_modulele() return error",errmessage,-1);
	if(curr->hmod == NULL)
		return ERROR_MESSAGE("mod_flash_extended_attribute","hmod is NULL",errmessage,-2);
	int (*flash)(int)= (int(*)(int))dlsym(curr->hmod,"flash");
	if(flash == NULL)
		return ERROR_MESSAGE("mod_flash_extended_attribute","Cannot found flash function in module",errmessage,-3);
	int flashret = (*flash)(mid);
	if( flashret < 0){
		char cError[200]; snprintf(cError,199,"Flash function return error: %d",flashret);
		return ERROR_MESSAGE("mod_flash_extended_attribute",cError,errmessage,-4);
	}
	return flashret;
}


int mod_printinfo()
{
	OUTPUT_MESSAGE("\n=== Modules information ===\n");
	if(root == NULL){
		OUTPUT_MESSAGE("List of Modules is empty\n\n");
		return 0;
	}
	else
		OUTPUT_MESSAGE("\n");
	OUTPUT_MESSAGE("\n=== Modules Table ===\nmid\t\titerations\tnames\t\tpriority\toverstep\tfile\n");
	OUTPUT_MESSAGE("------------------------------------------------------------------------------------------------------\n");
	for(curr = root ; curr != NULL ; curr = curr->next)
		OUTPUT_MESSAGE("%d\t\t%d\t\t%d\t\t%d\t\t%d\t\t%s\n",
			       curr->mid,curr->iterations,curr->names,curr->priority,curr->overstep,curr->filename);
	OUTPUT_MESSAGE("\n");
	return 0;
}

int mod_is_acctive(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) < 0) 
		return ERROR_MESSAGE("mod_is_acctive","find_modulele() return error",errmessage,-1);
	if( curr->hmod == NULL ) return 0;
	return 1;
}
