#ifndef __MODULE_H_
#define __MODULE_H_ 

#include <math.h>
#include <neurocad/engine/memmanager.h>
#include <neurocad/engine/nce_common.h>

typedef union __dptr {
	datum *dp;
	datum **dpp;
	void   *vp;
} dptr;

/**main module varibales**/
typedef struct __module {
	unsigned long oid;	//object, module and block id
	int  mid, bid;
	struct __module *next;	//using for common tree
	dptr   parvar[];	//set of pointers of parameters and variables
} module;

/**
This main makros must be included in all module structured data (block).
Module is server atom and used in cell, compartment, links(sysnapses), and ather
modeling objecs.
**/
#define __MODULE	unsigned long oid; \
			int mid, bid; \
			module  *next;
			

/**Types of modules**/
enum __mt{
	__cell = 0,	//phenominal neuron
	__cell2h,	//Hodgkin and Huxley neuron or compartment
	__channel,	//channel for Hodgkin and Huxley neurons
	__iondynamic,	//dynamic of ion concentration
	__synapse,	//synapse
	__port,		//port for multijobs
	__iomod		//input - output modules (cells)
}; 

/**Module multiloading**/
enum __ml{
	__single = 0,	//module can be inserted in model only one time
	__multi		//multiloding module
};

/**Common function for all modules**/
int init( int imid );
int release( int imid );
module *add( int imid, module *next, void *memory_handle);
int calculate( int imid );
int slack( int imid );
int jump( int imid );
int restore( int imid, int hfile, int oid, int bid);
int save( int imid, int hfile, int oid, int bid);

/*
Switch and antiswitch for global synchronization
Those variables have definition in the simmanager.c of NeuroCAD Engine
*/
extern int _sw, _asw, _bufdeep;

/*
Global timestep and timecurrent
Those variables have definition in the simmanager.c of NeuroCAD Engine
*/
extern double time_step, time_cur;

/*
Macroses for switch and antiswitch variables
*/
#ifndef GN_SW
	#define GN_SW(__NAME) (currentblk->parvar[__NAME].dp[_sw])
#endif //GN_SW
#ifndef GNASW
	#define GNASW(__NAME) (currentblk->parvar[__NAME].dp[_asw])
#endif //GNASW

#endif

