#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <neurocad/engine/nce_common.h>
#include <neurocad/engine/memmanager.h>

char sendmessage[__MAX_MESSAGE_LENGTH__];

char *bufdup( char * buff, char *str)
{
	if(buff != NULL) mem_free(buff);
	
	if(str == NULL) return NULL;
	int str_len = strlen(str);
	if(str_len < 1) return NULL;
	
	char *nb = (char*) mem_malloc(str_len+1);
	if(nb == NULL) return NULL;
	memset(nb,0, str_len+1);
	return strncpy(nb,str,str_len);
}

ssize_t web_write (int FILEDES, const void *BUFFER, size_t SIZE)
{
	size_t counter = 0; 
	ssize_t result;
	while(counter < (ssize_t) SIZE) {
		if( (result = write(FILEDES, (const void *)((char*)BUFFER + counter), SIZE - (size_t) counter)) < 0)
			return WEB_ERR;
		else counter += result;
	}
	return WEB_OK;
}

ssize_t web_read  (int FILEDES, const void *BUFFER, size_t SIZE)
{
	size_t counter = 0; 
	ssize_t result;
	while(counter < SIZE) {
		if( (result = read(FILEDES, ((char*)BUFFER + counter), SIZE - (size_t) counter)) <= 0)
			return WEB_ERR;
		else counter += (size_t)result;
	}
	return WEB_OK;
}

inline int send_error(int ret, char* message)
{
	static char cmd[COMMAND_CONTEXT_SIZE];
	memset(cmd, 0, COMMAND_CONTEXT_SIZE);
	cmd[0] = NC__RET_ERROR;
	*((int*)(cmd+1)) = strlen(message)+1;
	if(web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE)  == WEB_ERR) work_flag = 0;
	if(web_write(engine_context.hOutput,message, strlen(message)+1) == WEB_ERR) work_flag = 0;
	return ret;
}

inline int send_warning(int ret, char* message)
{
	static char cmd[COMMAND_CONTEXT_SIZE];
	memset(cmd, 0, COMMAND_CONTEXT_SIZE);
	cmd[0] = NC__RET_WARNING;
	*((int*)(cmd+1)) = strlen(message)+1;
	if(web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE)  == WEB_ERR) work_flag = 0;
	if(web_write(engine_context.hOutput,message, strlen(message)+1) == WEB_ERR) work_flag = 0;
	return ret;
}

inline void send_ok()
{
	static char cmd[COMMAND_CONTEXT_SIZE];
	memset(cmd, 0, COMMAND_CONTEXT_SIZE);
	cmd[0] = NC__RET_OK;
	if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) == WEB_ERR) work_flag = 0;
}

inline int send_message(int ret, char* message)
{
	static char cmd[COMMAND_CONTEXT_SIZE];
	memset(cmd, 0, COMMAND_CONTEXT_SIZE);
	cmd[0] = NC__RET_MESSAGE;
	*((int*)(cmd+1)) = strlen(message)+1;
	if(web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE)  == WEB_ERR) work_flag = 0;
	if(web_write(engine_context.hOutput,message, strlen(message)+1) == WEB_ERR) work_flag = 0;
	return ret;
}

inline int send_debug(int ret, char* message)
{
	static char cmd[COMMAND_CONTEXT_SIZE];
	memset(cmd, 0, COMMAND_CONTEXT_SIZE);
	cmd[0] = NC__RET_DEBUG;
	*((int*)(cmd+1)) = strlen(message)+1;
	if(web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE)  == WEB_ERR) work_flag = 0;
	if(web_write(engine_context.hOutput,message, strlen(message)+1) == WEB_ERR) work_flag = 0;
	return ret;
}

int ErrorMessage(const char *pFileName, const int Line, const char *pFuncsionName, const char *pErrorMessage, char *pParam, int ret)
{
	if(pParam == NULL)
		snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nERROR MESSAGE: NeuroCAD Engine (pid:%d)\n%s:%d: function %s() return error \"%d\" [%s]\n\n",
		getpid(),pFileName,Line,pFuncsionName,ret,pErrorMessage);
	else
		snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nERROR MESSAGE: NeuroCAD Engine (pid:%d)\n%s:%d: function %s(%s) return error \"%d\" [%s]\n\n",
		getpid(),pFileName,Line,pFuncsionName,pParam,ret,pErrorMessage);
	return send_error(ret,sendmessage);
}

int WarningMessage(const char *pFileName, const int Line, const char *pFuncsionName, const char *pErrorMessage, char *pParam, int ret)
{
	if(pParam == NULL)
		snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nERROR MESSAGE: NeuroCAD Engine (pid:%d)\n%s:%d: function %s() return error \"%d\" [%s]\n\n",
		getpid(),pFileName,Line,pFuncsionName,ret,pErrorMessage);
	else
		snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nERROR MESSAGE: NeuroCAD Engine (pid:%d)\n%s:%d: function %s(%s) return error \"%d\" [%s]\n\n",
		getpid(),pFileName,Line,pFuncsionName,pParam,ret,pErrorMessage);
	return send_warning(ret, sendmessage);

}

void ncedb_info(const char *modname, const char *modinfo)
{
	snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nDEBUG MESSAGE: NeuroCAD Engine stream (pid:%d)\n%s is report:%s\n",
		getpid(),
		(modname != NULL) ? modname:"Unknown",
		(modinfo != NULL) ? modinfo:"Unknown information");
	send_debug(0,sendmessage);
}

void ncedb_double(const char *modname, const char *modvar, double val)
{
	snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nDEBUG MESSAGE: NeuroCAD Engine stream (pid:%d)\n%s is report: %s = %g\n",
		getpid(),
		(modname != NULL) ? modname:"Unknown",
		(modvar != NULL)  ? modvar :"Unknown variable",
		val);
	send_debug(0,sendmessage);
}

void ncedb_int(const char *modname, const char *modvar, int val)
{
	snprintf(sendmessage,__MAX_MESSAGE_LENGTH__,
		"\nDEBUG MESSAGE: NeuroCAD Engine stream (pid:%d)\n%s is report: %s = %d[%x]\n",
		getpid(),
		(modname != NULL) ? modname:"Unknown",
		(modvar != NULL)  ? modvar :"Unknown variable",
		val,val);
	send_debug(0,sendmessage);
}

