#ifndef __NCAD_COMMON_H__
#define __NCAD_COMMON_H__

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <signal.h>
#include <neurocad/engine/libnce.h>

#ifdef HAVE_CONFIG_H
	#include "../config.h"
#endif
	#ifndef __RDWR_BUFFER_SIZE__
		#define __RDWR_BUFFER_SIZE__		4096
	#endif

	#ifndef __MEMPAGE_SIZE__
		#define __MEMPAGE_SIZE__		4096
	#endif
	#ifndef __MAX_MESSAGE_LENGTH__
		#define __MAX_MESSAGE_LENGTH__		512
	#endif

/**
The work_flag is flag of nce main loop. If work_flag
is set at zero the engine will be deaded.
**/
extern int work_flag;

/**
bufdup() clears the buff, if it isn't NULL, and allocates new memory and copies str there.
Return the pointer of new memory or NULL
 **/
extern char *bufdup( char * buff, char *str);

/**
web_write() and web_read() eliminate the problem of no full reading and writing. If read() or 
write() utilize less then SIZE the web_write() and web_read() will try to complete result.
functions return constant WEB_OK if transaction is successful or WEB_ERR in anouther case.
 **/
#define	WEB_OK	1
#define WEB_ERR	0
extern ssize_t web_write (int FILEDES, const void *BUFFER, size_t SIZE);
extern ssize_t web_read  (int FILEDES, const void *BUFFER, size_t SIZE);

/**
The set of function sending the messages text though binary pipe.
 **/
extern inline int send_error(int ret, char* message);
extern inline int send_warning(int ret, char* message);
extern inline void send_ok();
extern inline int send_message(int ret, char* message);
extern inline int send_debug(int ret, char* message);

/** Functions and Makroses for TEXT printing out messages system**/
extern char sendmessage[__MAX_MESSAGE_LENGTH__];
extern int ErrorMessage(const char *pFileName, const int Line, const char *pFuncsionName, const char *pErrorMessage, char *pParam, int ret);
extern int WarningMessage(const char *pFileName, const int Line, const char *pFuncsionName, const char *pErrorMessage, char *pParam, int ret);


#ifndef __SILENCE__
	#define ERROR_MESSAGE(__Func,__Message,__Param,__Ret) ErrorMessage(__FILE__,__LINE__,__Func,__Message,__Param,__Ret)
	#define WARNING_MESSAGE(__Func,__Message,__Param,__Ret) WarningMessage(__FILE__,__LINE__,__Func,__Message,__Param,__Ret)
	#define OUTPUT_MESSAGE(...) send_message(snprintf(sendmessage, __MAX_MESSAGE_LENGTH__ ,__VA_ARGS__), sendmessage)

#else
	#define ERROR_MESSAGE(__Func,__Message,__Param,__Ret) __Ret
	#define WARNING_MESSAGE(__Func,__Message,__Param,__Ret) __Ret
	#define OUTPUT_MESSAGE(...)
#endif



/**
engine_context is structure for engine environment (see libnce.h)
**/
extern nce_context engine_context;

/** Functions and Makroses for DEBUG printing out messages system**/
extern void ncedb_info(const char *modname, const char *modinfo);
extern void ncedb_double(const char *modname, const char *modvar, double val);
extern void ncedb_int(const char *modname, const char *modvar, int val);


/*Makroses for debug*/
#ifdef __DBFLAG__
	#define DB_INFO(__MOD, __INFO) 		ncedb_info(__MOD, __INFO)
	#define DB_DOUBLE(__MOD, __VAR, __VAL)	ncedb_double(__MOD,  __VAR, __VAL)
	#define DB_INT(__MOD, __VAR, __VAL)	ncedb_int(__MOD,  __VAR, __VAL)
#else
	#define DB_INFO(__MOD, __INFO)
	#define DB_DOUBLE(__MOD, __VAR, __VAL)
	#define DB_INT(__MOD, __VAR, __VAL)
#endif


#endif
