#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <neurocad/engine/watchmanager.h>
#include <neurocad/engine/simmanager.h>
#include <neurocad/engine/memmanager.h>

typedef struct __watchitem{
	int id;				//id of watchered item
	int mode;			//Mode of item
	datum *data;			//pointer on watched data
	struct __watchitem * next;	//the base for linked list of items
} watchitem;

typedef struct __watcher {
	unsigned int counter;		//number of watch items
	int id, fd, pid;		//Wotcher and file IDs
	int frame_counter;		//Counter of frames; set by watcher_init()
	int active;			//Active Flag. If it zero watcher is passive
	int init;			//INIT Flag. If it zero watcher must be init. Set by watcher_init() kill by watchitem_add() watchitem_kill()
	datum period;			//updating period and lastupdate time
	int period_mod, period_cnt;	//Module of period in setps and counter
	watchitem * root;		//the linked list of items
	datum *databuffer;		//Buffer of sending data
	struct __watcher * next;	//the base for linked list of watchers
	struct iovec vec[5];		//Structure for vector output
} watcher;

watcher *root = NULL;

static watcher *Find(int wid){
	char err[200]; memset(err,0,200);
	snprintf(err,199,"%d",wid);
	
	watcher *wscan = root;
	while((wscan != NULL) &&(wscan->id != wid)) wscan = wscan->next;
	if(wscan == NULL)
		return (watcher *)ERROR_MESSAGE("Find","Cannot find watcher",err,0);
	return wscan;
}

static int watcher_init(int wid)
{
	char fnarg[200]; memset(fnarg,0,200);
	snprintf(fnarg,199,"%d",wid);
	watcher *sw = Find(wid);
	if(sw == NULL)
		return ERROR_MESSAGE("watcher_init","Cannot found watcher",fnarg,-1);
	sw->frame_counter = 0;
	sw->init = -1;
	if( (sw->period_mod = (int) floor(sw->period/time_step) ) <= 0){
		sw->period_mod = 0;
		watchitem *si=sw->root;
		for(; si!= NULL; si=si->next) si->mode = WATCHER_ITEM_LAST;
		WARNING_MESSAGE("watcher_init","Set watcher mode in WATCHER_ITEM_LAST",fnarg,sw->id);
	}
	sw->period_cnt = 0;
	sw->period = (double)sw->period_mod * time_step;

	if( sw->databuffer != NULL ) mem_free(sw->databuffer);
	if( (sw->databuffer = (datum*) mem_malloc( sw->counter * sizeof( datum ) ) ) == NULL )
		return ERROR_MESSAGE("watcher_init","Cannot malloc memory for data buffer",fnarg,-2);
	memset(sw->databuffer, 0, sw->counter * sizeof( datum ));

	sw->pid = getpid();
	sw->vec[0].iov_base = (void*) (&sw->frame_counter);
	sw->vec[0].iov_len  = sizeof( int );
	sw->vec[1].iov_base = (void*) (&sw->id);
	sw->vec[1].iov_len  = sizeof( int );
	sw->vec[2].iov_base = (void*) (&sw->pid);
	sw->vec[2].iov_len  = sizeof( int );
	sw->vec[3].iov_base = (void*) (&time_cur);
	sw->vec[3].iov_len  = sizeof( double );
	sw->vec[4].iov_base = (void*) (sw->databuffer);
	sw->vec[4].iov_len  = sw->counter * sizeof( datum );

	struct flock outlock;
	memset(&outlock, 0, sizeof(outlock));
	
	outlock.l_type=F_WRLCK;
	fcntl(sw->fd,F_SETLKW,&outlock);
	
	write(sw->fd, &sw->init, sizeof(int));		//The Header Marker! sw->init = -1
	write(sw->fd, &sw->id, sizeof(int));		//The Watcher ID!
	write(sw->fd, &sw->pid, sizeof(int));		//The Process ID! sw->init = pid
	write(sw->fd, &sw->period, sizeof(double));	//updating period
	write(sw->fd, &sw->counter, sizeof(int));	//NWV
	
	watchitem *wit = sw->root;
	for(; wit != NULL; wit = wit->next){
		write(sw->fd, &wit->id, sizeof(int));	//ID
		write(sw->fd, &wit->mode, sizeof(int));	//Mode
	}
	outlock.l_type=F_UNLCK;
	fcntl(sw->fd,F_SETLKW,&outlock);
	sw->init = 1;
	return 0;
}

int watcher_add(int ifd, double period)
{
	char fnarg[200]; memset(fnarg,0,200);
	snprintf(fnarg,199,"%d, %f",ifd,period);
	if( ifd < 0 )
		return ERROR_MESSAGE("watcher_new","Watcher stream handle under zero",fnarg,-1);
	watcher *nw = (watcher*) mem_malloc(sizeof(watcher));
	if(nw == NULL)
		return ERROR_MESSAGE("watcher_new","Cannot malloc memory for new watcher",fnarg,-2);
	nw->period = period;
	nw->fd = ifd;
	nw->root = NULL;
	nw->counter = 0;
	nw->active = 1;
	nw->init = 0;
	nw->databuffer = NULL;

	nw->next = root;
	root = nw;
	return (nw->id = (root->next == NULL) ? 0 : root->next->id + 1);
}

int watcher_kill(int wid)
{
	char fnarg[20]; memset(fnarg,0,20);
	snprintf(fnarg,19,"%d",wid);
	
	watcher *wscan = root, *wdel;
	watchitem *delwatchit;
	
	if(root->id == wid){
		wdel = root;
		root = root->next;
		while(wdel->root != NULL){
			delwatchit = wdel->root;
			wdel->root = wdel->root->next;
			mem_free(delwatchit);
		}
		mem_free(wdel);
		return 0;
	}
	while((wscan->next != NULL) && (wscan->next->id != wid)) wscan = wscan->next;
	if(wscan->next == NULL)
		return ERROR_MESSAGE("watcher_del","Cannot found watcher",fnarg,-1);
	wdel = wscan->next;
	wscan->next = wscan->next->next;
	while(wdel->root != NULL){
		delwatchit = wdel->root;
		wdel->root = wdel->root->next;
		mem_free(delwatchit);
	}
	mem_free(wdel);
	return 0;
}

int watcher_killall()
{
	while (root != NULL) 
		if(watcher_kill(root->id) < 0 )
			return ERROR_MESSAGE("watcher_killall","watcher_dell() return error",NULL,-1);
	return 0;
}

int watchitem_add(int wid, int id, int item_mode, datum *data)
{
	char fnarg[200]; memset(fnarg,0,200);
	snprintf(fnarg,199,"%d, %d, %d, %ld",wid, id, item_mode, (long int) data);
	watcher *wi_add = Find(wid);
	if(wi_add == NULL)
		return ERROR_MESSAGE("watchitem_add","Find return error",fnarg,-1);
	watchitem *nwi = (watchitem *) mem_malloc(sizeof(watchitem));
	if(nwi == NULL)
		return ERROR_MESSAGE("watchitem_add","Cannot malloc memory for new watchitem",fnarg,-4);
	nwi->id = id;
	nwi->data = data;
	nwi->mode = item_mode;
	nwi->next = wi_add->root;
	wi_add->root = nwi;
	wi_add->counter++;
	wi_add->init = 0;
	return 0;
}

int watchitem_kill(int wid, int id)
{
	char fnarg[200]; memset(fnarg,0,200);
	snprintf(fnarg,199,"%d, %d",wid,id);
	watcher *wi_dell = Find(wid);
	if(wi_dell == NULL)
		return ERROR_MESSAGE("watchitem_free","Find return error",fnarg,-1);
	watchitem *dwi, *swi;
	if( wi_dell->root->id == id ){
		dwi = wi_dell->root;
		wi_dell->root = wi_dell->root->next;
	}
	else{
		for(swi=wi_dell->root; (swi->next != NULL)&&(swi->next->id != id); swi = swi->next);
		if(swi->next == NULL)
			return ERROR_MESSAGE("watchitem_free","Cannot find Watch Item",fnarg, -3);
		dwi = swi->next;
		swi = swi->next->next;
	}
	mem_free(dwi);
	--wi_dell->counter;
	wi_dell->init = 0;
	return 0;
	
}

int watcher_set_mode(int wid, int mode)
{
	char fnarg[200]; memset(fnarg,0,200);
	snprintf(fnarg,199,"%d",wid);
	watcher *sw = Find(wid);
	if(sw == NULL)
		return ERROR_MESSAGE("watcher_set_mode","Cannot found watcher",fnarg,-1);
	sw->active = (mode & WATCHER_MODE_ACTIVE) ? 1 : 0;
	return 1;
}

void watcher_restart()
{
	if(root == NULL) return;
	watcher *sw = root;
	for(;sw != NULL; sw = sw->next){
		sw->period_mod = (int) floor(sw->period/time_step);
		sw->period_cnt = 0;
		sw->period = (double)sw->period_mod * time_step;
		sw->init = 0;
	}
}


int watcher_printinfo()
{
	return 0;
}

void watch()
{
	if(root == NULL) return;
	watcher *sw = root;
	watchitem *si;
	struct flock outlock;
	memset(&outlock, 0, sizeof(outlock));
	int cnt;

	for(;sw != NULL; sw = sw->next){
		sw->period_cnt++;
		if( (!sw->init)   && ( watcher_init(sw->id) < 0) ) continue;
		if(!sw->active) continue;
		if (sw->period_mod > sw->period_cnt) { 
			for(si=sw->root, cnt = 0; si!= NULL; si=si->next, cnt++)
				switch(si->mode){
					default:
					case WATCHER_ITEM_LAST:
						sw->databuffer[cnt] = *(si->data + _sw);
						break;
					case WATCHER_ITEM_MINIMAL:
						if( sw->databuffer[cnt] > *(si->data + _sw))
							sw->databuffer[cnt] = *(si->data + _sw);
						break;
					case WATCHER_ITEM_MAXIMAL:
						if( sw->databuffer[cnt] < *(si->data + _sw))
							sw->databuffer[cnt] = *(si->data + _sw);
						break;
					case WATCHER_ITEM_AVERAGE:
						sw->databuffer[cnt] += *(si->data + _sw);
						break;
				}
		}
		else {
			for(si=sw->root, cnt = 0; si!= NULL; si=si->next, cnt++)
				if( si->mode ==  WATCHER_ITEM_AVERAGE) 
					sw->databuffer[cnt] /= (datum)sw->period_mod;
			outlock.l_type=F_WRLCK;
			fcntl(sw->fd,F_SETLKW,&outlock);
			writev(sw->fd, sw->vec, 5);
			outlock.l_type=F_UNLCK;
			fcntl(sw->fd,F_SETLKW,&outlock);
			
			sw->period_cnt = 0;
			sw->frame_counter++;
			for(si=sw->root, cnt = 0; si!= NULL; si=si->next, cnt++)
				sw->databuffer[cnt] = *(si->data + _sw);
		}
	}
}
