/*
 * file: cell_h2comp_eulexp.c
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_mcapacitance = 0,
	_var_voltage,
	_var_incurrent,
	_var_vcurrent,
	_var_conduct,
	_var_GE,
	_var_lastfire,
#ifdef __DBFLAG__
	_db_var_steadystate_v,
	_db_var_tau_v,
#endif
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype	= __cell2h;
const int	moditer		= 1;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;

const char	modname[] 	= "Example of Hodgkin-Huxley cell or compartment module";
const char	moddescription[]= "Description of module.\nThis module contains the Hodgkin-Huxley neuron (or compartment) solved potential-current equation by fourth-order Runge-Kutta method without channels.\nYou have to insert any channels in the model! See, for example, ion_K_slow.so\nFor module designer: NeuroCAD Worker make two doubles for each parameters or variable. Every step of simulation the globale variables _sw(switch) and _asw(antiswitch) switch to each aver, i.e. _sw=1/_asw=0 or _sw=0/_asw=1. Usage it for conveyer solution passings: parvar[_var_....].dp[_asw] = .... [_var_....].dp[_sw]\0";

const char	modpar[] 	= "Membrane capacitance\0";

#ifdef __DBFLAG__
const char	modvar[] 	= "Membrane potential\tInput current\tVoltage-dependent cell current\tCell conductance\tTotal product of channel conductances and channel reversal potentials\tLast fire time\t__DB__ voltage steady-state\t__DB__ voltage time constant\0";
#else
const char	modvar[] 	= "Membrane potential\tInput current\tVoltage-dependent cell current\tCell conductance\tTotal product of channel conductances and channel reversal potentials\tLast fire time\0";
#endif


static int mid;	/*module id*/
static block *rootblk;


int init( int imid )
{
	mid = imid;
	rootblk  = NULL;
	return 0;
}

int release ( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk =NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	memset( newblk, 0, sizeof( block ) );
	newblk->nextblk = rootblk;
	rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}


int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	block *currentblk;
	double conduct, tau, vinf, lastpot;
	for( currentblk = rootblk; currentblk != NULL; currentblk = currentblk->nextblk){
		conduct = GN_SW(_var_conduct);
		if( conduct != 0. ){
			tau = GN_SW( _par_mcapacitance )/conduct;
			vinf = ( GN_SW( _var_GE )-GN_SW( _var_incurrent ))/conduct;
			GNASW( _var_voltage ) = 
			( GN_SW( _var_voltage )-vinf )*exp( -time_step/tau )+vinf;
		}
#ifdef __DBFLAG__
		else{
			vinf = 0.;
			tau = 0.;
		}
#endif
		lastpot = (_sw == 0) ?(currentblk->parvar[_var_voltage].dp[_bufdeep -1]) :(currentblk->parvar[_var_voltage].dp[_sw - 1]);
		if( ( (GNASW( _var_voltage ) - GN_SW( _var_voltage )) < 0.0) && ( (GN_SW( _var_voltage ) - lastpot) > 0.0) && (GN_SW( _var_voltage ) > 0.0) )
			GNASW(_var_lastfire) = time_cur;
		else
			GNASW(_var_lastfire) = GN_SW(_var_lastfire);
		GNASW(_var_vcurrent) = 0.;
		GNASW(_var_incurrent) = 0.;
		GNASW(_var_conduct) = 0.;
		GNASW(_var_GE) = 0.;
#ifdef __DBFLAG__
		GNASW( _db_var_steadystate_v )	= vinf;
		GNASW( _db_var_tau_v )		= tau;
#endif
	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
const enum __ml modmulti        = __single;
