/*
 * file: conc_eulexp.c
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_conc_out = 0,
	_par_conc_in_min,
	_par_k,
	_par_alpha,
	
	_var_potential,
	_var_current,
	_par_conc,
#ifdef __DBFLAG__
	_db_var_steadystate_ca,
	_db_var_tau_ca,
#endif
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype	= __cell2h;
const int	moditer		= 1;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;

const char	modname[] 	= "Ca-concentration. De Schutter and Bower.\0";
const char	moddescription[]= "Description of module.\nThis module contains ????\nThe model description was taken from paper\nAn Active Membrane Model of the Cerebellar Purkinje Cell I. Simulation of Current Clamps in Slice.\nEric De Schutter and James M. Bower\nJ. Neurophysiol. Vol. 71, No. 1, 1994.\0";

const char	modpar[] 	= "Outside calcium concentration\tMinimal inside calcium concentration\tCoefficient k\tCoefficient alpha\0";

#ifdef __DBFLAG__
const char	modvar[] 	= "Ca potential\tCalcium current\tInside calcium concentration\t__DB__ Ca inside steady-state\t__DB__ Ca inside time constant\0";
#else
const char	modvar[] 	= "Ca potential\tCalcium current\tInside calcium concentration\0";
#endif

static int mid;	/*module id*/
static block *rootblk;


int init( int imid )
{
	mid = imid;
	rootblk  = NULL;
	return 0;
}

int release ( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk =NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}


int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	block *currentblk;
	double inInf, tau, incon;
	for( currentblk = rootblk; currentblk != NULL; currentblk = currentblk->nextblk){
		inInf = GN_SW(_par_conc_in_min) - GN_SW(_par_alpha) * GN_SW(_var_current) / GN_SW(_par_k);
		tau = 1.0/( GN_SW(_par_k) );
		incon = GNASW(_par_conc) = ( GN_SW( _par_conc )-inInf )*exp( -time_step/tau )+inInf;
		GNASW(_var_potential) = 13.27 * log( GNASW( _par_conc_out ) / incon );
//		GNASW(_var_potential) = 135.0;
		GNASW(_var_current) = 0.0;
#ifdef __DBFLAG__
		GNASW( _db_var_steadystate_ca )	= inInf;
		GNASW( _db_var_tau_ca )		= tau;
#endif
	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
const enum __ml modmulti        = __single;
