/*
 * file: ion_Leak_eulexp.c
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#include <stdlib.h>
#include <neurocad/module.h>

enum __chan_parvar{
	_par_mconduct = 0,
	_par_rev_pot,

	_var_voltage,
	_var_current,
	_var_total_vcurrent,
	_var_total_conduct,
	_var_total_GE,
	__end_par_var__,
};

typedef struct __block{
	__MODULE
	dptr parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt modtype = __channel;
const int moditer = 1;
const int modprior = 2;
const int modnames = __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;

const char modname[] = "Leak curent. De Schutter and Bower.\0";

const char moddescription[]= "The module description.\nThis module contains the implementation of leak current.\nThe description was taken from paper\nAn Active Membrane Model of the Cerebellar Purkinje Cell I. Simulation of Current Clamps in Slice\nEric De Schutter and James M. Bower\nJ. Neurophysiol. Vol. 71, No. 1, 1994.\0";

const char modpar[] = "Maximal conductance\tReversal potential\0";
const char modvar[] = "Membrane potential\tChannel current\tVoltage-dependent cell current\tCell conductance\tTotal product of channel conductances and channel reversal potentials\0";

static int mid = -1;
static block *rootblk = NULL;
static int iteration = 0;

int init( int imid )
{
	mid = imid;
	rootblk = NULL;
	iteration = 0;
	return 0;
}

int release ( int imid )
{
	rootblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = ( block *)mem_g_malloc( mem_handle, sizeof( block ) );
	if( newblk == NULL )
		return ( module *) NULL;
	newblk->nextblk = rootblk;
	rootblk = newblk;
	int cnt; for( cnt = 0; cnt < __end_par_var__; rootblk->parvar[cnt++].dp = NULL );
	rootblk->mid = mid;
	rootblk->next = next;
	return ( module *)rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	block *currentblk = rootblk;
	double 	voltage, revpotential, newcur, conduct;
	for( ; currentblk != NULL; currentblk = currentblk->nextblk ){
		voltage = GN_SW( _var_voltage );
		revpotential = GN_SW( _par_rev_pot );
		conduct = GN_SW( _par_mconduct );

		newcur = ( voltage-revpotential )*conduct;

		GNASW( _var_total_conduct ) += conduct;
		GNASW( _var_total_GE ) += conduct*revpotential;
		GNASW( _var_current ) = newcur;
		GNASW( _var_total_vcurrent ) += newcur;
	}
	return iteration;
}

int slack( int imid )
{
	if( rootblk != NULL )
		return iteration;
	return -1;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
const enum __ml modmulti        = __single;
