/*
 * file: ion_icom_eulexp.c
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#include <stdlib.h>
#include <neurocad/module.h>

enum __chan_parvar{
        _par_conduct_host2client = 0,
	_par_conduct_client2host,
	
	_var_voltage_host,
	_var_voltage_client,
        _var_total_vcurrent_host,
        _var_total_vcurrent_client,
	_var_total_conduct_host,
	_var_total_conduct_client,
	_var_total_GE_host,
	_var_total_GE_client,
        __end_par_var__,
};

typedef struct __block{
        __MODULE
        dptr parvar[__end_par_var__];
        struct __block *nextblk;
} block;

const enum __mt modtype = __channel;
const int moditer = 1;
const int modprior = 2;
const int modnames = __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;

const char modname[] = "Intercommunication channel.\0";

const char moddescription[]= "The module description.\nThis module contains the implementation of \"ion channel\" which provides intercommunications between two compartments.\0";

const char modpar[] = "Maximal conductance host2client\tMaximal conductance client2host\0";
const char modvar[] = "Membrane potential host\tMembrane potential client\tVoltage-dependent compartment current host\tVoltage-dependent compartment current client\tCompartment conductance host\tCompartment conductance client\tTotal product of channel conductances and channel reversal potentials host\tTotal product of channel conductances and channel reversal potentials client\0";

static int mid = -1;
static block *rootblk = NULL;
static int iteration = 0;

int init( int imid )
{
	mid = imid;
	rootblk = NULL;
	iteration = 0;
	return 0;
}

int release ( int imid )
{
	rootblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = ( block *)mem_g_malloc( mem_handle, sizeof( block ) );
	if( newblk == NULL )
		return ( module *) NULL;
	newblk->nextblk = rootblk;
	rootblk = newblk;
	int cnt; for( cnt = 0; cnt < __end_par_var__; rootblk->parvar[cnt++].dp = NULL );
	rootblk->mid = mid;
	rootblk->next = next;
	return ( module *)rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	block *currentblk = rootblk;
	double 	voltage_host, voltage_client, diff_v,
		newcur_host, newcur_client,
		conduct_host2client, conduct_client2host;
	for( ; currentblk != NULL; currentblk = currentblk->nextblk ){
		voltage_host = GN_SW( _var_voltage_host );
		voltage_client = GN_SW( _var_voltage_client );
		conduct_host2client = GN_SW( _par_conduct_host2client );
		conduct_client2host = GN_SW( _par_conduct_client2host );

		diff_v = voltage_host-voltage_client;
		newcur_host = conduct_client2host*diff_v;
		newcur_client = -conduct_host2client*diff_v;

		GNASW( _var_total_conduct_host ) += conduct_client2host;
		GNASW( _var_total_conduct_client ) += conduct_host2client;
		GNASW( _var_total_GE_host ) += conduct_client2host*voltage_client;
		GNASW( _var_total_GE_client ) += conduct_host2client*voltage_host;
		GNASW( _var_total_vcurrent_host ) += newcur_host;
		GNASW( _var_total_vcurrent_client ) += newcur_client;

	}
	return iteration;
}

int slack( int imid )
{
	if( rootblk != NULL )
		return iteration;
	return -1;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
const enum __ml modmulti        = __single;
