#include <unistd.h>
#include <stdlib.h>
#include <neurocad/module.h>

enum __chan_parvar{
	_par_mconduct = 0,
	_par_rev_pot,
	_par_ca_con,
	_par_temperature,
	_var_voltage,
	_var_total_current,
	_var_current,
        _var_total_conduct,
        _var_total_GE,
	_var_gate_m,
#ifdef __DBFLAG__
	_var_m_int,
	_var_m_tau,
#endif
	__end_par_var__,
};




typedef struct __block{
	__MODULE
	dptr parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt modtype = __channel;
const enum __ml modmulti = __single;
const int moditer  = 1;
const int modprior = 2;
const int modnames = __end_par_var__;
const int modblocksize = sizeof(block);
const double modweight = -1.0;

const char modname[] = "The Ca2 -dependent K channel";
const char moddescription[]= "This module contains the implementation of the calcium dependent potassium channel.\nThe channel description was taken from paper\nMainen Z.K. Sejnowski T.J. Nature v.382 96(model I)with correction from http://senselab.med.yale.edu/senselab/ModelDB/ShowModel.asp?model=2488\nThe different equasion are solved by exponentioal Euler method and compatible with compartment/single cell module.\0";

const char modpar[] =
		"Maximal conductance [mS/cm^2]"
		"\tReversal potential [mV]"
		"\tInside calcium concentration [mM]"
		"\ttemperature coefficient"
		"\0";
const char modvar[] =
		"Membrane potential [mV]"
		"\tCell current [mkA/cm^2]"
		"\tChannel current [mkA/cm^2]"
		"\tCell conductance"
		"\tg*E"
		"\tGate []"
#ifdef __DBFLAG__
		"\tActivation steady state"
		"\tActivation time constant"
#endif
		"\0";

static int mid = -1;
static block *rootblk = NULL;


int init( int imid )
{
	mid = imid;
	rootblk = NULL;
	return 0;
}

int release( int imid )
{
	rootblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = ( block *)mem_g_malloc(mem_handle, sizeof( block ));
	if( newblk == NULL )
		return ( module *) NULL;
	newblk->nextblk = rootblk;
	rootblk = newblk;
	int cnt; for( cnt = 0; cnt < __end_par_var__; rootblk->parvar[cnt++].dp = NULL );
	rootblk->mid = mid;
	rootblk->next = next;
	return ( module *)rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	block *currentblk = rootblk;
	datum voltage;
	datum newconduct, newcur, rev;
	datum newgatem, alpham, tm, m;

	for( ; currentblk != NULL; currentblk = currentblk->nextblk ){
		rev = GN_SW( _par_rev_pot );
		voltage = GN_SW( _var_voltage );
		newconduct = GN_SW( _par_mconduct );
		
		alpham = 0.01 * GN_SW(_par_ca_con);
		tm = 1./(alpham+0.02);
		m = alpham*tm;
		tm *= GN_SW( _par_temperature );
		GNASW( _var_gate_m ) = newgatem = 
			( GN_SW( _var_gate_m )-m )*exp( -time_step /tm )+m;
		newconduct *= newgatem;
		
		GNASW( _var_total_conduct ) += newconduct;
		GNASW( _var_total_GE ) += newconduct*rev;
		GNASW( _var_current ) = newcur = newconduct * ( voltage-rev );
		GNASW( _var_total_current ) += newcur;
#ifdef __DBFLAG__
		GNASW( _var_m_int) = m;
		GNASW( _var_m_tau) = tm;
#endif
		

	}
	return 0;
}



int slack( int imid )

{
	if( rootblk != NULL )
		return 0;
	return -1;
}




int restore( int iMID, int hFile, int oid, int bid )

{
	return 0;
}




int save( int iMID, int hFile, int oid, int bid )

{
	return 0;
}

