#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_mcapacity = 0,
	_var_voltage,
	_var_incurrent,
	_var_chcurrent,
	_var_lastfire,
	__end_par_var__
};

typedef struct __block{
	__MODULE
			dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __cell2h;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Hodgkin-Huxley cell or compartment module";
const char	modpar[] 	= "membrane capacity\0";
const char	modvar[] 	= "voltage\tinput current\tchannels current\tlast fire time\0";
const char	moddescription[]= "Description of module.\nThis module contains the Hodgkin-Huxley neuron(or compartment) solved potential-current equation by fourth-order Runge-Kutta method without channels.\nYou have to insert any channels in the model! See, for example, ion_K_slow.so\nFor module designer: NeuroCAD Worker make two doubles for each parameters or variable. Every step of simulation the globale variables _sw(switch) and _asw(antiswitch) switch to each aver, i.e. _sw=1/_asw=0 or _sw=0/_asw=1. Usage it for conveyer solution passings: parvar[_var_....].dp[_asw] = .... [_var_....].dp[_sw] \0";



static int mid;	/*module id*/
static block *rootblk, *currentblk;


int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}


int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	double voltage, current, capacity;
	for(currentblk=rootblk; currentblk != NULL; currentblk=currentblk->nextblk){
		voltage = GN_SW( _var_voltage );
		current = GN_SW( _var_incurrent )+GN_SW( _var_chcurrent );
		capacity = GN_SW( _par_mcapacity );
		
		GNASW(_var_voltage) = voltage-time_step*current/capacity;
		GNASW(_var_chcurrent) = 0.0;
		GNASW(_var_incurrent) = 0.0;

	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
