#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_mcapacity = 0,
	_var_voltage,
	_var_incurrent,
	_var_chcurrent,
	_var_lastfire,
	__end_par_var__
};

typedef struct __block{
	__MODULE
			dptr	parvar[__end_par_var__];
	struct __block *nextblk;
	/*Passes of solution*/
	double result;
} block;

const enum __mt	modtype		= __cell2h;
const enum __ml	modmulti	= __single;
const int	moditer		= 4;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Hodgkin-Huxley cell or compartment module";
const char	modpar[] 	= "membrane capacity\0";
const char	modvar[] 	= "voltage\tinput current\tchannels current\tlast fire time\0";
const char	moddescription[]= "Description of module.\nThis module contains the Hodgkin-Huxley neuron(or compartment) solved potential-current equation by fourth-order Runge-Kutta method without channels.\nYou have to insert any channels in the model! See, for example, ion_K_slow.so\nFor module designer: NeuroCAD Worker make two doubles for each parameters or variable. Every step of simulation the globale variables _sw(switch) and _asw(antiswitch) switch to each aver, i.e. _sw=1/_asw=0 or _sw=0/_asw=1. Usage it for conveyer solution passings: parvar[_var_....) = .... [_var_....) \0";

static int mid;	/*module id*/
static block *rootblk, *currentblk;
static int iteration;


int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	iteration = 0;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * pNewStruct = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(pNewStruct == NULL) return (module*) NULL; /*Cannot malloc memory*/
	pNewStruct->nextblk = rootblk;
	currentblk = rootblk = pNewStruct;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module*) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	if(iteration > 3) return -2;
	(iteration == 0) ? iteration=3 : iteration-- ;
	datum voltage;
	switch(iteration){
	/*The First Iteration*/
	case 3:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_voltage) = currentblk->result = GN_SW(_var_voltage) -
				time_step*(GN_SW(_var_incurrent) + GN_SW(_var_chcurrent))
				/(GN_SW(_par_mcapacity));
			GNASW(_var_chcurrent) = 0.0;
			GNASW(_var_incurrent) = 0.0;
		}
		return iteration;
		break;
	/*The Second and Third Iteration*/
	case 2:
	case 1:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_voltage) = voltage = GN_SW(_var_voltage) -
				time_step*2.0*(GNASW(_var_incurrent) + GNASW(_var_chcurrent))
				/(GN_SW(_par_mcapacity)+GNASW(_par_mcapacity));
			currentblk->result += 2.0 * voltage;
		}
		return iteration;
		break;
	case 0:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk)
			GNASW(_var_voltage) = ( currentblk->result + (GN_SW(_var_voltage) -time_step*(GNASW(_var_incurrent) + GNASW(_var_chcurrent)) /(GNASW(_par_mcapacity))))/6.0;
		return iteration;
		break;
	default: return -3;
	}
	return -4;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
