#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_threshold = 0,
	_par_timeconst,
	_par_capacity,
	_var_voltage,
	_var_incurrent,
	_var_lastfire,
	__end_par_var__
};

typedef struct __block{
	__MODULE
			dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __cell;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of phenology cell module (Integrate - and - Fire)";
const char	modpar[] 	= "threshold potential\tmembrane time constant\tmembrane capacity\0";
const char	modvar[] 	= "voltage\tinput current\tlast fire time\0";
const char	moddescription[]= "Description of module.\nThis module contains the simpel Integrate - and - Fire neuron solved by Euler method.\n.......";

static int mid;	/*module id*/
static block *rootblk, *currentblk;

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_s_malloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;	rootblk->bid = (rootblk->next == NULL) ? 0 : rootblk->next->bid + 1;

	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
		GNASW(_var_voltage) += GN_SW(_var_incurrent)/ GN_SW(_par_capacity) + 
			GN_SW(_var_voltage) *(1.0 - time_step /GN_SW(_par_timeconst));
		GNASW(_var_incurrent) = 0.0;
		if(GNASW(_var_voltage) > GN_SW(_par_threshold)){
			GNASW(_var_voltage) = 0.;
			GNASW(_var_lastfire) = time_cur;
		}
		else
			GNASW(_var_lastfire) = -1.0;
	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
