#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_a = 0,
	_par_b,
	_par_c,
	_par_d,
	_var_v,
	_var_u,
	_var_i,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	double p_fire, th_int;
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __cell;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "E.M. Izhikevich quadratic neuron";
const char	modpar[] 	= "a\tb\tc\td\0";
const char	modvar[] 	= "v\tu\ti\0";
const char	moddescription[]= "Description of module.\nThe model quadratic neuron supposed by Eugene M. Izhikevich\n See for details: Eugene M. Izhikevich\"Which Model to Use for Cortical Spiking Neurons\"IEEE TRANSACTIONS ON NEURAL NETWORKS, VOL. 15, NO. 5, SEPTEMBER 2004";

static int mid;	/*module id*/
static block *rootblk, *currentblk;

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_s_malloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *pMod, void *hmemory )
{
	block * newblk = (block *) mem_g_malloc( hmemory, sizeof( block ));
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;	rootblk->bid = (rootblk->next == NULL) ? 0 : rootblk->next->bid + 1;

	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = pMod;
	rootblk->p_fire = 0.0;
	rootblk->th_int = 0.0;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	double v, u, i;
	for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
		v = GN_SW(_var_v);
		u = GN_SW(_var_u);
		i = GN_SW(_var_i);
		GNASW(_var_i) = 0.0;
		v += time_step * 0.5 * (0.04*v*v + 5.0*v + 140.0 - u + i);
		v += time_step * 0.5 * (0.04*v*v + 5.0*v + 140.0 - u + i);
		if(v >= 30.0){
			GNASW(_var_v) = GN_SW(_par_c);
			GNASW(_var_u) = u + GN_SW(_par_d);
			continue;
		}
		GNASW(_var_v) = v;
		GNASW(_var_u) = u + time_step * GN_SW(_par_a) * ( GN_SW(_par_b) * v - u );
	}
	return 0;
}


int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
