#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_cap = 0,
	_par_timeconst,
	//threshold
	_par_th_init,
	_par_th_rising,
	_par_th_falling,
	_par_th_firing,
	//after polarization
	_par_ap_sing,
	_par_ap_res,
	_par_ap_cap,
	//spike generator
	_par_sp_rev_pot,
	_par_sp_rising,
	_par_sp_falling,
	_var_voltage,
	_var_th_voltage,
	_var_ap_voltage,
	_var_incurrent,
	_var_lastfire,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	double p_fire, th_int;
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __cell;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 1;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Modified Integrate - and - Fire neuron";
const char	modpar[] 	= "main compartment capacity\tmembrane time constant\tthreshold potential\tthreshold rising time\tthreshold falling time\tthreshold firing constant\tafter polarization single amplitude\tafter polarization resistance\tafter polarization capacity\tspike generator reversal potential\tspike generator rising time\tspike generator fallingtime\0";
const char	modvar[] 	= "voltage\tthreshold voltage\tafter polarization voltage\tinput current\tlast fire time\0";
const char	moddescription[]= "Description of module.\nThe originaly model of Modified Integrate - and - Fire neuron by Ruben Tikidji - Hamburyan.";

static int mid;	/*module id*/
static block *rootblk, *currentblk;

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_s_malloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;	rootblk->bid = (rootblk->next == NULL) ? 0 : rootblk->next->bid + 1;

	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	rootblk->p_fire = 0.0;
	rootblk->th_int = 0.0;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	double du_ap, u, u_ap, u_th, ap_r, th_time;
	for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
		u = GN_SW(_var_voltage);
		u_ap = GN_SW(_var_ap_voltage);
		u_th = GN_SW(_var_th_voltage);
		ap_r = GN_SW(_par_ap_res);
		du_ap = u  - u_ap;
		GNASW(_var_voltage) = u + ( GN_SW(_var_incurrent) + u_ap/ap_r - u * ( 1.0/ap_r + GN_SW(_par_cap)/GN_SW(_par_timeconst) )) * time_step / GN_SW(_par_cap);
		GNASW(_var_incurrent) = 0.0;
		th_time = 1.0;
		GNASW(_var_lastfire) = GN_SW(_var_lastfire);
		if( currentblk->p_fire < ( GN_SW(_par_sp_rising) + time_step) ){
			GNASW(_var_voltage) += (GN_SW(_par_sp_rev_pot) - u ) * time_step / GN_SW(_par_sp_rising);
			GNASW(_var_ap_voltage) = u_ap + GN_SW(_par_ap_sing) * time_step /( GN_SW(_par_sp_rising) + GN_SW(_par_sp_falling) );
			th_time = GN_SW(_par_th_firing);
		}
		else if( currentblk->p_fire < ( GN_SW(_par_sp_rising) + GN_SW(_par_sp_falling) + time_step) ){
			GNASW(_var_voltage) += (u - GN_SW(_par_sp_rev_pot) ) * time_step / GN_SW(_par_sp_falling);
			GNASW(_var_ap_voltage) = u_ap + GN_SW(_par_ap_sing) * time_step /( GN_SW(_par_sp_rising) + GN_SW(_par_sp_falling) );
			th_time = GN_SW(_par_th_firing);
		}
		else if(u > u_th){
			GNASW(_var_lastfire) = time_cur;
			currentblk->p_fire = 0.0;
		}
		else
			GNASW(_var_ap_voltage) = u_ap + du_ap * time_step /(GN_SW(_par_ap_cap) * ap_r);

		if((GN_SW(_par_th_rising) > 0.0) & (GN_SW(_par_th_falling) > 0.0)){
			double dth = u - currentblk->th_int;
			currentblk->th_int += (dth < 0) ?
				dth * th_time * time_step / GN_SW(_par_th_falling):
				dth * th_time * time_step / GN_SW(_par_th_rising);
			GNASW(_var_th_voltage) = GN_SW(_par_th_init) + currentblk->th_int;
		}
		else GNASW(_var_th_voltage) = GN_SW(_par_th_init);
		currentblk->p_fire += time_step;
		
	}
	return 0;
}


int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
