#include <neurocad/module.h>
enum {
	_par_tau_m = 0,
	_par_tau_R,
	_par_tau_E,
	_par_c_beta,
	_par_k_beta,
	_par_tau_Ca,
	_par_a_cyc,
	_par_a_is,
	_par_n_is,
	_par_tau_is,
	_var_l,
	_var_R,
	_var_E,
	_var_cG,
	_var_alpha,
	_var_beta,
	_var_Ca,
	_var_i_os,
	_var_g_out,
	_var_v_out,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __channel;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Cone photo receptor";
const char	modpar[] 	= 
	"Outer segment time constant[ms]"
	"\tR* time constconstant[ms]"
	"\tE* time constconstant[ms]"
	"\tcGMP hydrolysis rate constant in darkness[1/ms]"
	"\tcGMP hydrolysis rate constant [1/(ms td)]"
	"\tCa2+ time constant[ms]"
	"\tGC activation scaling coefficient[]"
	"\tscaling constant of membrane nonlinearity[]"
	"\tparamter of membrane nonlinearity[]"
	"\ttime constant of membrane nonlinearity[ms]"
	"\0";
const char	modvar[] 	= 
	"Light[tds]"
	"\tR*[]"
	"\tE*[]"
	"\tcGMP[]"
	"\tcGMP Ca2+ gain[]"
	"\tcGMP rate constant[]"
	"\tCa2+[]"
	"\tcurrent output segment[]"
	"\tInner segment conductance[]"
	"\tInner segment voltage[mV]"
	"\0";
const char	moddescription[]= "Description of module.\nThis module contains the model of cone photoreceptor descriped by JH vanHateren, TD Lamb \"The photocurrent response of human cone is fast and monophasic\" BMC Neuroscience 2006, 7:34 doi:10.1186/1471-2202-7-34\nWarning: apparent Hill coefficient of GC activation and apparent Hill coefficient of CNG activation is fixed in n_c=4, n_x=1\0";

#define POW_NX(X) (X)
#define POW_NC(X) ((X) * (X) * (X) * (X))

static int mid;	/*module id*/
static block *rootblk, *currentblk;
int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if( rootblk == NULL ) return -1;
	block *currentblk = rootblk;
	double lp, rp, ep, cap, cgp, alphap, betap, iosp, goutp, voutp;
	for( ; currentblk != NULL; currentblk = currentblk->nextblk ){
		lp	= GN_SW( _var_l );
		rp	= GN_SW(_var_R);
		ep	= GN_SW(_var_E);
		cap	= GN_SW(_var_Ca);
		cgp	= GN_SW(_var_cG);
		betap	= GN_SW(_var_beta);
		alphap	= GN_SW(_var_alpha) / betap;
		iosp	= GN_SW(_var_i_os);
		goutp	= GN_SW(_var_g_out);
		voutp	= GN_SW(_var_v_out);
		
		GNASW(_var_R) = lp + (rp - lp) * exp( -time_step / GN_SW(_par_tau_R) );
		GNASW(_var_E) = rp + (ep - rp) * exp( -time_step / GN_SW(_par_tau_E) );
		
		GNASW(_var_beta) =  GN_SW(_par_c_beta) + GN_SW(_par_k_beta) * ep ;
		GNASW(_var_alpha) = 1.0/(1.0+POW_NC(GN_SW(_par_a_cyc)*cap));
		GNASW(_var_cG) = cgp 
				= alphap + (cgp - alphap) * exp(-time_step * betap );
		
		GNASW(_var_i_os) = POW_NX( cgp);
		GNASW(_var_Ca)= iosp + (iosp - cap ) * exp( -time_step / GN_SW(_par_tau_Ca) );
		iosp /= goutp;
		GNASW(_var_v_out) = iosp + (iosp - voutp) * exp( -time_step / GN_SW(_par_tau_m) );
		voutp = GN_SW(_par_a_is) * pow(voutp, GN_SW(_par_n_is) );
		GNASW(_var_g_out) = voutp + (voutp - goutp) * exp( -time_step / GN_SW(_par_tau_is) );

	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}

