/*
file:			noise_modulus.c
Author:			V. Vasilkov			vasilkov_va@mail.ru
Description of programs:
			This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
			License as published by the Free Software Foundation;	
General information:
			This file contains a modules, which generate noises with limited amplituds and periods.
			The model appends the noise value to the 'control' variable.
*/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <neurocad/module.h>
#include <time.h>

enum {
	_par_a_min = 0,
	_par_a_max,
	_par_t_min,
	_par_t_max,
	_var_control,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	double addampn, last_tamp;
	int taun_int;
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __iomod;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 3;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "noise";
const char	modpar[] 	= "a_min\ta_max\tt_min\tt_max\0";
const char	modvar[] 	= "control\0";
const char	moddescription[]= "Description of module.\nThe model generate noises with limited amplituds and periods.The model appends the noise value to the \'control variable\'.";

static int mid;	/*module id*/
static block *rootblk, *currentblk;

	

int init( int imid )
{
	srand ((unsigned)time(NULL));
	
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_s_malloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *pMod, void *hmemory )
{
	block * newblk = (block *) mem_g_malloc( hmemory, sizeof( block ));
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;	rootblk->bid = (rootblk->next == NULL) ? 0 : rootblk->next->bid + 1;

	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = pMod;
	rootblk->taun_int = -1;
	rootblk->addampn = 0.0;
	rootblk->last_tamp = 0.0;
	return (module *) rootblk;
}


int calculate( int imid )
{
	
	if(rootblk == NULL) return -1;
	double na, nt;
	for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
		if(currentblk->taun_int <= 0){
			nt = (GN_SW(_par_t_min) + (GN_SW(_par_t_max) - GN_SW(_par_t_min)) * (( (double) rand()) / (RAND_MAX + 1.0))); 
			currentblk->taun_int = (int) floor(nt / time_step);
			
			na = (GN_SW(_par_a_min) + (GN_SW(_par_a_max) - GN_SW(_par_a_min)) * (( (double) rand()) / (RAND_MAX + 1.0)));
			currentblk->addampn = (na - currentblk->last_tamp) / currentblk->taun_int;
		}
		currentblk->last_tamp += currentblk->addampn;
		GNASW(_var_control) += currentblk->last_tamp;
		currentblk->taun_int--;
		
	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
