/*
file:			threshold_discriminations_modulus.c
Author:			V. Vasilkov			vasilkov_va@mail.ru		
Description of programs:
	This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
	License as published by the Free Software Foundation;	
General information:
	This file contains a modules, which binarised the input datas and returns once if it's value exceeds the threshold.
*/

#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_threshold = 0,
	_par_monitoring,
	_var_output,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	double p_fire, th_int;
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __iomod;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 3;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "threshold discriminations";
const char	modpar[] 	= "threshold\tmonitoring\0";
const char	modvar[] 	= "output\0";
const char	moddescription[]= "Description of module.\nThis  module binarised the input datas and returns once if it's value exceeds the threshold.";

static int mid;	/*module id*/
static block *rootblk, *currentblk;

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_s_malloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *pMod, void *hmemory )
{
	block * newblk = (block *) mem_g_malloc( hmemory, sizeof( block ));
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;	rootblk->bid = (rootblk->next == NULL) ? 0 : rootblk->next->bid + 1;

	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = pMod;
	rootblk->p_fire = 0.0;
	rootblk->th_int = 0.0;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	
	for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk)
		GNASW(_var_output) =  (GN_SW(_par_monitoring) >= GN_SW(_par_threshold)) ? 1.0 : 0.0;
	return 0;
}


int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
