#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <neurocad/module.h>

enum {
	_par_ctrlvar = 0,
	__end_par_var__	
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __iomod;
const enum __ml	modmulti	= __multi;
const int	moditer		= 1;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of input - output module";
const char	modpar[] 	= "";
const char	modvar[] 	= "control variable";
const char	moddescription[]= "Module Description.\nThis module contains the driver which set the parameters (extracted from file, fifo or mem) to selecte controled variable.\n";

typedef struct __module_ctx{
//Intended Attribute
	FILE *hDriver;
	double lasttime, nexttime;
	double lastval, nextval;
	int mid;	/*module id*/
	int blkcnt; /*blocks counter*/
	block *rootblk, *currentblk;
	struct __module_ctx *next;
} module_ctx;


//Extended Attribute
char *ioFileName;
static module_ctx *root_ctx = NULL, *current_ctx;

int init( int imid )
{
	if( root_ctx == NULL )
		root_ctx = current_ctx = (module_ctx *) mem_malloc( sizeof( module_ctx ) );
	else {
		current_ctx = root_ctx;
		while(current_ctx->next != NULL) current_ctx = current_ctx->next;
		current_ctx->next = (module_ctx *) mem_malloc( sizeof( module_ctx ) );
		current_ctx = current_ctx->next;
	}
	if( current_ctx == NULL ) return -1;
	memset(current_ctx, 0,  sizeof( module_ctx )  );
	current_ctx->mid = imid;
	return 0;
}

int release( int imid )
{
	//WARING!!!
	static module_ctx *del_ctx;
	del_ctx = root_ctx;
	while( (del_ctx != NULL) && (del_ctx->mid != imid) ) del_ctx = del_ctx->next;
	if( del_ctx == NULL ) return -1;
	if( del_ctx == root_ctx)
		root_ctx = root_ctx->next;
	else{
		current_ctx = root_ctx;
		while(current_ctx->next != del_ctx) current_ctx = current_ctx->next;
		if(current_ctx->next == NULL) return -2;
		current_ctx->next = del_ctx->next;
	}
	// All BLOCKs of current ctx must release by mem_g_free() as one block
	if (del_ctx->hDriver != NULL) fclose(del_ctx->hDriver);
	mem_free(del_ctx);
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	current_ctx = root_ctx;
	while( (current_ctx->next != NULL) && (current_ctx->mid != imid)) current_ctx = current_ctx->next;
	if( current_ctx == NULL ) return (module *) NULL;
	
	block * newblk = (block *) mem_g_malloc(mem_handle, sizeof( block ));
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = current_ctx->rootblk;
	current_ctx->currentblk = current_ctx->rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)current_ctx->rootblk->parvar[cnt].dp=NULL;
	current_ctx->rootblk->mid = imid;
	current_ctx->rootblk->next = next;
	current_ctx->blkcnt++;
	return (module *) current_ctx->rootblk;
}

int calculate( int imid )
{

//	if(root_ctx == NULL) return -1;
	current_ctx = root_ctx;
	while( (current_ctx != NULL) &&  (current_ctx->mid != imid)) 
		current_ctx = current_ctx->next;
	
	if(current_ctx == NULL ) return -1;
	if(current_ctx->rootblk == NULL) return -2;
	if(current_ctx->lasttime > time_cur) return 0;
	if(current_ctx->nexttime < 0.0){
		for(current_ctx->currentblk = current_ctx->rootblk; (current_ctx->currentblk != NULL); current_ctx->currentblk = current_ctx->currentblk->nextblk){
			current_ctx->currentblk->parvar[_par_ctrlvar].dp[_asw] = current_ctx->nextval;
		}
		return 0;
	}
	if((current_ctx->nexttime < time_cur)&&(!feof(current_ctx->hDriver))){
		current_ctx->lasttime = current_ctx->nexttime;
		current_ctx->lastval  = current_ctx->nextval;
		if(fscanf(current_ctx->hDriver,"%lf %lf", &current_ctx->nexttime, &current_ctx->nextval) < 2){
			current_ctx->nexttime = -1.5;
			current_ctx->nextval = current_ctx->lastval;
			fclose(current_ctx->hDriver);
			current_ctx->hDriver = NULL;
			return 0;
		}
	}
	double cval = current_ctx->lastval + (time_cur - current_ctx->lasttime) * (current_ctx->nextval - current_ctx->lastval) / 
				(current_ctx->nexttime - current_ctx->lasttime);
	for(current_ctx->currentblk = current_ctx->rootblk; (current_ctx->currentblk != NULL); current_ctx->currentblk = current_ctx->currentblk->nextblk){
		current_ctx->currentblk->parvar[_par_ctrlvar].dp[_asw] = cval;
	}
	return 0;
}

int slack( int imid )
{
	current_ctx = root_ctx;
	while( (current_ctx->next != NULL) && (current_ctx->mid != imid)) current_ctx = current_ctx->next;
	if( current_ctx == NULL ) return -1;
	if(current_ctx->rootblk == NULL) return -2;
	return 0;
}

int jump( int imid )
{
	current_ctx = root_ctx;
	while( (current_ctx->next != NULL) && (current_ctx->mid != imid)) current_ctx = current_ctx->next;
	if( current_ctx == NULL ) return -1;
	
	if(current_ctx->rootblk == NULL) return -2;
	if(current_ctx->lasttime > time_cur) return 0;
	if(current_ctx->nexttime < 0.0){
		for(current_ctx->currentblk = current_ctx->rootblk; (current_ctx->currentblk != NULL); current_ctx->currentblk = current_ctx->currentblk->nextblk){
			current_ctx->currentblk->parvar[_par_ctrlvar].dp[_asw] = current_ctx->nextval;
		}
		return 0;
	}
	double cval = current_ctx->lastval + (time_cur - current_ctx->lasttime) * (current_ctx->nextval - current_ctx->lastval) / 
				(current_ctx->nexttime - current_ctx->lasttime);
	for(current_ctx->currentblk = current_ctx->rootblk; (current_ctx->currentblk != NULL); current_ctx->currentblk = current_ctx->currentblk->nextblk){
		current_ctx->currentblk->parvar[_par_ctrlvar].dp[_asw] = cval;
	}
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int flash( int imid )
{
	if(ioFileName == NULL) return -1;
	FILE *hDrv = fopen(ioFileName,"r");
	mem_free(ioFileName);
	ioFileName = NULL;
	
	if(hDrv == NULL) return -2;
	
	current_ctx = root_ctx;
	while( (current_ctx->next != NULL) && (current_ctx->mid != imid)) current_ctx = current_ctx->next;
	if( current_ctx == NULL ) return -3;
	current_ctx->hDriver = hDrv;
		
	if(fscanf(current_ctx->hDriver,"%lf %lf", &current_ctx->lasttime, &current_ctx->lastval) < 2)return -4;
	if(fscanf(current_ctx->hDriver,"%lf %lf", &current_ctx->nexttime, &current_ctx->nextval) < 2)return -5;
	return 0;
}
