#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <neurocad/module.h>

enum {
	_par_mconduct = 0,
	_par_rev_pot,
	_var_voltage,
	_var_chcurrent,
	_var_Kslgate,
	_var_Kslcur,
	__end_par_var__	
};

typedef struct __block{
	__MODULE
			dptr	parvar[__end_par_var__];
	struct __block *nextblk;
	/*Passes of gate solution*/
	double ps0, ps1, ps2, ps3;
	/*Passes of K slow current*/
	double cKps0, cKps1;
	/*temporary storage for voltage*/
	double tmpvolt;
} block;

const enum __mt	modtype		= __channel;
const enum __ml	modmulti	= __single;
const int	moditer		= 4;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Slow K channel for cell_2h model";
const char	modpar[] 	= "K-slow max conductivity\tK-slow reversal potential\0";
const char	modvar[] 	= "voltage\tchannels current\tK-slow current\tK-slow gate\0";
const char	moddescription[]= "Description of module.\nThis module contains Slow K channel for the Hodgkin-Huxley neuron(or compartment)[See for example cell_h2.so]. The main equation for \"K slow gate\"(n) is solved by fourth-order Runge-Kutta method.\0";




static int mid;	/*module id*/
static block *rootblk, *currentblk;
static int iteration;

/*KT/zF lg() => 58mV lg() => 25.18882mV ln() => 0.025V ln()*/
#define _KT_zF_		25.18882
/*activasion/inactivation functions*/
#define __ALPHA(__U) (0.01 *((__U) + 34.0)/(1.0 - exp(-((__U)+34.0)/10.0) ))
#define __BETA(__U)  (0.125 * exp(- ((__U)  + 44.0)/25.0))
#define __TIMES4(__M) ((__M) * (__M) * (__M) * (__M))


int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	iteration = 0;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * pNewStruct = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(pNewStruct == NULL) return (module*) NULL; /*Cannot malloc memory*/
	pNewStruct->nextblk = rootblk;
	currentblk = rootblk = pNewStruct;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	iteration = 0;
	return (module*) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	if(iteration > 3) return -2;
	(iteration == 0) ? iteration = 3 : iteration-- ;
	switch(iteration){
	//The First Iteration//
	case 3:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_Kslgate) = currentblk->ps0 = GN_SW(_var_Kslgate) +
				time_step*(__ALPHA(GN_SW(_var_voltage))*(1.0 -  GN_SW(_var_Kslgate)) -
				__BETA(GN_SW(_var_voltage))* GN_SW(_var_Kslgate));
			GNASW(_var_Kslcur) = currentblk->cKps0 = GN_SW(_par_mconduct) * __TIMES4(currentblk->ps0) *
				(GN_SW(_var_voltage) - GN_SW(_par_rev_pot));
			GNASW(_var_chcurrent) += currentblk->cKps0;
			//We have to store voltage for next itaration becouse this module hase second order modprior//
			currentblk->tmpvolt = GNASW(_var_voltage);

		DB_DOUBLE("K Slow channel", "Pass 1  __ALPHA(GN_SW(_var_voltage))",__ALPHA(GN_SW(_var_voltage)));
		DB_DOUBLE("K Slow channel", "Pass 1  __BETA(GN_SW(_var_voltage))",__BETA(GN_SW(_var_voltage)));
		DB_DOUBLE("K Slow channel", "Pass 1  GN_SW(_var_voltage)",GN_SW(_var_voltage));
		DB_DOUBLE("K Slow channel", "Pass 1  __TIMES4(currentblk->ps0)",__TIMES4(currentblk->ps0));
		DB_DOUBLE("K Slow channel", "Pass 1  GNASW(_var_Kslgate)",GNASW(_var_Kslgate));
		DB_DOUBLE("K Slow channel", "Pass 1  GNASW(_var_Kslcur)",GNASW(_var_Kslcur));

		}
		return iteration;
		break;
	//The Second Iteration//
	case 2:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_Kslgate) = currentblk->ps1 = GN_SW(_var_Kslgate) +
				time_step*(__ALPHA( (currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) *
				(1.0 - (GN_SW(_var_Kslgate) + GNASW(_var_Kslgate))/2.0) -
				__BETA((currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) *
				(GN_SW(_var_Kslgate) + GNASW(_var_Kslgate))/2.0);
			GNASW(_var_Kslcur) = currentblk->cKps1 = (GNASW(_par_mconduct) +
				GN_SW(_par_mconduct)) * __TIMES4((currentblk->ps1 + GN_SW(_var_Kslgate))) * 
				(currentblk->tmpvolt + GN_SW(_var_voltage) -
				GN_SW(_par_rev_pot) - GNASW(_par_rev_pot))/64.0;
			GNASW(_var_chcurrent) += currentblk->cKps1 - currentblk->cKps0;
			currentblk->tmpvolt = GNASW(_var_voltage);

		}
		return iteration;
		break;
	//The Third Iteration//
	case 1:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_Kslgate) = currentblk->ps2 = GN_SW(_var_Kslgate) +
				time_step*(__ALPHA( (currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) *
				(1.0 - (GN_SW(_var_Kslgate) + GNASW(_var_Kslgate))/2.0) -
				__BETA((currentblk->tmpvolt + GN_SW(_var_voltage))/2.0) *
				(GN_SW(_var_Kslgate) + GNASW(_var_Kslgate))/2.0);
			GNASW(_var_Kslcur) = currentblk->cKps0 = (GNASW(_par_mconduct) +
				GN_SW(_par_mconduct)) * __TIMES4((currentblk->ps2 + GN_SW(_var_Kslgate))) *
				(currentblk->tmpvolt + GN_SW(_var_voltage)- 
				GN_SW(_par_rev_pot) - GNASW(_par_rev_pot))/64.0;
			GNASW(_var_chcurrent) += currentblk->cKps0 - currentblk->cKps1;
			currentblk->tmpvolt = GNASW(_var_voltage);
		}
		return iteration;
		break;
	case 0:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			currentblk->ps3 = GN_SW(_var_Kslgate) +
				time_step*( __ALPHA( currentblk->tmpvolt) * (1.0 -  GNASW(_var_Kslgate)) -
				__BETA(currentblk->tmpvolt) * GNASW(_var_Kslgate));
			GNASW(_var_Kslgate) = (currentblk->ps0 + 2.0 * currentblk->ps1 + 2.0 * currentblk->ps2 + currentblk->ps3)/6.0;
			GNASW(_var_Kslcur) = GNASW(_par_mconduct) * 
				__TIMES4(GNASW(_var_Kslgate)) * (GNASW(_var_voltage) - GNASW(_par_rev_pot));
			GNASW(_var_chcurrent) += GNASW(_var_Kslcur) - currentblk->cKps0;
		}
		return iteration;
		break;
	default: return -3;
	}
	return -4;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	switch(iteration){
	case 3:
	case 2:
	case 1:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk)
			currentblk->tmpvolt = GNASW(_var_voltage);
		return iteration;
		break;
	case 0:
		return iteration;
		break;
	default: return -3;
	}
	return -4;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
