#include <stdlib.h>
#include <unistd.h>
#include <neurocad/module.h>

enum {
	_par_mconduct = 0,
	_par_rev_pot,
	_var_voltage,
	_var_chcurrent,
	_var_current,
	_var_total_conduct,
	_var_total_GE,
	__end_par_var__
};

typedef struct __block{
	__MODULE
			dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __channel;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Leak channel";
const char	modpar[] 	= "Leak conductivity\tLeak potential\0";
const char	modvar[] 	= "voltage\tchannels current\tLeak current\tTotal conductance\tGE\0";
const char	moddescription[]= "Description of module.\nThis module contains Leak channel for the Hodgkin-Huxley neuron(or compartment) solved equation by second-order Runge-Kutta method.\nSee for example cell_h2.so\0";


static int mid;	/*module id*/
static block *rootblk, *currentblk;

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_s_malloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if( rootblk == NULL ) return -1;
	block *currentblk = rootblk;
	double newcur, maxconduct, revpotential;
	for( ; currentblk != NULL; currentblk = currentblk->nextblk ){
		maxconduct = GN_SW( _par_mconduct );
		revpotential = GN_SW( _par_rev_pot );
		newcur = maxconduct*( GN_SW( _var_voltage ) - revpotential);
		GNASW( _var_current ) = newcur;
		GNASW( _var_chcurrent ) += newcur;
		GNASW( _var_total_conduct ) += maxconduct;
		GNASW( _var_total_GE ) += maxconduct * revpotential;

	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
