#include <stdlib.h>
#include <unistd.h>
#include <neurocad/module.h>

enum {
	_par_mconduct = 0,
	_par_LEAKpot,
	_var_LEAKcur,
	_var_voltage,
	_var_chcurrent,
	__end_par_var__
};

typedef struct __block{
	__MODULE
			dptr	parvar[__end_par_var__];
	struct __block *nextblk;
	/*temporary storage for voltage*/
	double tmpvolt;
	/*Passes of Leak current*/
	double cLps0, cLps1;
} block;

const enum __mt	modtype		= __channel;
const enum __ml	modmulti	= __single;
const int	moditer		= 4;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Leak channel";
const char	modpar[] 	= "Leak conductivity\tLeak potential\0";
const char	modvar[] 	= "Leak current\tvoltage\tchannels current\0";
const char	moddescription[]= "Description of module.\nThis module contains Leak channel for the Hodgkin-Huxley neuron(or compartment) solved equation by fourth-order Runge-Kutta method.\nSee for example cell_h2.so\0";

static int mid;	/*module id*/
static block *rootblk, *currentblk;
static int iteration;

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	iteration = 0;
	return 0;
}

int release( int imid )
{
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * pNewStruct = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(pNewStruct == NULL) return (module*) NULL; /*Cannot malloc memory*/
	pNewStruct->nextblk = rootblk;
	currentblk = rootblk = pNewStruct;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module*) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	if(iteration > 3) return -2;
	(iteration == 0) ? iteration=3 : iteration-- ;
	switch(iteration){
	case 3:
		for(currentblk=rootblk; currentblk != NULL; currentblk=currentblk->nextblk){
			GNASW(_var_LEAKcur)  = currentblk->cLps0 = GN_SW(_par_mconduct) *
				(GN_SW(_var_voltage) - GN_SW(_par_LEAKpot));
			GNASW(_var_chcurrent) += currentblk->cLps0;
			/*We have to store voltage for next itaration becouse this module hase second order modprior*/
			currentblk->tmpvolt = GNASW(_var_voltage);
		}
		return iteration;
		break;
	case 2:
		for(currentblk=rootblk; currentblk != NULL; currentblk=currentblk->nextblk){
			GNASW(_var_LEAKcur)  = currentblk->cLps1 = 
				(GN_SW(_par_mconduct) + GNASW(_par_mconduct)) *
				(currentblk->tmpvolt + currentblk->tmpvolt - GN_SW(_par_LEAKpot) - GNASW(_par_LEAKpot))/4.0;
			GNASW(_var_chcurrent) += currentblk->cLps1 - currentblk->cLps0;
			currentblk->tmpvolt = GNASW(_var_voltage);
		}
		return iteration;
		break;
	case 1:
		for(currentblk=rootblk; currentblk != NULL; currentblk=currentblk->nextblk){
			GNASW(_var_LEAKcur)  = currentblk->cLps0 =
				 (GN_SW(_par_mconduct) + GNASW(_par_mconduct)) *
				(currentblk->tmpvolt + currentblk->tmpvolt - GN_SW(_par_LEAKpot) - GNASW(_par_LEAKpot))/4.0;
			GNASW(_var_chcurrent) += currentblk->cLps0 - currentblk->cLps1;
		}
		return iteration;
		break;
	case 0:
		for(currentblk=rootblk; currentblk != NULL; currentblk=currentblk->nextblk){
			GNASW(_var_LEAKcur) = GNASW(_par_mconduct) * (GNASW(_var_voltage) - GNASW(_par_LEAKpot));
			GNASW(_var_chcurrent) += GNASW(_var_LEAKcur) - currentblk->cLps0;
		}
		return iteration;
		break;
	default:
		return -3;
		break;
	}
	return -4;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
