#include <stdlib.h>
#include <neurocad/module.h>
enum {
	_par_mconduct = 0,
	_par_rev_pot,
	_var_voltage,
	_var_chcurrent,
	_var_current,
	_var_gate_m,
	_var_gate_h,
	_var_total_conduct,
	_var_total_GE,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype		= __channel;
const enum __ml	modmulti	= __single;
const int	moditer		= 1;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Fast Na channel for cell_2h model";
const char	modpar[] 	= "Na-fast max conductivity\tNa-fast reversal potential\0";
const char	modvar[] 	= "voltage\tchannels current\tNa-fast current\tNa-fast activating gate\tNa-fast inactivating gate\tTotal conductance\tGE\0";
const char	moddescription[]= "Description of module.\nThis module contains Fast Na channel for the Hodgkin-Huxley neuron(or compartment)[See for example cell_h2.so]. The main equation for \"Na fast activating gate\"(m) and \"Na fast inactivating gate\"(h) are solved by fourth-order Runge-Kutta method.\0";

static int mid;	/*module id*/
static block *rootblk, *currentblk;

/*KT/zF lg() => 58mV lg() => 25.18882mV ln() => 0.025V ln()*/
#define _KT_zF_		25.18882
#define __ALPHA_M(__U) ( 0.09 * ((__U) + 38.0)/(1.0 - exp(- ((__U) + 38.0)/8.0)) )
#define __BETA_M(__U)  ( -0.062 * ((__U) + 38.0)/(1.0 - exp( ((__U) + 38.0)/8.0)) )
#define __ALPHA_H(__U) ( 0.016 * exp(- ((__U) + 55.0)/15.0))
#define __BETA_H(__U)  ( 2.07 /( exp( -((__U) - 17.0)/21.0 ) + 1.0) )

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_smalloc().
	//Thus all resources must be free by memmanager.c
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * newblk = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(newblk == NULL) return (module *) NULL; //Cannot malloc memory
	newblk->nextblk = rootblk;
	currentblk = rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module *) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	block *currentblk = rootblk;
	double voltage, alpham, tm, m, newgatem, alphah, th, h, newgateh, newcur, newconduct, revpotential;
	for( ; currentblk != NULL; currentblk = currentblk->nextblk ){
		voltage = GN_SW( _var_voltage );
		
		alpham = __ALPHA_M( voltage );
		tm = 1./( alpham+__BETA_M( voltage ) );
		m = alpham*tm;
		newgatem = ( GN_SW( _var_gate_m )-m )*exp( -time_step /tm )+m;
		
		alphah = __ALPHA_H( voltage );
		th = 1./( alphah+__BETA_H( voltage ) );
		h = alphah*th;
		newgateh = ( GN_SW( _var_gate_h )-h )*exp( -time_step /th )+h;
		
		newconduct =  GN_SW( _par_mconduct ) * pow( newgatem, 3.0 ) * newgateh;
		revpotential = GN_SW( _par_rev_pot );
		newcur = newconduct * ( voltage - revpotential );

		GNASW( _var_total_conduct ) += newconduct;
		GNASW( _var_total_GE ) += newconduct*revpotential;
		
		GNASW( _var_gate_m ) = newgatem;
		GNASW( _var_gate_h ) = newgateh;
		GNASW( _var_current ) = newcur;
		GNASW( _var_chcurrent ) += newcur;

	}
	return 0;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
