#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_mconduct = 0,
	_par_rev_pot,
	_var_voltage,
	_var_chcurrent,
	_var_NAfsactgate,
	_var_NAfsinactgate,
	_var_NAfscur,
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
	/*Passes of gate solutions*/
	double gMps0, gMps1, gMps2, gMps3;
	double gHps0, gHps1, gHps2, gHps3;
	/*Passes of Na fast current*/
	double cNAps0, cNAps1;
	/*temporary storage for voltage*/
	double tmpvolt;
} block;

const enum __mt	modtype		= __channel;
const enum __ml	modmulti	= __single;
const int	moditer		= 4;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;
const char	modname[] 	= "Example of Fast Na channel for cell_2h model";
const char	modpar[] 	= "Na-fast max conductivity\tNa-fast reversal potential\0";
const char	modvar[] 	= "voltage\tchannels current\tNa-fast activating gate\tNa-fast inactivating gate\tNa-fast current\0";
const char	moddescription[]= "Description of module.\nThis module contains Fast Na channel for the Hodgkin-Huxley neuron(or compartment)[See for example cell_h2.so]. The main equation for \"Na fast activating gate\"(m) and \"Na fast inactivating gate\"(h) are solved by fourth-order Runge-Kutta method.\0";



static int mid;	/*module id*/
static block *rootblk, *currentblk;
static int iteration;

/*KT/zF lg() => 58mV lg() => 25.18882mV ln() => 0.025V ln()*/
#define _KT_zF_		25.18882
#define __ALPHA_M(__U) ( 0.09 * ((__U) + 38.0)/(1.0 - exp(- ((__U) + 38.0)/8.0)) )
#define __BETA_M(__U)  ( -0.062 * ((__U) + 38.0)/(1.0 - exp( ((__U) + 38.0)/8.0)) )
#define __ALPHA_H(__U) ( 0.016 * exp(- ((__U) + 55.0)/15.0))
#define __BETA_H(__U)  ( 2.07 /( exp( -((__U) - 17.0)/21.0 ) + 1.0) )
#define __TIMES3(__M)  ((__M) * (__M) * (__M))

int init( int imid )
{
	mid = imid;
	rootblk = currentblk = NULL;
	iteration = 0;
	return 0;
}

int release( int imid )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk = currentblk = NULL;
	return 0;
}

module *add( int imid, module *next, void *mem_handle )
{
	block * pNewStruct = (block *) mem_g_malloc( mem_handle, sizeof( block ) );
	if(pNewStruct == NULL) return (module*) NULL; /*Cannot malloc memory*/
	pNewStruct->nextblk = rootblk;
	currentblk = rootblk = pNewStruct;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = next;
	return (module*) rootblk;
}

int calculate( int imid )
{
	if(rootblk == NULL) return -1;
	if(iteration > 3) return -2;
	(iteration == 0) ? iteration=3 : iteration-- ;
	switch(iteration){
	/*The First Iteration*/
	case 3:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_NAfsactgate) = currentblk->gMps0 = GN_SW(_var_NAfsactgate) +
				time_step*(__ALPHA_M(GN_SW(_var_voltage))*(1.0 - GN_SW(_var_NAfsactgate)) -
				__BETA_M(GN_SW(_var_voltage))* GN_SW(_var_NAfsactgate));
			GNASW(_var_NAfsinactgate) = currentblk->gHps0 = GN_SW(_var_NAfsinactgate) +
				time_step*(__ALPHA_H(GN_SW(_var_voltage))*(1.0 - GN_SW(_var_NAfsinactgate)) -
				__BETA_H(GN_SW(_var_voltage))* GN_SW(_var_NAfsinactgate));
			GNASW(_var_NAfscur) = currentblk->cNAps0 = GN_SW(_par_mconduct) * 
				__TIMES3(currentblk->gMps0) * currentblk->gHps0 *
				(GN_SW(_var_voltage) - GN_SW(_par_rev_pot));
			GNASW(_var_chcurrent) += currentblk->cNAps0;
			/*We have to store voltage for next itaration becouse this module hase second order modprior*/
			currentblk->tmpvolt = GNASW(_var_voltage);
	
		}
		return iteration;
		break;
	/*The Second Iteration*/
	case 2:	
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_NAfsactgate) = currentblk->gMps1 = GN_SW(_var_NAfsactgate) +
				time_step*(__ALPHA_M((currentblk->tmpvolt + GN_SW(_var_voltage))/2.0) * 
				(1.0 - (GNASW(_var_NAfsactgate) + GN_SW(_var_NAfsactgate))/2.0) -
				__BETA_M((currentblk->tmpvolt + GN_SW(_var_voltage))/2.0) *
				(GNASW(_var_NAfsactgate) + GN_SW(_var_NAfsactgate))/2.0);
			GNASW(_var_NAfsinactgate) = currentblk->gHps1 = GN_SW(_var_NAfsinactgate) +
				time_step*(__ALPHA_H((currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) * 
				(1.0 - (GNASW(_var_NAfsinactgate) + GN_SW(_var_NAfsinactgate))/2.0) -
				__BETA_H((currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) *
				(GNASW(_var_NAfsinactgate) + GN_SW(_var_NAfsinactgate))/2.0);
			GNASW(_var_NAfscur) = currentblk->cNAps1 = 
				(GNASW(_par_mconduct) + GN_SW(_par_mconduct)) * 
				__TIMES3((currentblk->gMps1 + GN_SW(_var_NAfsactgate))) *
				(currentblk->gHps1 + GN_SW(_var_NAfsinactgate)) *
				(currentblk->tmpvolt + GN_SW(_var_voltage) -
				GN_SW(_par_rev_pot) - GNASW(_par_rev_pot))/64.0;
			GNASW(_var_chcurrent) += currentblk->cNAps1 - currentblk->cNAps0;
			currentblk->tmpvolt = GNASW(_var_voltage);
		}
		return iteration;
		break;
	/*The Third Iteration*/
	case 1:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			GNASW(_var_NAfsactgate) = currentblk->gMps2 = GN_SW(_var_NAfsactgate) +
				time_step*(__ALPHA_M((currentblk->tmpvolt + GN_SW(_var_voltage))/2.0) * 
				(1.0 - (GNASW(_var_NAfsactgate) + GN_SW(_var_NAfsactgate))/2.0) -
				__BETA_M((currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) *
				(GNASW(_var_NAfsactgate) + GN_SW(_var_NAfsactgate))/2.0);
			GNASW(_var_NAfsinactgate) = currentblk->gHps2 = GN_SW(_var_NAfsinactgate) +
				time_step*(__ALPHA_H((currentblk->tmpvolt+ GN_SW(_var_voltage))/2.0) * 
				(1.0 - (GNASW(_var_NAfsinactgate) + GN_SW(_var_NAfsinactgate))/2.0) -
				__BETA_H((currentblk->tmpvolt + GN_SW(_var_voltage))/2.0) *
				(GNASW(_var_NAfsinactgate) + GN_SW(_var_NAfsinactgate))/2.0);
			GNASW(_var_NAfscur) = currentblk->cNAps0 = 
				(GNASW(_par_mconduct) + GN_SW(_par_mconduct)) * 
				__TIMES3((currentblk->gMps2 + GN_SW(_var_NAfsactgate))) *
				(currentblk->gHps2 + GN_SW(_var_NAfsinactgate)) *
				(currentblk->tmpvolt + GN_SW(_var_voltage) -
				GN_SW(_par_rev_pot) - GNASW(_par_rev_pot))/2.0;
			GNASW(_var_chcurrent) += currentblk->cNAps0 - currentblk->cNAps1;
			currentblk->tmpvolt = GNASW(_var_voltage);
		}
		return iteration;
		break;
	case 0:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk){
			currentblk->gMps3 = GN_SW(_var_NAfsactgate) +
				time_step*(__ALPHA_M(currentblk->tmpvolt) * (1.0 - GNASW(_var_NAfsactgate)) -
				__BETA_M(currentblk->tmpvolt) * GNASW(_var_NAfsactgate));
			currentblk->gHps3 = GN_SW(_var_NAfsinactgate) +
				time_step*(__ALPHA_H(currentblk->tmpvolt) * (1.0 - GNASW(_var_NAfsinactgate)) -
				__BETA_H(currentblk->tmpvolt) * GNASW(_var_NAfsinactgate));
			
			GNASW(_var_NAfsactgate) = (currentblk->gMps0 + 2.0 * currentblk->gMps1 + 2.0 * currentblk->gMps2 + currentblk->gMps3)/6.0;
			GNASW(_var_NAfsinactgate) = (currentblk->gHps0 + 2.0 * currentblk->gHps1 + 2.0 * currentblk->gHps2 + currentblk->gHps3)/6.0;
			GNASW(_var_NAfscur) = GNASW(_par_mconduct) * __TIMES3(GNASW(_var_NAfsactgate)) * GNASW(_var_NAfsinactgate)
				*(GNASW(_var_voltage) - GNASW(_par_rev_pot));
			GNASW(_var_chcurrent) += GNASW(_var_NAfscur) - currentblk->cNAps0;
		}
		return iteration;
		break;
	default: return -3;
	}
	return -4;
}

int slack( int imid )
{
	if(rootblk == NULL) return -1;
	switch(iteration){
	case 2:
	case 1:
		for(currentblk=rootblk; (currentblk != NULL); currentblk=currentblk->nextblk)
			currentblk->tmpvolt = GNASW(_var_voltage);
		return iteration;
		break;
	case 0:
		return iteration;
		break;
	default: return -3;
	}
	return -4;
}

int restore( int imid, int hFile, int oid, int bid )
{
	return 0;
}

int save( int imid, int hFile, int oid, int bid )
{
	return 0;
}
