#include <stdlib.h>
#include <neurocad/module.h>

enum {
	_par_mconduct = 0,
	_par_rev_pot,
	_par_alpha,
	_par_tau,
	_var_pre_voltage,
	_var_voltage,
	_var_incurrent,
	_var_current,
	_var_gate,
#ifdef __DBFLAG__
	_var_steadystate,
#endif /* __DBFLAG__*/
	__end_par_var__
};

typedef struct __block{
	__MODULE
	dptr	parvar[__end_par_var__];
	struct __block *nextblk;
} block;

const enum __mt	modtype	= __cell2h;
const enum __ml modmulti = __single;
const int	moditer		= 1;
const int	modprior	= 2;
const int	modnames	= __end_par_var__;
const int	modblocksize	= sizeof(block);
const double	modweight	= -1.0;

const char	modname[] 	= "AMPA and GABAa synapse";
const char	moddescription[]= "Description of module.\nThis module contains the AMPA and GABAa synapse without time delay.\nThe different equasion are solved by Euler exponentioal method\0";

const char	modpar[] 	= 
		"Maximal conductance"
		"\tReversal potential"
		"\tGain constant alpha"
		"\tTime Constant"
		"\0";

const char	modvar[] 	= 
		"Presynaptic membrane potential"
		"\tMembrane potential"
		"\tInput current"
		"\tSynapstic current"
		"\tSynapstic gate"
#ifdef __DBFLAG__
		"\tSteady state"
#endif /* __DBFLAG__*/
		"\0";


static int mid;	/*module id*/
static block *rootblk;


int init( int iMID )
{
	mid = iMID;
	rootblk  = NULL;
	return 0;
}

int release ( int iModID )
{
	//The module uses the static memory function mem_smalloc()
	//Thus all resources must be free by memmanager.c;
	rootblk =NULL;
	return 0;
}

module *add( int iMID, module *pMod, void *memory_handle )
{
	block * newblk = (block *) mem_g_malloc(memory_handle, sizeof( block ));
	if(newblk == NULL) return (module *) NULL; /*Cannot malloc memory*/
	memset( newblk, 0, sizeof( block ) );
	newblk->nextblk = rootblk;
	rootblk = newblk;
	int cnt; for(cnt=0; cnt<__end_par_var__; cnt++)rootblk->parvar[cnt].dp=NULL;
	rootblk->mid = mid;
	rootblk->next = pMod;
	return (module *) rootblk;
}


int calculate( int iMID )
{
	if(rootblk == NULL) return -1;
	block *currentblk;
	double newcur, newgate, inf, tau;
	for( currentblk = rootblk; currentblk != NULL; currentblk = currentblk->nextblk){
		tau = GN_SW(_par_tau);
		inf =  GN_SW(_par_alpha) * 1.0/(1.0 + exp( -(GN_SW(_var_pre_voltage)-20.0)/10.0 ) ) / tau ;
		newgate = GN_SW(_var_gate);
		GNASW(_var_gate) = newgate = inf + ( newgate - inf ) * exp ( - time_step / tau);
		GNASW(_var_current) = newcur  = GN_SW(_par_mconduct) * newgate * ( GN_SW( _var_voltage ) - GN_SW(_par_rev_pot ) );
		GNASW(_var_incurrent) += newcur;
#ifdef __DBFLAG__
		GNASW( _var_steadystate ) = inf;
#endif /* __DBFLAG__*/
		
	}
	return 0;
}

int slack( int iMID )
{
	if(rootblk == NULL) return -1;
	return 0;
}

int restore( int iMID, int hFile, int oid, int bid )
{
	return 0;
}

int save( int iMID, int hFile, int oid, int bid )
{
	return 0;
}
