#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dlfcn.h>
#include <neurocad/module.h>

int _sw, _asw, _bufdeep;
datum time_step, time_cur;

int main( int cc, char **cv )
{
	if( cc < 2){
		printf("USAGE: %s <module name>\n", cv[0]);
		return 0;
	}
	
	void *hmod;
	if( (hmod = dlopen(cv[cc-1], RTLD_LAZY)) == NULL ){
		fprintf(stderr,"Can not attach module \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	
	enum __mt *modtype;
	enum __ml *modmulti;
	if( (modtype = (enum __mt *)dlsym(hmod,"modtype")) == NULL){
		fprintf(stderr,"Can not found \"modtype\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (modmulti = (enum __ml *)dlsym(hmod,"modmulti")) == NULL){
		fprintf(stderr,"Can not found \"modmulti\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	
	int *moditer, *modprior, *modnames, *modblocksize ;
	if( (moditer = (int *)dlsym(hmod,"moditer")) == NULL){
		fprintf(stderr,"Can not found \"moditer\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (modprior = (int *)dlsym(hmod,"modprior")) == NULL){
		fprintf(stderr,"Can not found \"modprior\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (modnames = (int *)dlsym(hmod,"modnames")) == NULL){
		fprintf(stderr,"Can not found \"modnames\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (modblocksize = (int *)dlsym(hmod,"modblocksize")) == NULL){
		fprintf(stderr,"Can not found \"modblocksize\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	
	double *modweight;
	if( (modweight = (double *)dlsym(hmod,"modweight")) == NULL){
		fprintf(stderr,"Can not found \"modweight\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	
	char *modname, *modpar, *modvar, *moddescription;
	if( (modname = (char *)dlsym(hmod,"modname")) == NULL){
		fprintf(stderr,"Can not found \"modname\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (modpar = (char *)dlsym(hmod,"modpar")) == NULL){
		fprintf(stderr,"Can not found \"modpar\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (modvar = (char *)dlsym(hmod,"modvar")) == NULL){
		fprintf(stderr,"Can not found \"modvar\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	if( (moddescription = (char *)dlsym(hmod,"moddescription")) == NULL){
		fprintf(stderr,"Can not found \"moddescription\"  in \"%s\" library. dlerror() returns: %s\n\n",cv[cc-1], dlerror());
		return 1;
	}
	puts("<module>");
	printf("<lable>%s</lable>\n",modname);
	printf("<file>%s</file>\n",cv[cc-1]);
	printf("<description>\n%s\n</description>\n",moddescription);
	switch (*modtype){
		case __cell:
			printf("<type>cell</type>\n");
			break;
		case __cell2h:
			printf("<type>Hodgkin-Huxley cell</type>\n");
			break;
		case __channel:
			printf("<type>ion channel</type>\n");
			break;
		case __iondynamic:
			printf("<type>ion dynamic</type>\n");
			break;
		case __synapse:
			printf("<type>synapse</type>\n");
			break;
		case __port:
			printf("<type>port</type>\n");
			break;
		case __iomod:
			printf("<type>input-output module</type>\n");
			break;
		default:
			printf("<type>unknown</type>\n");
			break;
	}
	switch (*modmulti){
		case __single:
			printf("<multiload>no</multiload>\n");
			break;
		case __multi:
			printf("<multiload>yes</multiload>\n");
			break;
		default:
			printf("<multiload>unknown</multiload>\n");
			break;
	}
	printf("<iterations>%d</iterations>\n<prioprity>%d</prioprity>\n<names>%d</names>\n<blocksize>%d</blocksize>\n", *moditer, *modprior, *modnames, *modblocksize);
	puts("<parameters>");
	int cnt, edflg=0;
	for(cnt = 0; modpar[cnt] != '\0'; cnt++){
		edflg = 1;
		if(cnt == 0)
			printf("<name>%c",*modpar);
		else if(modpar[cnt] == '\t')
			printf("</name>\n<name>");
		else
			putchar(modpar[cnt]);
	}
	if(edflg)puts("</name>");
	puts("</parameters>");
	
	puts("<variables>");
	edflg=0;
	for(cnt = 0; modvar[cnt] != '\0'; cnt++){
		edflg = 1;
		if(cnt == 0)
			printf("<name>%c", *modvar);
		else if(modvar[cnt] == '\t')
			printf("</name>\n<name>");
		else
			putchar(modvar[cnt]);
	}
	if(edflg)puts("</name>");
	puts("</variables>");
	
	puts("</module>");
	return 0;
}
