#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/sysinfo.h>

#include <neurocad/engine/libnce.h>
#include <neurocad/engine/nce_common.h>
#include "getparam.m"


static nce_context nce_ctx;


static int usage(char *progname)
{
	fprintf(stderr,"\nUSAGE: %s [if=<filename>] [of=<filename>]\n",progname);
	fprintf(stderr,"%s -h or --help              \t\t- this message\n",progname);
	return 0;
}



int main (int cv, char **vv)
{
	int exitcode = 0;
	// scan a command line
	char 	**sv = NULL, 
		*iffilename = NULL, 
		*offilename = NULL,
		*wbffilename= NULL;

	memset(&nce_ctx,0,sizeof(nce_ctx));
	nce_ctx.hInput  =  0;
	nce_ctx.hOutput =  1;
	nce_ctx.hWatch  = -1;
	nce_ctx.id = getpid();
	

	for( sv = vv; *sv != NULL; sv++ ){
		__FIND_PS__( *sv, "if=" ,  iffilename );
		__FIND_PS__( *sv, "of=" ,  offilename );
		__FIND_PS__( *sv, "wbf=", wbffilename );
		if( (!strcmp(*sv, "-h")) || (!strcmp(*sv, "--help")) )
			return usage(vv[0]);
		
	}
	
	fprintf(stderr,"\n===============================================================");
	fprintf(stderr,"\n                     NeuroCAD engine wrapper.");
	fprintf(stderr,"\n                             2004 - 2007");
	fprintf(stderr,"\n===============================================================");
	fprintf(stderr,"\n Copyright(C) Ruben Tikidji - Hanburyan (rth@nisms.krinc.ru)");
	fprintf(stderr,"\n Copyright(C) Sergey Markin             (markin@nisms.krinc.ru)");
	fprintf(stderr,"\n                           Licence: GPL");
	fprintf(stderr,"\n                     Use it without any wary!!!");
	if(iffilename != NULL){
		nce_ctx.hInput = open(iffilename,O_RDONLY,0);
		if (nce_ctx.hInput < 0){
			fprintf(stderr,"\n ERROR: Cannot open input file: %s\n\n",iffilename);
			return 1;
		}
		fprintf(stderr,"\n %s uses %s as input stream \n",vv[0],iffilename);
	}

	if(offilename != NULL){
		nce_ctx.hOutput =  open(offilename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR );
		if(nce_ctx.hOutput < 0){
			fprintf(stderr,"\n ERROR: Cannot open output file: %s\n\n",offilename);
			return 1;
		}
		fprintf(stderr,"\n %s uses %s as output stream \n",vv[0],offilename);
	}

	if(wbffilename != NULL){
		nce_ctx.hWatch =  open(wbffilename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR );
		if(nce_ctx.hWatch < 0){
			fprintf(stderr,"\n ERROR: Cannot open output file: %s\n\n",wbffilename);
			return 1;
		}
		fprintf(stderr,"\n %s uses %s as Watcher stream \n",vv[0],wbffilename);
	}

	// get a system statistics
	struct sysinfo si;
	memset( &si, 0, sizeof( si ));
	if( sysinfo( &si ) < 0 ){
		fprintf( stderr, "Cannot receive the system statistics. \'sysinfo()\' returns an error.\n" );
		return -2;
	}
	nce_ctx.memsize_all = nce_ctx.memsize_free = 
		(( size_t )si.totalram )*(( size_t )si.mem_unit );
	exitcode =  nce( &nce_ctx );

	if(nce_ctx.hInput  > 2) close( nce_ctx.hInput );
	if(nce_ctx.hOutput > 2) close( nce_ctx.hOutput );
	if(nce_ctx.hWatch  > 2) close( nce_ctx.hWatch );
	return exitcode;
}


