#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <neurocad/engine/nce_common.h>
#include <neurocad/engine/memmanager.h>
#include <neurocad/assemblers/basic/xmlfn.h>

static int fd = -1, workdepth = 0, curattr = 0;
static xmlTextReaderPtr reader = NULL;
static char *text_buffer = NULL;

static int getnewbuffer(int size)
{
	if(text_buffer != NULL) mem_free(text_buffer);
	if( (text_buffer = (char*) mem_malloc(size)) == NULL )
		return ERROR_MESSAGE("getnewbuffer","Cannot maaloc memory for buffer",NULL,-1);
	return size;
}

static int str2buff(char *string)
{
	if(string == NULL)return ERROR_MESSAGE("str2buff","string is empty",string,-1);
	int len = strlen(string)+1;
	if( getnewbuffer(len) < 0) return ERROR_MESSAGE("str2buff","getnewbuffer() return error",string,-2);
	memset(text_buffer,0, len);
	if( strncpy(text_buffer,string,len) == NULL)return ERROR_MESSAGE("str2buff","Cannot copy string",string,-3);
	return 0;
}

int   xmlfn_init(int fileid)
{
	if((fd != -1) || (reader != NULL) )xmlfn_free();
	workdepth = 0; curattr = 0;
	if( (reader = xmlReaderForFd(fd = fileid,"neurocad/server/model",NULL,0)) == NULL){
		fd = -1;
		return ERROR_MESSAGE("xmlfn_init","xmlReaderForFd return NULL",NULL, -1);
	}
	return 0;
}

int   xmlfn_free()
{
	if(reader !=  NULL){
		xmlTextReaderClose(reader);
		xmlFreeTextReader(reader);
	}
	reader = NULL;
	fd = -1; workdepth = 0; curattr = 0;
	if(text_buffer !=  NULL) mem_free(text_buffer);
	text_buffer = NULL;
	return 0;
}

int   xmlfn_reinit()
{
	int fileid = fd;
	if(xmlfn_free() < 0) return ERROR_MESSAGE("xmlfn_reinit","Cannot clean context by xmlfn_free()", NULL, -1);
	lseek(fileid,0,SEEK_SET);
	if(xmlfn_init(fileid) < 0 )return ERROR_MESSAGE("xmlfn_reinit","Cannot creat context by xmlfn_init()", NULL, -2);
	return 0;
}

int   xmlfn_stepin()
{
	workdepth++;
	return workdepth;
}

int   xmlfn_stepout()
{
	workdepth--;
	if(workdepth < 0){
		workdepth = 0;
		return 0;
	}
	return workdepth;
}

char  *xmlfn_nextnode(const char *name)
{
	if(reader == NULL) return (char *) ERROR_MESSAGE("xmlfn_nextnode","nothining to read",(char*)name,0);
	curattr = 0;
	xmlChar *xmlText = NULL;
	if( name == NULL){
		while(xmlTextReaderRead(reader) == 1){
			if( xmlTextReaderDepth(reader) < workdepth) return NULL;
			if( xmlTextReaderNodeType(reader) == 1){
				if( xmlTextReaderDepth(reader) != workdepth) continue;
				if( (xmlText = xmlTextReaderName(reader)) == NULL)
					return (char *) ERROR_MESSAGE("xmlfn_nextnode","error of libxml",(char*)name,(int)NULL);
				if( str2buff((char*)xmlText) < 0)
					return (char *) ERROR_MESSAGE("xmlfn_nextnode","str2buff() return error",(char*)name,(int)NULL);
				xmlFree(xmlText);
				return text_buffer;
			}
		}
		return NULL;
	}
	//name != NULL
	while(xmlTextReaderRead(reader) == 1){
		if( xmlTextReaderDepth(reader) < workdepth) return NULL;
		if( xmlTextReaderNodeType(reader) == 1){
			if( xmlTextReaderDepth(reader) != workdepth) continue;
			if( (xmlText = xmlTextReaderName(reader)) == NULL)
				return (char *) ERROR_MESSAGE("xmlfn_nextnode","error of libxml",(char*)name,(int)NULL);
			if( xmlStrcmp((xmlChar*)name,xmlText) ){
				xmlFree(xmlText);
				continue;
			}
			if( str2buff((char*)xmlText) < 0)
				return (char *) ERROR_MESSAGE("xmlfn_nextnode","str2buff() return error",(char*)name,(int)NULL);
			xmlFree(xmlText);
			return text_buffer;
		}
	}
	return NULL;
}
/*
char  *xmlfn_nextlnode(const char *names)
{
}
*/
char *xmlfn_getprop(const char *name)
{
	if(reader == NULL) return (char *) ERROR_MESSAGE("xmlfn_getprop","nothining to read",(char*)name,(int)NULL);
	int allattr;
	xmlChar *NodeText = NULL;
	
	//looking begin or end element
	do{
		if( xmlTextReaderNodeType(reader) == 1 ) break;
		if( xmlTextReaderNodeType(reader) == 15) break;
	}while(xmlTextReaderRead(reader) == 1);
	
	if(( xmlTextReaderNodeType(reader) != 1 )&&( xmlTextReaderNodeType(reader) != 15)) return NULL;
	if(name != NULL){
		NodeText = xmlTextReaderGetAttribute(reader,(xmlChar *) name);
		if(NodeText == NULL) return NULL;
	}
	else{
		if(xmlTextReaderHasAttributes(reader) != 1) return NULL;
		allattr = xmlTextReaderAttributeCount(reader);
		if( (NodeText = xmlTextReaderGetAttributeNo(reader,curattr)) == NULL) return NULL;
		((curattr < 0) || (curattr >=allattr)) ? curattr = 0 : curattr++;
	}
	if( str2buff((char*)NodeText) < 0) return (char *) ERROR_MESSAGE("xmlfn_getprop","str2buff() return error",(char*)name,(int)NULL);
	xmlFree(NodeText);
	return text_buffer;
}

char *xmlfn_propscan()
{
	if(reader == NULL) return (char *) ERROR_MESSAGE("xmlfn_scanprop","nothining to read",NULL,(int)NULL);
	xmlChar *xmlText = NULL;
	do{
		if( xmlTextReaderNodeType(reader) == 1 ) break;
		if( xmlTextReaderNodeType(reader) == 2 ) break;
		if( xmlTextReaderDepth(reader) < workdepth ) return NULL;	
	}while(xmlTextReaderRead(reader) == 1);
	
	if( ( xmlTextReaderNodeType(reader) == 1 )||(xmlTextReaderHasAttributes(reader) == 1) ){
		if( xmlTextReaderMoveToFirstAttribute(reader) != 1) return NULL;
	}
	else if ( xmlTextReaderNodeType(reader) == 2){
		if(xmlTextReaderMoveToNextAttribute(reader) != 1){
			xmlTextReaderMoveToElement(reader);
			return NULL;
		}
	}
	else
		return NULL;
	
	if( (xmlText = xmlTextReaderName(reader)) == NULL)
		return (char *) ERROR_MESSAGE("xmlfn_scanprop","error of libxml",NULL,(int)NULL);
	if( str2buff((char*)xmlText) < 0)
		return (char *) ERROR_MESSAGE("xmlfn_scanprop","str2buff() return error",NULL,(int)NULL);
	xmlFree(xmlText);
	return text_buffer;
}

char *xmlfn_propscan_getvalue()
{
	if(reader == NULL) return (char *) ERROR_MESSAGE("xmlfn_scanprop_getvalue","nothining to read",NULL,(int)NULL);
	xmlChar *xmlText = NULL;
	if( xmlTextReaderNodeType(reader) != 2)
		return (char *) ERROR_MESSAGE("xmlfn_scanprop_getvalue","Pointer isnot attrebute",NULL,(int)NULL);
	if( (xmlText = xmlTextReaderValue(reader)) == NULL)
		return (char *) ERROR_MESSAGE("xmlfn_scanprop_getvalue","error of libxml",NULL,(int)NULL);
	if( str2buff((char*)xmlText) < 0)
		return (char *) ERROR_MESSAGE("xmlfn_scanprop_getvalue","str2buff() return error",NULL,(int)NULL);
	xmlFree(xmlText);
	return text_buffer;	
}

char *xmlfn_getcontent()
{
	if(reader == NULL) return (char *) ERROR_MESSAGE("xmlfn_getcontent","nothining to read",NULL,(int)NULL);
	xmlChar *xmlText = NULL;
	//looking content
	do {
		if(xmlTextReaderNodeType(reader) != 3) continue;
		if(xmlTextReaderDepth(reader) != (workdepth + 1) ) continue;
		
		if( (xmlText = xmlTextReaderValue(reader)) == NULL)
			return (char *) ERROR_MESSAGE("xmlfn_getcontent","str2buff() return error",NULL,(int)NULL);
		if( str2buff((char*)xmlText) < 0) 
			return (char *) ERROR_MESSAGE("xmlfn_getcontent","str2buff() return error",NULL,(int)NULL);
		xmlFree(xmlText);
		return text_buffer;
	}while(xmlTextReaderRead(reader) == 1);
	return NULL;
}

