#include <neurocad/engine/nce_common.h>
#include <neurocad/engine/simmanager.h>
#include <neurocad/engine/asmmanager.h>

static void *asmhandle = NULL;

int (*asm_model_grow_pointer)() = NULL;
int (*asm_messenger_pointer)(int pars_command, int command_length) = NULL;
int (*asm_printinfo_pointer)() = NULL;

int asm_init(const char *asmname)
{
	asm_drop();
	char chfd[256], errormessage[512]; 
	memset(errormessage,0, 512);
	if( asmname == NULL )
		return ERROR_MESSAGE("asm_init","`assembler name\' is NULL",chfd, -1);
	memset(chfd, 0, 256); snprintf(chfd, 255, "%s",asmname);
	
	if( ( asmhandle = dlopen(asmname, RTLD_LAZY) ) == NULL ){
		snprintf(errormessage,511,"Can not attach assembler library. dlerror() returns: %s",dlerror());
		return ERROR_MESSAGE("asm_init",errormessage,chfd, -2);
	}
	
	if( ( asm_model_grow_pointer = dlsym(asmhandle,"asm_model_grow") ) == NULL ){
		snprintf(errormessage,511,"Can not found asm_model_grow() function in assembler library. dlerror() returns: %s",dlerror());
		return ERROR_MESSAGE("asm_init",errormessage,chfd, -3);
	}
	
	if( ( asm_printinfo_pointer = dlsym(asmhandle,"asm_printinfo") ) == NULL ){
		snprintf(errormessage,511,"Can not found asm_printinfo() function in assembler library. dlerror() returns: %s",dlerror());
		return ERROR_MESSAGE("asm_init",errormessage,chfd, -3);
	}
	
	if( ( asm_messenger_pointer = dlsym(asmhandle,"asm_messenger") ) == NULL ){
		snprintf(errormessage,511,"Can not found asm_messenger() function in assembler library. dlerror() returns: %s",dlerror());
		return ERROR_MESSAGE("asm_init",errormessage,chfd, -3);
	}

	return 0;
}

int asm_drop()
{
	if( asmhandle == NULL ) return 0;
	dlclose(asmhandle);
	asmhandle		= NULL;
	asm_model_grow_pointer	= NULL;
	asm_messenger_pointer	= NULL;
	asm_printinfo_pointer	= NULL;
	return 0;
}

