#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/un.h>
#include <sys/sendfile.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <neurocad/engine/libnce.h>
#include <neurocad/engine/nce_common.h>
#include <neurocad/engine/asmmanager.h>
#include <neurocad/engine/memmanager.h>
#include <neurocad/engine/modmanager.h>
#include <neurocad/engine/simmanager.h>
#include <neurocad/engine/watchmanager.h>

nce_context engine_context;
int work_flag = 1;

void nce_stop_simulation_handler ( int signal )
{
	sim_flag = 0;
	return;
}

void nce_get_simulation_status ( int signal )
{
	return;
}


#define clean_pipe() { \
	while(((*command.length)--) > 0) \
		read(engine_context.hInput, &empty,  sizeof(char)); \
}



static int nce_assembler_command(command_header command){
	char *socket__file_name = NULL, empty;
	int ret;
	switch(*command.command){
		case NCE_ASSEMBLER_LOAD:
			socket__file_name = (char*) mem_malloc (*command.length + 1);
			memset(socket__file_name, 0, *command.length + 1);
			if(socket__file_name == NULL)
				return ERROR_MESSAGE("nce_assembler_command","Cannot mem_malloc memory for assembler name string",NULL, -6);
			if ( web_read(engine_context.hInput, socket__file_name, *command.length) != WEB_OK){
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_assembler_command", "Input is broken!",NULL, 12 );
			}
			if( asm_init(socket__file_name) < 0 ){
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_assembler_command","Cannot attach assembler library",NULL, 13 );
			}
			mem_free(socket__file_name);
			socket__file_name = NULL;
			return 0;
			break;
		case NCE_ASSEMBLER_COMMAND:
			if( *command.length < 1 ){
				clean_pipe();
				return WARNING_MESSAGE("nce_assembler_command", "Erroneous format of NCE_ASSEMBLER_COMMAND", NULL, 14);
			}
			if( web_read(engine_context.hInput, &empty, 1 ) != WEB_OK )
				return ERROR_MESSAGE("nce_assembler_command","Cannot read assemblers command. Input is broken!",NULL, -7 );
			if( asm_messenger_pointer == NULL ) 
				return WARNING_MESSAGE("nce_assembler_command", "Assembler isn't load", NULL, 15 );
			if( (ret = (*asm_messenger_pointer)((int)empty, (*command.length) - 1 )) < 0 )
				return WARNING_MESSAGE("nce_assembler_command", "The asm_messanger() return error", NULL, 16 );
			return 0;
			break;
		case NCE_ASSEMBLER_KILL:
			clean_pipe();
			if(asm_drop() < 0)
				return ERROR_MESSAGE("nce_assembler_command", "The function asm_drop() return an error", NULL, -8);
			return 0;
			break;
			
	}
	return ERROR_MESSAGE("nce_assembler_command","Unknown command",NULL, -9 );
}

static int nce_simulation_command(command_header command, char*cmd){
	char empty;
	datum d_value;
	int    i_value, ret;	
	switch(*command.command){
		case NCE_SIMUALTION_STEP:
			clean_pipe();
			if(sim_step() < 0)
				return ERROR_MESSAGE("nce_simulation_command", "The function sim_step() return an error", NULL, -10);
			return 0;
			break;
		case NCE_SIMULATION_START:
			if(*command.length != sizeof( datum )){
				clean_pipe();
				return WARNING_MESSAGE("nce_simulation_command", "Erroneous format of NCE_SIMULATION_START command", NULL, 17);
			}
			if( web_read(engine_context.hInput, &d_value, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Input is broken!",NULL, 18 );
			if( (ret = sim_start(d_value)) >= 0 ){
				*command.command = NCE_RET_SIMULATION_IS_END;
				*command.length  = sizeof(int);
				if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
					return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 19 );
				if( web_write(engine_context.hOutput,&ret, sizeof(int))  != WEB_OK)
					return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 20 );
				return 0;
			}
			else
				return ERROR_MESSAGE("nce_simulation_command", "sim_start() returns an error",NULL, -11 );
			break;
		case NCE_SIMULATION_GO:
			if(*command.length != sizeof( int )){
				clean_pipe();
				return WARNING_MESSAGE("nce_simulation_command", "Erroneous format of NCE_SIMULATION_GO command", NULL, 21 );
			}
			if( web_read(engine_context.hInput, &i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Input is broken!",NULL, 22 );
			if( (ret = sim_go(i_value)) >= 0 ){
				*command.command = NCE_RET_SIMULATION_IS_END;
				*command.length  = sizeof(int);
				if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
					return WARNING_MESSAGE("nce_simulation_command", "Output is broken! ",NULL, 23 );
				if( web_write(engine_context.hOutput,&ret, sizeof(int)) != WEB_OK)
					return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 24 );
				return 0;
			}
			else
				return ERROR_MESSAGE("nce_simulation_command", "sim_go() returns an error",NULL, -12 );
			break;
		case NCE_SIMULATION_UNTIL:
			if(*command.length != sizeof( datum )){
				clean_pipe();
				return WARNING_MESSAGE("nce_simulation_command", "Erroneous format of NCE_SIMULATION_UNTIL command", NULL, 25 );
			}
			if( web_read(engine_context.hInput, &d_value, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken",NULL, 26 );
			if( (ret = sim_until(d_value)) >= 0 ){
				*command.command = NCE_RET_SIMULATION_IS_END;
				*command.length  = sizeof(int);
				if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE)  != WEB_OK)
					return WARNING_MESSAGE("nce_simulation_command", "Output is broken",NULL, 27 );
				if( web_write(engine_context.hOutput,&ret, sizeof(int))  != WEB_OK)
					return WARNING_MESSAGE("nce_simulation_command", "Output is broken",NULL, 28 );
				return 0;
			}
			else
				return ERROR_MESSAGE("nce_simulation_command", "sim_until() returns an error",NULL, -13 );
			break;
		case NCE_SIMULATION_DB_ITERATION:
			clean_pipe();
			return ( sim_db_iter() >= 0) ? 0 : ERROR_MESSAGE("nce_simulation_command", "sim_db_iter() returns an error",NULL, -14 );
			break;
		case NCE_SIMULATION_DB_MODULE_ITERATION:
			if(*command.length != sizeof( int )){
				clean_pipe();
				return WARNING_MESSAGE("nce_simulation_command", "Erroneous format of NCE_SIMULATION_DB_MODULE_ITERATION command", NULL, 29 );
			}
			if( web_read(engine_context.hInput, &i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Input is broken!",NULL, 30 );
			return (sim_db_mod_iter(i_value) >= 0 ) ? 0 : ERROR_MESSAGE("nce_simulation_command", "sim_db_mod_iter() returns an error",NULL, -15);
			break;
		case NCE_SIMULATION_DB_NEXT_MODULE_ITERATION:
			clean_pipe();
			return ( sim_db_next_mod_iter() >= 0) ? 0 : ERROR_MESSAGE("nce_simulation_command", "sim_db_next_mod_iter() returns an error",NULL, -16 );
			break;
		case NCE_SIMULATION_GET_BUFFDEEP:
			clean_pipe();
			*command.command = NCE_RET_SIMULATION_GET_BUFFDEEP;
			*command.length  = sizeof(int);
			if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE)  != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken",NULL, 31 );
			if( web_write(engine_context.hOutput,&nce_bufdeep, sizeof(int))  != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken",NULL, 32 );
			return 0;
			break;
		case NCE_SIMULATION_GET_MODEL_TIME:
			clean_pipe();
			*command.command = NCE_RET_SIMULATION_GET_MODEL_TIME;
			*command.length  = sizeof(datum);
			if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 33 );
			if( web_write(engine_context.hOutput,&nce_time_cur, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 34 );
			return 0;
			break;
		case NCE_SIMULATION_SET_MODEL_TIME:
			if(*command.length != sizeof( datum ) ){
				clean_pipe();
				return WARNING_MESSAGE("nce_simulation_command", "Erroneous format of NCE_SIMULATION_SET_MODEL_TIME command", NULL, 35 );
			}
			if( web_read(engine_context.hInput, &nce_time_cur, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Input is broken",NULL, 36 );
			watcher_restart();
			return 0;
			break;
		case NCE_SIMULATION_GET_MODEL_TIMESTEP:
			clean_pipe();
			*command.command = NCE_RET_SIMULATION_GET_MODEL_TIMESTEP;
			*command.length  = sizeof(datum);
			if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 37 );
			if( web_write(engine_context.hOutput,&nce_time_step, sizeof(datum)) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Output is broken!",NULL, 38 );
			return 0;
			break;
		case NCE_SIMULATION_SET_MODEL_TIMESTEP:
			if(*command.length != sizeof( datum ) ){
				clean_pipe();
				return WARNING_MESSAGE("nce_simulation_command", "Erroneous format of NCE_SIMULATION_SET_MODEL_TIMESTEP command", NULL, 39 );
			}
			if( web_read(engine_context.hInput, &nce_time_step, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_simulation_command", "Input is broken",NULL, 40 );
			watcher_restart();
			return 0;
			break;
	}
	return ERROR_MESSAGE("nce_simulation_command","Unknown command",NULL, -17 );
}

static int nce_watcher_command(command_header command, char*cmd){
	char *socket__file_name = NULL;
	char empty;
	datum d_value;
	int    i_value, ii_value, socket__file_id;
	struct sockaddr_un loc_sock;
	loc_sock.sun_family = AF_LOCAL;

	switch(*command.command){
		case NCE_WATCHER_ADD:
			if(*command.length != sizeof( datum )){
				clean_pipe();
				return WARNING_MESSAGE("nce_watcher_command", "Erroneous format of NCE_WATCHER_ADD command", NULL, 41);
			}
			if( web_read(engine_context.hInput, &d_value, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Input is broken!",NULL, 42 );
			if( (i_value = watcher_add(engine_context.hWatch, d_value)) < 0 )
				return WARNING_MESSAGE("nce_watcher_command", "Input is broken!",NULL, 43 );
			*command.command = NCE_RET_WATCHER_ADD;
			*command.length = sizeof(int);
			if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Output is broken!",NULL, 44 );
			if( web_write(engine_context.hOutput,&i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Output is broken!",NULL, 45 );
			return 0;
			break;				
		case NCE_WATCHER_ADD_LOCAL_SOCKET:
			if(*command.length <= sizeof( datum )){
				clean_pipe();
				return WARNING_MESSAGE("nce_watcher_command", "Erroneous format of NCE_WATCHER_ADD_FILE command", NULL, 46 );
			}
			if( web_read(engine_context.hInput, &d_value, sizeof(datum) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Input is broken!",NULL, 47 );
			if( (socket__file_name = (char*) mem_malloc (*command.length + 1 - sizeof(datum) ) ) == NULL)
				return ERROR_MESSAGE("nce","Cannot mem_malloc memory for socket name string",NULL, -18 );
			memset(socket__file_name, 0, *command.length + 1 - sizeof(datum));
			if( web_read(engine_context.hInput, socket__file_name, *command.length - sizeof(datum)) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Input is broken!",NULL, 48 );
			if( (socket__file_id = socket (PF_LOCAL, SOCK_STREAM, 0)) < 0 ){
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_watcher_command", "Cannot open local socket", NULL, 49);
			}
			strcpy(loc_sock.sun_path,socket__file_name);
			if ( connect(socket__file_id, (struct sockaddr*)&loc_sock, sizeof(loc_sock)) < 0){
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_watcher_command", "Cannot connect local socket", NULL, 50);
			}
			if( (i_value = watcher_add(socket__file_id, d_value)) < 0 ){
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_watcher_command", "nce_watcher_add() returns an error",NULL, 51 );
			}
			mem_free(socket__file_name);
			socket__file_name = NULL;
			*(command.command) = NCE_RET_WATCHER_ADD;
			*command.length = sizeof(int);
			if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Output is broken",NULL, 52 );
			if( web_write(engine_context.hOutput,&i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Output is broken",NULL, 53 );
			return 0;
			break;
		case NCE_WATCHER_ADD_FILE:
			if(*command.length <= sizeof( datum )){
				clean_pipe();
				return WARNING_MESSAGE("nce_watcher_command", "Erroneous format of NCE_WATCHER_ADD_FILE command", NULL, 54 );
			}
			if( web_read(engine_context.hInput, &d_value, sizeof(datum)) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Input is broken!",NULL, 55 );
			if( (socket__file_name = (char*) mem_malloc (*command.length + 1 - sizeof(datum) )) == NULL)
				return ERROR_MESSAGE("nce_watcher_command","Cannot mem_malloc memory for socket name string",NULL, -19);
			memset(socket__file_name, 0, *command.length + 1 - sizeof(datum));
			if( web_read(engine_context.hInput, socket__file_name, (*command.length) - sizeof(datum)) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Input is broken!",NULL, 56 );
			if( (socket__file_id =open(socket__file_name, O_CREAT | O_WRONLY, S_IWUSR | S_IRUSR | S_IWGRP | S_IRGRP)) < 0) {
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_watcher_command", "Cannot open file", NULL, 57);
			}
			if( (i_value = watcher_add(socket__file_id, d_value)) < 0 ){
				mem_free(socket__file_name);
				return WARNING_MESSAGE("nce_watcher_command", "nce_watcher_add() returns an error",NULL, 58 );
			}
			mem_free(socket__file_name);
			socket__file_name = NULL;
			*(command.command) = NCE_RET_WATCHER_ADD;
			*command.length = sizeof(int);
			if( web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Output is broken",NULL, 59 );
			if( web_write(engine_context.hOutput,&i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Output is broken",NULL, 60 );
			return 0;
			break;
		case NCE_WATCHER_SET_MODE:
			if(*command.length != (sizeof( int ) * 2)){
				clean_pipe();
				return WARNING_MESSAGE("nce_watcher_command", "Erroneous format of NCE_WATCHER_SET_MODE command", NULL, 61 );
			}
			if( web_read(engine_context.hInput, &i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Intput is broken",NULL, 62 );
			if( web_read(engine_context.hInput, &ii_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Intput is broken",NULL, 63 );
			if( watcher_set_mode(i_value, ii_value) < 0 )
				return WARNING_MESSAGE("nce_watcher_command", "Function watcher_set_mode() returns an error",NULL, 64 );
			return 0;
			break;
		case NCE_WATCHER_KILL:
			if(*command.length != sizeof( int )){
				clean_pipe();
				return WARNING_MESSAGE("nce_watcher_command", "Erroneous format of NCE_WATCHER_KILL command", NULL, 65 );
			}
			if( web_read(engine_context.hInput, &i_value, sizeof(int) ) != WEB_OK)
				return WARNING_MESSAGE("nce_watcher_command", "Intput is broken",NULL, 66 );
			if( watcher_kill(i_value) < 0 )
				return WARNING_MESSAGE("nce_watcher_command", "Function watcher_kill() returns an error",NULL, 67 );
			return 0;
			break;
		case NCE_WATCHER_KILL_ALL:
			clean_pipe();
			if( watcher_killall() < 0)
				return WARNING_MESSAGE("nce_watcher_command", "Function watcher_killall() returns an error",NULL, 68 );
			return 0;
			break;
	}
	return ERROR_MESSAGE("nce_watcher_command","Unknown command",NULL, -20 );
}

int nce(nce_context *context)
{
	engine_context = *context;
	static char cmd[COMMAND_CONTEXT_SIZE];
	command_header command;
	command.command = cmd;
	command.length  = (int*)(cmd+1);
	
	char empty;
	
	//Engine signals
	struct sigaction stop_signal, simulation_status;
	//Stop signal
	memset(&stop_signal, 0, sizeof(stop_signal));
	stop_signal.sa_handler = &nce_stop_simulation_handler;
	sigaction(NCE_STOP_SIMULATION, &stop_signal, NULL);
	//Get simulation status
	memset(&simulation_status, 0, sizeof(simulation_status) );
	simulation_status.sa_handler = &nce_get_simulation_status;
	sigaction(NCE_GET_SIMULATION_STATUS,&simulation_status, NULL);
	
	while(work_flag){
		if (web_read(engine_context.hInput, cmd, COMMAND_CONTEXT_SIZE) == WEB_ERR) break;
		switch(*command.command){
			case NCE_SYSTEM_GET_ID:
				clean_pipe();
				*(command.command) = NCE_RET_SYSTEM_GET_ID;
				*(command.length)  = sizeof (int);
				if(web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE) != WEB_OK){
					WARNING_MESSAGE("nce", "Output is broken",NULL, -2);
					break;
				}
				int pid = engine_context.id;
				if( web_write(engine_context.hOutput,&pid, sizeof(int)) != WEB_OK){
					WARNING_MESSAGE("nce", "output is broken",NULL, -2);
					break;
				}
				send_ok();
				break;
			case NCE_SYSTEM_QUIT:
				sim_flag = 0;
				work_flag = 0;
				break;
			case NCE_ASSEMBLER_LOAD:
			case NCE_ASSEMBLER_COMMAND:
			case NCE_ASSEMBLER_KILL:
				if( nce_assembler_command(command) == 0) send_ok();
				break;

			case NCE_SIMUALTION_STEP:
			case NCE_SIMULATION_START:
			case NCE_SIMULATION_GO:
			case NCE_SIMULATION_UNTIL:
			case NCE_SIMULATION_DB_ITERATION:
			case NCE_SIMULATION_DB_MODULE_ITERATION:
			case NCE_SIMULATION_DB_NEXT_MODULE_ITERATION:
			case NCE_SIMULATION_GET_BUFFDEEP:
			case NCE_SIMULATION_GET_MODEL_TIME:
			case NCE_SIMULATION_SET_MODEL_TIME:
			case NCE_SIMULATION_GET_MODEL_TIMESTEP:
			case NCE_SIMULATION_SET_MODEL_TIMESTEP:
				if( nce_simulation_command(command, cmd) == 0) send_ok();
				break;

			case NCE_WATCHER_ADD:
			case NCE_WATCHER_ADD_LOCAL_SOCKET:
			case NCE_WATCHER_ADD_FILE:
			case NCE_WATCHER_SET_MODE:
			case NCE_WATCHER_KILL:
			case NCE_WATCHER_KILL_ALL:
				if( nce_watcher_command(command, cmd) == 0) send_ok();
				break;
				
			case NCE_PRINT_INFO_ASM:
				clean_pipe();
				if( asm_printinfo_pointer != NULL){
					(*asm_printinfo_pointer)();
					send_ok();
				}
				else
					WARNING_MESSAGE("nce", "Assembler isn't load",NULL, 69 );
				break;
			case NCE_PRINT_INFO_MODULES:
				clean_pipe();
				mod_printinfo();
				send_ok();
				break;
			case NCE_PRINT_INFO_SIMULATOR:
				clean_pipe();
				sim_printinfo();
				send_ok();
				break;
			case NCE_PRINT_INFO_MEMORY:
				clean_pipe();
				mem_printinfo();
				send_ok();
				break;
			case NCE_PRINT_INFO_WATCHER:
				clean_pipe();
				watcher_printinfo();
				send_ok();
				break;
			default:
				clean_pipe();
				WARNING_MESSAGE("nce", "Unknown command",NULL, 70);
				break;
		}
	}
	if(asm_drop() < 0) return ERROR_MESSAGE("nce", "Cannot destroy model", NULL, -21);
	memset(cmd, 0, COMMAND_CONTEXT_SIZE);
	cmd[0] = NCE_RET_DEAD;
	web_write(engine_context.hOutput,cmd, COMMAND_CONTEXT_SIZE);
	return 0;
}
