#ifndef LIBNCE_H
#define LIBNCE_H

typedef struct {
	int hInput;		//Handle for input binary stream,
				//used for send commands, models and ...
	int hOutput;		//Output stream, returns commands results, 
				//messages and debugs. For all types of text
				//streams heading by indentifer NC__RET_XXX.
	int hWatch;		//Binary stream of watchered data
	size_t memsize_all;
	size_t memsize_allocated;
	size_t memsize_free;
	pid_t  id;		//Engine identifier
} nce_context;


typedef struct {
	char *command;
	int *length;
} command_header;

typedef double datum;

#define COMMAND_CONTEXT_SIZE (sizeof(char) + sizeof(int))

//Signals for stoping, .... simulation 
#ifndef NCE_STOP_SIMULATION
#	define NCE_STOP_SIMULATION SIGUSR1
#endif
#ifndef NCE_GET_SIMULATION_STATUS
#	define NCE_GET_SIMULATION_STATUS SIGUSR2
#endif


enum engine_returns{
	NC__RET_OK = 0,
	NC__RET_MESSAGE,
	NC__RET_DEBUG,
	NC__RET_WARNING,
	NC__RET_ERROR,
	NCE_RET_DEAD,
	NCE_RET_SYSTEM_GET_ID,
	NCE_RET_SIMULATION_IS_END,
	NCE_RET_SIMULATION_GET_BUFFDEEP,
	NCE_RET_SIMULATION_GET_MODEL_TIME,
	NCE_RET_SIMULATION_GET_MODEL_TIMESTEP,
	NCE_RET_WATCHER_ADD,
	NCE_RET_SEPARATOR
};

enum engine_commands {
//System commands
	NCE_SYSTEM_GET_ID = 0,
	NCE_SYSTEM_QUIT,
//Assembler functions
	NCE_ASSEMBLER_LOAD,
	NCE_ASSEMBLER_COMMAND,
	NCE_ASSEMBLER_KILL,
//Simulation commands
	NCE_SIMUALTION_STEP,
	NCE_SIMULATION_START,
	NCE_SIMULATION_GO,
	NCE_SIMULATION_UNTIL,
	NCE_SIMULATION_DB_ITERATION,
	NCE_SIMULATION_DB_MODULE_ITERATION,
	NCE_SIMULATION_DB_NEXT_MODULE_ITERATION,
	NCE_SIMULATION_GET_BUFFDEEP,
	NCE_SIMULATION_GET_MODEL_TIME,
	NCE_SIMULATION_SET_MODEL_TIME,
	NCE_SIMULATION_GET_MODEL_TIMESTEP,
	NCE_SIMULATION_SET_MODEL_TIMESTEP,
//Data Watcher (stream data ....) commands
	NCE_WATCHER_ADD,
	NCE_WATCHER_ADD_LOCAL_SOCKET,
	NCE_WATCHER_ADD_FILE,
	NCE_WATCHER_SET_MODE,
	NCE_WATCHER_KILL,
	NCE_WATCHER_KILL_ALL,
//Commands for information and engine testing
	NCE_PRINT_INFO,
	NCE_PRINT_INFO_ASM,
	NCE_PRINT_INFO_MODULES,
	NCE_PRINT_INFO_SIMULATOR,
	NCE_PRINT_INFO_MEMORY,
	NCE_PRINT_INFO_WATCHER,
//Set the enum of server commands or upper programme in
	NCE_SEPARATOR
};

//Mode of watcher
#define WATCHER_MODE_PASSIVE	0x0000
#define WATCHER_MODE_ACTIVE	0x0001
//Mode of watcher item
#define WATCHER_ITEM_LAST	0x0000
#define WATCHER_ITEM_MINIMAL	0x0001
#define WATCHER_ITEM_MAXIMAL	0x0002
#define WATCHER_ITEM_AVERAGE	0x0003

#ifdef  __cplusplus
	extern "C" { int nce(nce_context *context); }
#else
	int nce(nce_context *context);
#endif

#endif
