#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <neurocad/engine/memmanager.h>

typedef struct pro_mem_g_header {
	size_t size;				// The total size  of memory in granule
	char *root, *curr;			//root and current allocated memory pointers
	struct pro_mem_g_header *next, *prev;	//linked list pointers
} mem_g_header, *mghp;

static mghp mem_root = NULL;

//Local function for finding granule. Looking for handle pointer in the
//granular memory linked list
static mghp find_granule(mghp handle)
{
#ifndef FAST_NOACCURACY_FLAG
	return handle;
#else
	mghp scaner = mem_root;
	for(; scner != NULL; scaner = scaner->prev)
		if( scaner == handle ) return handle;
	return scaner;
#endif
}

void *mem_malloc(size_t size)
{
	if(size > engine_context.memsize_free) return NULL;
	void *ptr = malloc (size + sizeof(size_t));
	if(ptr == NULL) return ptr;
	engine_context.memsize_allocated += size + sizeof(size_t);
	engine_context.memsize_free -= size + sizeof(size_t);
	*((size_t*)ptr) = size;
	return (void*)( (char*)ptr + sizeof(size_t));
}

void  mem_free(void *ptr)
{
	if(ptr == NULL) return;
	ptr = (void*)( (char*)ptr - sizeof(size_t));
	size_t size = *((size_t *)ptr);
	engine_context.memsize_allocated -= size + sizeof(size_t);
	engine_context.memsize_free += size + sizeof(size_t);
	free(ptr);
}

void	*mem_g_init (size_t allsize)
{
	char fnarg[20]; memset(fnarg,0,20); snprintf(fnarg,19,"%ld", (long int) allsize);
	mghp new_mem_header = (mghp) mem_malloc ( sizeof(mem_g_header) + allsize + 1 );
	if(new_mem_header  == NULL )
		return (void*)ERROR_MESSAGE("mem_g_init", "Cannot malloc memory", fnarg, (int)NULL);
	memset(new_mem_header, 0, sizeof(mem_g_header) + allsize +1 );
	new_mem_header->size = allsize + 1;
	new_mem_header->root = new_mem_header->curr = ((char*)new_mem_header) + sizeof(mem_g_header);
	new_mem_header->prev = mem_root;
	new_mem_header->next = NULL;
	if( mem_root != NULL )mem_root->next = new_mem_header;
	mem_root = new_mem_header;
	return (void*)mem_root;
}

void	*mem_g_malloc(void *handle, size_t size)
{
	char fnarg[20]; memset(fnarg,0,20); snprintf(fnarg,19,"%ld, %ld",(long int)handle, (long int) size);
	mghp mem_handle = (mghp)handle;
	if (find_granule( mem_handle ) == NULL )
		return (void*) ERROR_MESSAGE("mem_g_malloc", "Memory does not init or release",fnarg, (int)NULL);
	if ( mem_handle->size <= ( (size_t)(mem_handle->curr - mem_handle->root) + size ) )
		return (void*) ERROR_MESSAGE("mem_g_malloc", "The required memory size is out of initiated memory size",fnarg, (int)NULL);
	void *result = (void*)mem_handle->curr;
	mem_handle->curr += size;
	return result;
}

int	 mem_g_release(void *handle)
{
	char fnarg[20]; memset(fnarg,0,20); snprintf(fnarg,19,"%ld",(long int)handle);
	mghp mem_handle = (mghp)handle;
	if (find_granule( mem_handle ) == NULL )
		return WARNING_MESSAGE("mem_g_release", "Memory does not init or already release",fnarg, -1);
	if( mem_handle->next != NULL ) mem_handle->next->prev = mem_handle->prev;
	if( mem_handle->prev != NULL ) mem_handle->prev->next = mem_handle->next;
	if( mem_handle == mem_root ) mem_root = mem_handle->prev;
	mem_free(handle);
	return 0;
}

void	*mem_g_resize(void *handle, size_t newsize)
{
	char fnarg[20]; memset(fnarg,0,20); snprintf(fnarg,19,"%ld, %ld",(long int)handle, (long int) newsize);
	mghp old_mem_header = (mghp)handle;
	if (find_granule( old_mem_header ) == NULL )
		return (void*) ERROR_MESSAGE("mem_g_resize", "Memory does not init or release",fnarg, (int)NULL);
	mghp new_mem_header = (mghp) mem_malloc ( sizeof(mem_g_header) + newsize );
	if(new_mem_header  == NULL )
		return (void*)ERROR_MESSAGE("mem_g_resize", "Cannot malloc memory", fnarg, (int)NULL);
	memset(new_mem_header, 0, sizeof(mem_g_header) + newsize );
	new_mem_header->size = newsize;
	new_mem_header->root = ((char*)new_mem_header) + sizeof(mem_g_header);
	if( old_mem_header->size < newsize){
		memcpy(new_mem_header->root, old_mem_header->root, old_mem_header->size);
		new_mem_header->curr = new_mem_header->root + 
			(size_t)(old_mem_header->curr  - old_mem_header->root);
	}
	else {
		memcpy(new_mem_header->root, old_mem_header->root, newsize);
		new_mem_header->curr =
			( (size_t)(old_mem_header->curr  - old_mem_header->root) > newsize ) ? 
			new_mem_header->root + newsize : 
			new_mem_header->root + (size_t)(old_mem_header->curr  - old_mem_header->root);
	}
		
	new_mem_header->prev = old_mem_header->prev;
	new_mem_header->next = old_mem_header->next;
	if(old_mem_header == mem_root) mem_root = new_mem_header;
	if(mem_g_release(old_mem_header) != 0) 
		WARNING_MESSAGE("mem_g_resize", "Memory is not release",fnarg, 0);
	return (void*)new_mem_header;
}

size_t	 mem_g_getfreesize(void *handle)
{
	char fnarg[20]; memset(fnarg,0,20); snprintf(fnarg,19,"%ld",(long int)handle);
	mghp mem_handle = (mghp)handle;
	if (find_granule( mem_handle ) == NULL )
		return (size_t) WARNING_MESSAGE("mem_g_getfreesize", "Memory does not init or release",fnarg, 0);
	return mem_handle->size - (size_t)(mem_handle->curr - mem_handle->root);
}


int  mem_printinfo()
{
	OUTPUT_MESSAGE("=== Memory manager information ===\n");
	OUTPUT_MESSAGE("\t Total memory:             %lld[%lld kb, %lld Mb]\n",
		       (unsigned long long int) engine_context.memsize_all, 
		       (unsigned long long int) engine_context.memsize_all/1024,
		       (unsigned long long int) engine_context.memsize_all/1024/1024);
	OUTPUT_MESSAGE("\t Total allocated memory:   %lld[%lld kb, %lld Mb]\n",
		       (unsigned long long int)engine_context.memsize_allocated,
		       (unsigned long long int)engine_context.memsize_allocated/1024,
		       (unsigned long long int)engine_context.memsize_allocated/1024/1024);
	OUTPUT_MESSAGE("\t Total free memory:        %lld[%lld kb, %lld Mb]\n",
		       (unsigned long long int)engine_context.memsize_free,
		       (unsigned long long int)engine_context.memsize_free/1024,
		       (unsigned long long int)engine_context.memsize_free/1024/1024);
	OUTPUT_MESSAGE("\t=== Granule list ===\n");
	mghp mem_handle = mem_root;
	size_t granuleaccount = (size_t)0;
	for(; mem_handle != NULL; mem_handle = mem_handle->prev){
		size_t alloc = (size_t)(mem_handle->curr - mem_handle->root);
		OUTPUT_MESSAGE("\t	Allocated:           %lld[%lld kb, %lld Mb]\n",
			       (unsigned long long int)mem_handle->size,
			       (unsigned long long int)mem_handle->size/1024,
			       (unsigned long long int)mem_handle->size/1024/1024);
		OUTPUT_MESSAGE("\t	Used:                %lld[%lld kb, %lld Mb]\n",
			       (unsigned long long int)alloc,
			       (unsigned long long int)alloc/1024,
			       (unsigned long long int)alloc/1024/1024);
		granuleaccount += mem_handle->size + sizeof( mem_g_header );
	}
	granuleaccount = engine_context.memsize_allocated - granuleaccount;
	OUTPUT_MESSAGE("\t Allocated ungranular:     %lld[%lld kb, %lld Mb]\n",
		       (unsigned long long int)granuleaccount,
		       (unsigned long long int)granuleaccount/1024,
		       (unsigned long long int)granuleaccount/1024/1024);
	return 0;
}

void	mem_g_dbdump(void *handle)
{
	char fnarg[20]; memset(fnarg,0,20); snprintf(fnarg,19,"%ld",(long int)handle);
	mghp mem_handle = (mghp)handle;
	if (find_granule( mem_handle ) == NULL ) {
		WARNING_MESSAGE("mem_g_dbdump", "Memory does not init or release",fnarg, 0);
		return;
	}
	size_t alloc = (size_t)(mem_handle->curr - mem_handle->root), counter;
	OUTPUT_MESSAGE("DUMP of %lld memory granule\n", (unsigned long long int)handle);
	OUTPUT_MESSAGE("   Allocated:           %lld[%lld kb, %lld Mb]\n",
		       (unsigned long long int)mem_handle->size,
		       (unsigned long long int)mem_handle->size/1024,
		       (unsigned long long int)mem_handle->size/1024/1024);
	OUTPUT_MESSAGE("   Used:                %lld[%lld kb, %lld Mb]\n",
		       (unsigned long long int)alloc,
		       (unsigned long long int)alloc/1024,
		       (unsigned long long int)alloc/1024/1024);
	OUTPUT_MESSAGE("DUMP:");
	for(counter = 0; counter < mem_handle->size; counter++){
		if( (counter % 0x10) == 0) OUTPUT_MESSAGE("\n%06lX:",(long int)counter);
		OUTPUT_MESSAGE(" %02X:",mem_handle->root[counter]);
	}
}

