#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <neurocad/engine/modmanager.h>
#include <neurocad/engine/nce_common.h>

static mip root = NULL;
static mip curr = NULL;
int mod_number = 0;


inline static int find_modulele(int imodid)
{
	char errmessage[10]; memset(errmessage, 0,10); snprintf(errmessage,9,"%d",imodid);
	if(root == NULL) return ERROR_MESSAGE("find_modulele","List of Modules is empty root = NULL",errmessage, -1);
	if((curr == NULL) || (curr->mid > imodid))curr = root;
	while((curr != NULL) && (curr->mid != imodid)) curr = curr->next;
	if(curr == NULL) return ERROR_MESSAGE("find_modulele","Can not found  module",errmessage, -2);
	return 0;
}

inline static void free_module(mip dell)
{
	if( dell->release != NULL ){
		(*dell->release)();
		dlclose(dell->hmod);
	}
	if( dell->filename != NULL )
		mem_free(dell->filename);
	mem_free(dell);
}


int mod_insert(char *mod_file_name, int mid)
{
	char errmessage[512]; memset(errmessage,0,512);
	snprintf(errmessage,200,"%s, %d",mod_file_name,mid);
	char dlmessage[512]; memset(dlmessage,0,512);
	int *intpointer, cnt; datum *datumpointer;

	mip newitem = (mip)mem_malloc(sizeof(moditem));
	if(newitem == NULL)
		return ERROR_MESSAGE("mod_insert","Can not malloc memory for New Module Item",errmessage,-1);
	memset(newitem, sizeof(moditem), 0);
	if( (newitem->hmod = dlopen(mod_file_name, RTLD_LAZY)) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot dlopen file! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-2);
	}
	enum ml *pmodmultiload = (enum ml*)dlsym(newitem->hmod,"modmulti");
	if( pmodmultiload == NULL ){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modmultiload! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	else	newitem->multiload = *pmodmultiload;
	if( (intpointer = dlsym(newitem->hmod,"moditer")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read moditer! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	else	newitem->iterations = *intpointer;
	if( (intpointer = dlsym(newitem->hmod,"modprior")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modprior! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	else	newitem->priority = *intpointer;
	if( (intpointer = dlsym(newitem->hmod,"modallnames")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modallnames! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	else	newitem->allnames = *intpointer;
	if( (intpointer = dlsym(newitem->hmod,"modblocksize")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modblocksize! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	else	newitem->blocksize = *intpointer;
	if( (datumpointer = dlsym(newitem->hmod,"modweight")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modweight! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	else	newitem->weight = *datumpointer;
	if( (newitem->names = (nameitem *)dlsym(newitem->hmod,"modnames")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modnames! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->names = (nameitem *)dlsym(newitem->hmod,"modnames")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot read modnames! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->init = (int (*) (int)) dlsym(newitem->hmod,"init")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'init\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->release = (int (*) ()) dlsym(newitem->hmod,"release")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'release\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->add = (module * (*) (int, module *, void*)) dlsym(newitem->hmod,"add")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'add\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->calculate = (int(*)(int)) dlsym(newitem->hmod,"calculate")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'calculate\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->slack = (int(*)(int)) dlsym(newitem->hmod,"slack")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'slack\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->restore = (int(*)(int,int)) dlsym(newitem->hmod,"restore")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'restore\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	if( (newitem->save = (int(*)(int,int)) dlsym(newitem->hmod,"save")) == NULL){
		free_module(newitem);
		snprintf(dlmessage, 511, "Cannot find function \'save\'! `dlerror\' return: %s",dlerror());
		return ERROR_MESSAGE("mod_insert",dlmessage,errmessage,-3);
	}
	newitem->jump = (int(*)(int))dlsym(newitem->hmod,"jump");

	mip scan = root;
	for(; (scan != NULL) && ( strcmp(scan->filename,mod_file_name) ); scan = scan->next);
	if( (scan != NULL) && ( scan->multiload != multi ) ){
		free_module(newitem);
		return ERROR_MESSAGE("mod_insert","Loaded module cannot support multiloder mode",errmessage,-4);
	}

	newitem->mid = mid;
	newitem->overstep = 0;
	newitem->isused = 0;
	newitem->next = NULL;

	for(cnt=0; cnt < newitem->allnames; cnt++){
		if(newitem->names[cnt].name != NULL) continue;
		free_module(newitem);
		return ERROR_MESSAGE("mod_insert","Number of names more then real names in module",errmessage,-5);
	}
	if(root == NULL){
		root = newitem;
 		++mod_number;
		return 0;
	}
	if(root->mid > mid){
		newitem->next = root;
		root = newitem;
 		++mod_number;
		return 0;
	}
	if(root->mid == mid) return ERROR_MESSAGE("mod_insert","Module with such id has been made",errmessage,-6);
	for(curr = root; (curr != NULL) && (curr->next != NULL) && (curr->next->mid < mid); curr=curr->next);
	if(curr->next == NULL){
		curr->next = newitem;
		curr->next->next = NULL;
 		++mod_number;
		return 0;
	}
	if(curr->next->mid == mid) return ERROR_MESSAGE("mod_insert","Module with such id has been made",errmessage,-6);
	newitem->next = curr->next;
	curr->next = newitem;
	++mod_number;
	return 0;
}

int mod_insert_empty(int mid)
{
	char errmessage[512]; memset(errmessage,0,512);
	snprintf(errmessage,200,"%d",mid);
	mip newitem = (mip)mem_malloc(sizeof(moditem));
	if(newitem == NULL)
		return ERROR_MESSAGE("mod_insert_empty","Can not malloc memory for New Module Item",errmessage,-1);
	newitem->mid = mid;
	newitem->filename = NULL;
	newitem->isused = newitem->overstep = 0;
	newitem->multiload = 0;
	newitem->iterations = 0;
	newitem->allnames = 0;
	newitem->priority = 0;
	newitem->blocksize = 0;
	newitem->weight = 0.0;

	newitem->init = NULL;
	newitem->release = NULL;
	newitem->add = NULL;
	newitem->calculate = NULL;
	newitem->slack = NULL;
	newitem->jump = NULL;
	newitem->restore = NULL;
	newitem->save = NULL;
	newitem->hmod = NULL;
	newitem->names = NULL;
	newitem->next = NULL;

	if(root == NULL){
		root = newitem;
 		++mod_number;
		return 0;
	}
	
	if(root->mid > mid){
		newitem->next = root;
		root = newitem;
 		++mod_number;
		return 0;
	}
	
	if(root->mid == mid) return ERROR_MESSAGE("mod_insert_empty","Module with such id has been made",errmessage,-2);
	
	for(curr = root; (curr != NULL) && (curr->next != NULL) && (curr->next->mid < mid); curr=curr->next);
	
	if(curr->next == NULL){
		curr->next = newitem;
		curr->next->next = NULL;
 		++mod_number;
		return 0;
	}

	if(curr->next->mid == mid) return ERROR_MESSAGE("mod_insert_empty","Module with such id has been made",errmessage,-3);
	
	newitem->next = curr->next;
	curr->next = newitem;
	++mod_number;
	return 0;

}

int mod_remove(int mid)
{
	char errmessage[201]; memset(errmessage,0,201);
	mip cur = root, dell = root;
	if(root->mid == mid){
		root = root->next;
		free_module(dell);
		 --mod_number;
		return 0;
	}
	while((cur->next != NULL)&&(cur->next->mid != mid)) cur = cur->next;
	if(cur->next == NULL){
		snprintf(errmessage,200,"%d",mid);
		return  ERROR_MESSAGE("mod_remove","Can not find mid",errmessage,-1);
	}
	dell = cur->next;
	cur->next = cur->next->next;
	free_module(dell);
	--mod_number;
	return 0;
}

int mod_removeall()
{
	if(root == NULL){
		WARNING_MESSAGE("mod_removeall","List of Modules is empty root = NULL",NULL,-1);
		return 0;
	}
	while(root != NULL){
		curr = root;
		root = root->next;
		free_module(curr);
	}
	return (mod_number = 0);

}

int mod_get_number()
{
	return (root == NULL) ? 0 :  mod_number;
}

mip mod_get_item(int mid)
{
	char errmessage[10];snprintf(errmessage,10,"%d",mid);
	mip cur = root;
	while((cur != NULL) && (cur->mid != mid))cur = cur->next;
	if(cur == NULL) return (mip)ERROR_MESSAGE("mod_get_item","Can not found  module",errmessage,(int)NULL);
	return cur;
}

mip mod_get_list()
{
	return root;
}

int mod_init(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) != 0)
		return ERROR_MESSAGE("mod_init","find_modulele() return error",errmessage,-1);
	if( (curr->add == NULL) || ( curr->calculate == NULL ) || ( curr->slack == NULL) ||
		( curr->restore == NULL ) || ( curr->save == NULL ) )
		return ERROR_MESSAGE("mod_init","Cannot find functions: add, calculate, slack restore and save",errmessage,2);
	if( curr->isused != 0) return 0;
	if( curr->init != NULL){
		int modinitret = (*curr->init)(curr->mid);
		if( modinitret < 0){
			char errmessageInint[100];snprintf(errmessageInint,100,"init() return error %d",modinitret);
			return ERROR_MESSAGE("mod_init",errmessageInint,errmessage,-3);
		}
	}
	curr->isused = 1;
	return 0;
}

int mod_is_acctive(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) != 0) 
		return ERROR_MESSAGE("mod_is_acctive","find_modulele() return error",errmessage,-1);
	return curr->isused;
}

int mod_set_extended_attribute(int mid, char *name, char *value)
{
	char errmessage[201]; memset(errmessage,0,201); snprintf(errmessage,200,"%d, %s, %s",mid,name,value);
	if(find_modulele(mid) != 0)
		return ERROR_MESSAGE("mod_set_extended_attribute","find_modulele() return error",errmessage,-1);
	if(curr->hmod == NULL)
		return ERROR_MESSAGE("mod_set_extended_attribute","hmod is NULL",errmessage,-2);
	char **pAttr = dlsym(curr->hmod,name);
	if(pAttr == NULL){
		char cError[200];
		snprintf(cError,199,"Cannot found attribute \"%s\" in <module id = %d>",name,mid);
		return ERROR_MESSAGE("mod_set_extended_attribute",cError,errmessage,-3);
	}
	if( (*pAttr = bufdup(NULL, value)) == NULL )
		return ERROR_MESSAGE("mod_set_extended_attribute","Cannot set extended attribute",errmessage,-4);
	return 0;
}


int mod_flash_extended_attribute(int mid)
{
	char errmessage[10]; memset(errmessage,0,10); snprintf(errmessage,9,"%d",mid);
	if(find_modulele(mid) < 0) 
		return ERROR_MESSAGE("mod_flash_extended_attribute","find_modulele() return error",errmessage, -1);
	if(curr->hmod == NULL)
		return ERROR_MESSAGE("mod_flash_extended_attribute","hmod is NULL",errmessage, -2);
	int (*flash)(int)= (int(*)(int))dlsym(curr->hmod,"flash");
	if(flash == NULL)
		return ERROR_MESSAGE("mod_flash_extended_attribute","Cannot found flash function in module",errmessage, -3);
	int flashret = (*flash)(mid);
	if( flashret != 0){
		char cError[200]; snprintf(cError,199,"Flash function return error: %d",flashret);
		return ERROR_MESSAGE("mod_flash_extended_attribute",cError,errmessage, -4);
	}
	return 0;
}


int mod_printinfo()
{
	OUTPUT_MESSAGE("\n=== Modules information ===\n");
	if(root == NULL){
		OUTPUT_MESSAGE("List of Modules is empty\n\n");
		return 0;
	}
	else
		OUTPUT_MESSAGE("\n");
	OUTPUT_MESSAGE("\n=== Modules Table ===\nmid\t\titerations\tnames\t\tpriority\toverstep\tfile\n");
	OUTPUT_MESSAGE("------------------------------------------------------------------------------------------------------\n");
	for(curr = root ; curr != NULL ; curr = curr->next)
		OUTPUT_MESSAGE("%d\t\t%d\t\t%d\t\t%d\t\t%d\t\t%s\n",
			       curr->mid,curr->iterations,curr->allnames,curr->priority,curr->overstep,curr->filename);
	OUTPUT_MESSAGE("\n");
	return 0;
}
