#ifndef __MODULE_H_
#define __MODULE_H_ 

#include <math.h>
#include <neurocad/engine/memmanager.h>
#include <neurocad/engine/nce_common.h>

typedef union pro_dptr {
	datum *dp;
	datum **dpp;
	void   *vp;
} dptr;

/**main module varibales**/
typedef struct pro_module {
	unsigned long oid;	//object, module and block id
	int  mid, bid;
	struct pro_module *next;	//using for common tree
	dptr   names[];		//set of names - parameters and variables
} module;

/**
This main makros must be included in all module structured data (block).
**/
#define MODULE		unsigned long oid; \
			int mid, bid; \
			module  *next;
			

/**Module multiloading**/
enum ml{
	single = 0,	//module can be inserted in model only one time
	multi		//multiloding module
};

/**Name mode**/
enum mn{
	rd = 0,		//the name only is read by modules
	wr,		//only is write by modules
	rdwr,		//bidirectal name
	tbl,		//the name is table (only for reading)
	rb		//ringing buffer
};

/**Define default size of name**/
#define alone 1

typedef struct {
	char		*name;
	enum mn		mode;
	int		size;
} nameitem;

/**Common function for all modules**/
int	init( int imid );
int	release( int imid );
module	*add( int imid, module *next, void *memory_handle);
int	calculate( int imid );
int	slack( int imid );
int	restore( int imid, int hfile);
int	save( int imid, int hfile );
/**optional functions**/
int	jump( int imid );
int	flash( int imid);

/**
Switch and antiswitch for global synchronization
Those variables have definition in the simmanager.c of NeuroCAD Engine
For each named variable Engine must allocated ncencence_bufdeep datum set. The ncencencence_sw (switch) set the point 
of current step data the ncencencence_asw points on the next step data (next step in future).
**/
extern int nce_sw, nce_asw, nce_bufdeep;

/**
Global timestep and timecurrent
Those variables have definition in the simmanager.c of NeuroCAD Engine
**/
extern double nce_time_step, nce_time_cur;

/**
Macroses for gatting value of switch and antiswitch variables
**/
#ifndef GN_SW
#define		GN_SW(NAME) (currentblk->names[NAME].dp[nce_sw])
#endif //GN_SW
#ifndef GNASW
#define 	GNASW(NAME) (currentblk->names[NAME].dp[nce_asw])
#endif //GNASW

#endif

