#ifndef NCAD_COMMON_H
#define NCAD_COMMON_H

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <signal.h>
#include <neurocad/engine/libnce.h>

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#ifndef RDWR_BUFFER_SIZE
#define 	RDWR_BUFFER_SIZE		4096
#endif
#ifndef MEMPAGE_SIZE
#define 	MEMPAGE_SIZE			4096
#endif
#ifndef MAX_MESSAGE_LENGTH
#define 	MAX_MESSAGE_LENGTH		512
#endif


/**
engine_context is structure for engine environment (see libnce.h)
**/
extern nce_context engine_context;

/**
The work_flag is flag of nce main loop. If work_flag
is set at zero the engine will be deaded.
**/
extern int work_flag;

/**
bufdup() clears the buff, if it isn't NULL, and allocates new memory and copies str there.
Return the pointer of new memory or NULL
 **/
extern char *bufdup( char * buff, char *str);

/**
web_write() and web_read() eliminate the problem of no full reading and writing. If read() or 
write() utilize less then SIZE the web_write() and web_read() will try to complete result.
functions return constant WEB_OK if transaction is successful or WEB_ERR in anouther case.
 **/
#define	WEB_OK	1
#define WEB_ERR	0
extern ssize_t web_write (int FILEDES, const void *BUFFER, size_t SIZE);
extern ssize_t web_read  (int FILEDES, const void *BUFFER, size_t SIZE);

/**
The set of function sending the messages text though binary pipe.
 **/
extern inline int send_error(int ret, char* message);
extern inline int send_warning(int ret, char* message);
extern inline void send_ok();
extern inline int send_message(int ret, char* message);
extern inline int send_debug(int ret, char* message);

/** Functions for TEXT printing out messages system**/
extern char sendmessage[MAX_MESSAGE_LENGTH];
extern int ErrorMessage(const char *pFileName, const int Line, const char *pFuncsionName, const char *pErrorMessage, char *pParam, int ret);
extern int WarningMessage(const char *pFileName, const int Line, const char *pFuncsionName, const char *pErrorMessage, char *pParam, int ret);

/** Functions for DEBUG printing out messages system**/
extern void ncedb_info(const char *modname, const char *modinfo);
extern void ncedb_double(const char *modname, const char *modvar, double val);
extern void ncedb_int(const char *modname, const char *modvar, int val);


/** Makroses for TEXT and DEBUG printing out messages system**/
#ifndef SILENCE
#define 	ERROR_MESSAGE(Func,Message,Param,Ret)	ErrorMessage(__FILE__,__LINE__,Func,Message,Param,Ret)
#define 	WARNING_MESSAGE(Func,Message,Param,Ret)	WarningMessage(__FILE__,__LINE__,Func,Message,Param,Ret)
#define 	OUTPUT_MESSAGE(...) 			send_message(snprintf(sendmessage, MAX_MESSAGE_LENGTH ,__VA_ARGS__), sendmessage)
#define 	DB_INFO(MOD, INFO) 			ncedb_info(MOD, INFO)
#define 	DB_DOUBLE(MOD, VAR, VAL)		ncedb_double(MOD,  VAR, VAL)
#define 	DB_INT(MOD, VAR, VAL)			ncedb_int(MOD,  VAR, VAL)
#else
#define 	ERROR_MESSAGE(Func,Message,Param,Ret)	Ret
#define 	WARNING_MESSAGE(Func,Message,Param,Ret) Ret
#define 	OUTPUT_MESSAGE(...)
#define 	DB_INFO(MOD, INFO)
#define 	DB_DOUBLE(MOD, VAR, VAL)
#define 	DB_INT(MOD, VAR, VAL)
#endif

#endif
