#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/resource.h>
#include <neurocad/engine/nce_common.h>
#include <neurocad/engine/simmanager.h>
#include <neurocad/engine/memmanager.h>
#include <neurocad/engine/asmmanager.h>

typedef struct pro_sch_iteration{
	int (*fn)(int);		//function for calculate, noload og jump
	int result;		//expected result of FN()
	int fl_step;		//end step operation flag
	int mid;		//module ID
	struct pro_sch_iteration *next;
} sch_iteration;

static sch_iteration *sch_root=NULL, *sch_curr=NULL;

/*Global variables*/
double nce_time_cur, nce_time_step;
int nce_sw = 0, nce_asw = 0, nce_bufdeep;
sig_atomic_t sim_flag = 0;

inline static int sim_add_sch_iter(int mid, int result, int *fl_step, int (*fn)() )
{
	if(sch_root == NULL)
		sch_root = sch_curr = (sch_iteration *) mem_malloc ( sizeof(sch_iteration) );
	else if(sch_curr == NULL){
		for(sch_curr = sch_root; sch_curr->next != NULL; sch_curr = sch_curr->next);
		sch_curr->next = (sch_iteration *) mem_malloc ( sizeof(sch_iteration) );
		sch_curr = sch_curr->next;
	}
	else if(sch_curr->next != NULL){
		for(sch_curr = sch_root; sch_curr->next != NULL; sch_curr = sch_curr->next);
		sch_curr->next = (sch_iteration *) mem_malloc ( sizeof(sch_iteration) );
		sch_curr = sch_curr->next;
	}
	else{
		sch_curr->next = (sch_iteration *) mem_malloc ( sizeof(sch_iteration) );
		sch_curr = sch_curr->next;
	}
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_add_sch_iter", "Cannot malloc memory for new schedule item",NULL, -1);
	sch_curr->next = NULL;
	sch_curr->fn = fn;
	sch_curr->result = result;
	sch_curr->fl_step = *fl_step;
	sch_curr->mid = mid;
	*fl_step = 0;
	return 0;
}

int sim_init()
{
	sim_flag = 0;
	int all_overstep = 1, min_priority = 0;
	int step_cnt, prior_cnt, iter_cnt, fl_step;
	mip modroot = mod_get_list(), modscan;
	modscan = modroot;
	if(modscan == NULL) return ERROR_MESSAGE("sim_init","mod_get_list() return NULL",NULL,-1);
	while(modscan != NULL){
		if(!mod_is_acctive(modscan->mid)){
			mip moddel = modscan; modscan = modscan->next;
			WARNING_MESSAGE("sim_init","Remove unused module",NULL,moddel->mid);
			if(mod_remove(moddel->mid) < 0)
				return ERROR_MESSAGE("sim_init","mod_remove() return error. Cannot remove unused module.",NULL,-3);
			continue;
		}
		if(modscan->overstep < 0){
			WARNING_MESSAGE("sim_init","module overstep below zero. Set by defaule 0",NULL,modscan->mid);
			modscan->overstep = 0;
		}
		all_overstep *= ( modscan->overstep + 1);
		if(modscan->priority > min_priority) min_priority = modscan->priority;
		modscan = modscan->next;
	}

	modscan = modroot;
	int step_max_iter[all_overstep];
	//count max iteration in steps
	for(step_cnt = 0; step_cnt < all_overstep ; step_cnt++){
		step_max_iter[step_cnt] = 0;
		for(modscan = modroot; modscan != NULL; modscan = modscan->next){
			if( step_cnt % (modscan->overstep + 1) ) continue;
			if(step_max_iter[step_cnt] < modscan->iterations)
				step_max_iter[step_cnt] = modscan->iterations;
		}
	}
	//fill table
	for(step_cnt = 0, fl_step=1; step_cnt < all_overstep ; step_cnt++, fl_step=1){
		for(iter_cnt = 0; iter_cnt < step_max_iter[step_cnt]; iter_cnt++){
			for(prior_cnt=0; prior_cnt <= min_priority; prior_cnt++){
				for(modscan = modroot; modscan != NULL; modscan = modscan->next){
					if(modscan->priority != prior_cnt) continue;
					if( step_cnt % (modscan->overstep + 1) ){
						if( (iter_cnt == 0) && (sim_add_sch_iter(modscan->mid,0, &fl_step,modscan->jump) < 0) )
								return ERROR_MESSAGE("sim_init", "sim_add_sch_iter() return error", NULL, -4);
					}
					else {
						if(modscan->iterations > iter_cnt){ 
							if(sim_add_sch_iter(modscan->mid,(modscan->iterations - iter_cnt - 1), &fl_step,modscan->calculate) < 0) 
								return ERROR_MESSAGE("sim_init", "sim_add_sch_iter() return error", NULL, -5);
						}
						else{
							if(sim_add_sch_iter(modscan->mid,0, &fl_step,modscan->slack) < 0) 
								return ERROR_MESSAGE("sim_init", "sim_add_sch_iter() return error", NULL, -6);

						}
					}
				}
			}
		}
	}
	sch_curr->next = sch_root;
	sch_curr = sch_root;
	nce_time_cur = 0.0;
	nce_sw = 0; nce_asw = 1;
	return 0;
}

int sim_free()
{
	if(sch_root == NULL)
		return WARNING_MESSAGE("sim_free", "Scheduller is empty", NULL, 0);
	sch_iteration *sch_dell = sch_root;
	sch_curr = sch_root;
	while(sch_dell != sch_root){
		sch_curr = sch_dell->next;
		mem_free(sch_dell);
		sch_dell = sch_curr;
	}
	sch_curr = sch_root = NULL;
	return 0;
}

int sim_step()
{
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_step", "Scheduller is empty", NULL, -1);
	char error[201]; memset(error,0,201);
	do{
		if( sch_curr->result != (*sch_curr->fn)(sch_curr->mid) ){
			snprintf(error,200,"Module id=\"%06d\" return unexpected result!",sch_curr->mid);
			sim_flag = 0;
			return ERROR_MESSAGE("sim_step", error, NULL, -2);
		}
		sch_curr = sch_curr->next;
	}while(!sch_curr->fl_step);
	watch();
	if( (asm_model_grow_pointer == NULL) || ( (*asm_model_grow_pointer)() < 0) ){
		sim_flag = 0;
		return ERROR_MESSAGE("sim_step", "model_grow() return ERROR", NULL, -3);
	}
	
	nce_time_cur += nce_time_step;
	nce_sw++; nce_asw++;
	if( nce_sw >= nce_bufdeep)  nce_sw =0;
	if(nce_asw >= nce_bufdeep) nce_asw =0;
	sim_flag = 0;
	return 0;
}

int sim_go(int numberSteps)
{
	int cnt=0;
	char error[201], arg[20]; memset(error,0,201); snprintf(arg,19,"%d",numberSteps);
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_go", "Scheduller is empty", arg, -1);
	struct rusage st, ed;
	getrusage(RUSAGE_SELF, &st);
	sim_flag = 1;
	while( (cnt < numberSteps) &&  sim_flag ){
		if( sch_curr->result != (*sch_curr->fn)(sch_curr->mid) ){
			snprintf(error,200,"Module id=\"%06d\" return unexpected result!",sch_curr->mid);
			return ERROR_MESSAGE("sim_go", error, arg, -2);
		}
		sch_curr = sch_curr->next;
		if( sch_curr->fl_step ){
			cnt++;
			watch();
			if( (asm_model_grow_pointer == NULL) || ( (*asm_model_grow_pointer)() < 0) ){
				sim_flag = 0;
				return ERROR_MESSAGE("sim_step", "model_grow() return ERROR", NULL, -3);
			}
			nce_time_cur += nce_time_step;
			nce_sw  = ( ++ nce_sw ) % nce_bufdeep;
			nce_asw = ( ++nce_asw ) % nce_bufdeep;
		}
	}
	sim_flag = 0;
	getrusage(RUSAGE_SELF, &ed);
	return (int)((((double)ed.ru_utime.tv_sec * 1e6 + (double)ed.ru_utime.tv_usec) - ((double)st.ru_utime.tv_sec * 1e6 + (double)st.ru_utime.tv_usec)) *1e-3);
}

int sim_until(double untilTime)
{
	char error[201], arg[40]; memset(error,0,201); snprintf(arg,39,"%lf",untilTime);
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_until", "Scheduller is empty", arg, -1);
	if( (nce_time_cur > untilTime) && (untilTime > 0.0) )
		return WARNING_MESSAGE("sim_until","Order empty timing",arg,0);
	struct rusage st, ed;
	getrusage(RUSAGE_SELF, &st);
	sim_flag = 1;
	while( ( (nce_time_cur < untilTime) || (untilTime < 0.0) ) && sim_flag ){
		if( sch_curr->result != (*sch_curr->fn)(sch_curr->mid) ){
			snprintf(error,200,"Module id=\"%06d\" return unexpected result!",sch_curr->mid);
			sim_flag = 0;
			return ERROR_MESSAGE("sim_until", error, arg, -2);
		}
		sch_curr = sch_curr->next;
		if( sch_curr->fl_step ){
			watch();
			if( (asm_model_grow_pointer == NULL) || ( (*asm_model_grow_pointer)() < 0) ){
				sim_flag = 0;
				return ERROR_MESSAGE("sim_step", "model_grow() return ERROR", NULL, -3);
			}
			nce_time_cur += nce_time_step;
			nce_sw  = ( ++ nce_sw ) % nce_bufdeep;
			nce_asw = ( ++nce_asw ) % nce_bufdeep;
		}
	}
	sim_flag = 0;
	getrusage(RUSAGE_SELF, &ed);
	return (int)((((double)ed.ru_utime.tv_sec * 1e6 + (double)ed.ru_utime.tv_usec) - ((double)st.ru_utime.tv_sec * 1e6 + (double)st.ru_utime.tv_usec)) *1e-3);
}

int sim_start(double continueTime)
{
	char error[201], arg[40]; memset(error,0,201); snprintf(arg,39,"%lf",continueTime);
	double untilTime = nce_time_cur + continueTime;
	int ret;
	if( (ret = sim_until(untilTime)) < 0){
		sim_flag = 0;
		return ERROR_MESSAGE("sim_start","sim_until() returns an error",arg, -1);
	}
	return ret;
}

int sim_stop()
{
	return sim_flag = 0;
}

int sim_db_iter()
{
	char error[201];
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_db_iter", "Scheduller is empty", NULL, -1);
	if( sch_curr->result != (*sch_curr->fn)(sch_curr->mid) ){
		snprintf(error,200,"Module id=\"%06d\" return unexpected result!",sch_curr->mid);
		sim_flag = 0;
		return ERROR_MESSAGE("sim_db_iter", error, NULL, -2);
	}
	sch_curr = sch_curr->next;
	if( sch_curr->fl_step ){
		watch();
		if( (asm_model_grow_pointer == NULL) || ( (*asm_model_grow_pointer)() < 0) ){
			sim_flag = 0;
			return ERROR_MESSAGE("sim_db_iter", "model_grow() return ERROR", NULL, -3);
		}
		nce_time_cur += nce_time_step;
		nce_sw  = ( ++ nce_sw ) % nce_bufdeep;
		nce_asw = ( ++nce_asw ) % nce_bufdeep;
	}
	return (sim_flag = 0);
}

int sim_db_mod_iter(int mid)
{
	char error[201], arg[40]; memset(error,0,201); snprintf(arg,39,"%d",mid);
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_db_mod_iter", "Scheduller is empty", arg, -1);
	if(!mod_is_acctive(mid))
		return WARNING_MESSAGE("sim_db_mod_iter", "Module isn't active", arg, -2);
	sim_flag = 1;
	do{
		if( sch_curr->result != (*sch_curr->fn)(sch_curr->mid) ){
			snprintf(error,200,"Module id=\"%06d\" return unexpected result!",sch_curr->mid);
			sim_flag = 0;
			return ERROR_MESSAGE("sim_db_mod_iter", error, arg, -3);
		}
		sch_curr = sch_curr->next;
		if( sch_curr->fl_step ){
			watch();
			if( (asm_model_grow_pointer == NULL) || ( (*asm_model_grow_pointer)() < 0) ){
				sim_flag = 0;
				return ERROR_MESSAGE("sim_db_mod_iter", "model_grow() returns an ERROR", arg, -4);
			}
			nce_time_cur += nce_time_step;
			nce_sw  = ( ++ nce_sw ) % nce_bufdeep;
			nce_asw = ( ++nce_asw ) % nce_bufdeep;
		}
	}while( sch_curr->mid != mid && sim_flag );
	sim_flag = 0;
	return 0;
}

int sim_db_next_mod_iter()
{
	char error[201];
	if(sch_curr == NULL)
		return ERROR_MESSAGE("sim_db_next_mod_iter", "Scheduller is empty", NULL, -1);
	if(sim_db_mod_iter(sch_curr->mid) < 0)
		return ERROR_MESSAGE("sim_db_next_mod_iter", "sim_db_mod_iter() returns an error", NULL, -2);
	return 0;
}


int sim_printinfo()
{
	OUTPUT_MESSAGE("\n=== Simulation scheduller information ===\n");
	if(sch_root == NULL){
		OUTPUT_MESSAGE("Scheduller root is empty\n\n");
		return 0;
	}
	else
		OUTPUT_MESSAGE("\n");
	int step = 1;
    	sch_iteration *sch_print = sch_root;
	OUTPUT_MESSAGE("\n\tMIDd\tOVERSTEP\tPRIOR\tITER\n");
	do{
		if(sch_print->fl_step) 
			OUTPUT_MESSAGE( "Step %d\n", step++ );
		mip mod = mod_get_item(sch_print->mid);
		if( mod == NULL)
			return ERROR_MESSAGE("sim_print_table","mod_get_item() returns an error",NULL,-2);
		OUTPUT_MESSAGE("\t%d\t%d\t%d\t%d",sch_print->mid, mod->overstep, mod->priority, mod->iterations);
		if(sch_print->fn == mod->calculate)
			OUTPUT_MESSAGE("\tCALCULATE\t%d\n",sch_print->result);
		else if(sch_print->fn == mod->slack)
			OUTPUT_MESSAGE("\tNO LOAD  \t%d\n",sch_print->result);
		else if (sch_print->fn == mod->jump)
			OUTPUT_MESSAGE("\tJUMPING  \t%d\n",sch_print->result);
		else
			OUTPUT_MESSAGE("\t_UNKNOWN_\t%d\n",sch_print->result);
	} while( (sch_print = sch_print->next) != sch_root);
	return 0;
}
