#ifndef MODMANAGER_H
#define MODMANAGER_H

#include "nce_common.h"
#include <neurocad/module.h>

extern int mod_number;

/**
Structure `moditem' holds handle, some information, attached variables and function for each 
inserted module library (*.so). The structures have lined up to the linked list by pointer next.
**/
typedef struct pro_moditem {
	char	*filename;
	void	*hmod;
	int 	isused;
	int	overstep;
	struct pro_moditem *next;
	//module parameters
	enum ml multiload;
	int	mid;
	int	allnames;
	int	iterations;
	int	priority;
	int	blocksize;
	datum	weight;
	nameitem *names;
	//module functions
	int	(*init)(int);
	int	(*release)();
	module*	(*add)( int, module *, void *);
	int 	(*calculate)( int );
	int	(*slack)( int );
	int	(*jump)( int );
	int	(*restore)(int, int );
	int	(*save)(int, int );
} moditem;

typedef moditem * mip; //module item poiter

/**
mod_insert() attaches and inserts the module `mod_file_name'in list, marking it as`mid'.
It try to connect all needed functions from the module library and return zero if result
is successful. In other cases function returs code of error (a negative number).
**/
int mod_insert(char *mod_file_name, int mid);

/**
mod_insert_empty() isn't attach any file and insert empty `moditem' with `mid' in modules list.
Use it if you would't like exploit external module library and prefer to apply internal functions.
Be careful! The all pointers in `moditem' structure will be set in NULL, therefore you have to init 
all pointer before start simulation and using the model. Function return zero if result is successful.
In other cases function returs code of error (a negative number).
**/
int mod_insert_empty(int mid);

/**
mod_remove() removes module morked `mid' from linked list of modules. Function return zero if result 
is successful. In other cases function returs code of error (a negative number).
**/
int mod_remove(int mid);

/**
mod_removeall() cleans up the modules list and returns zero in successful result
**/
int mod_removeall();

/**
mod_get_number() returns all modules amount
**/
int mod_get_number();

/**
mod_get_item() returns the pointer of moditem structure marked as `mid' 
or NULL if such item isn't found.
**/
mip mod_get_item(int mid);

/**
mod_get_list() returns the root of modules linked list
**/
mip mod_get_list();

/**
mod_init() initializes the module makred as `mid'. Returns zero if successfully
**/
int mod_init(int mid);


/**
mod_is_acctive() return TRUE(one) if module already initialized, FALSE(zero) if 
module hasn't initialized and value under zero in erroneous cases.
**/
int mod_is_acctive(int mid);


/**
mod_set_extended_attribute() set external attribute named `name' of module marked as `mid'
in value equal `value'  Function return zero if result is successful. In other cases 
function returs code of error (a negative number).
**/
int mod_set_extended_attribute(int mid, char *name, char *value);

/**
mod_flash_extended_attribute() calls the flashing function of module marked as `mid'.
The flash is constrains modules to fix external attributes and applay it. Function 
return zero if result is successful.
**/
int mod_flash_extended_attribute(int mid);


/**
mod_printinfo() prints out information about modmanager and always return zero
**/
int mod_printinfo();
#endif

